/*----------------------------------------------------------
	STOOD v5.0
	AADL rules checker 
	Legality rules: AADL 1.0 - SAE AS5506

	Copyright TNI Europe & Ellidiss Technologies 2005
	Pierre DISSAUX (pierre.dissaux@tni-world.com)
-------------------------------------------------------------*/
/**** changes ****/
/*-----------------------------------------------------------
initial version: 18 Mar 2005
-------------------------------------------------------------*/
 

/* 4.2 Packages */
/*
L001	The defining package name following the reserved word end must be identical 
	to the defining package name following the reserved word package.
L002	For each package there may be at most one public section declaration 
	and one private section declaration. These two sections may be declared 
	in a single package declaration or in two separate package declarations.
*/

/* 4.3 Component types */
/* 
L003	The defining identifier following the reserved word end must be identical 
	to the defining identifier that appears after the component category reserved word.
L004	The features, flows, and properties subclauses are optional. 
	If a subclause is present but empty, then the reserved word none 
	followed by a semi-colon must be present.
L005	A component type declaration that does not extend another component type 
	must not contain feature refinement declarations. 
L006	The category of the component type being extended must match 
	the category of the extending component type.
*/

/* 4.4 Component implementations */
/*
L007	The pair of identifiers separated by a dot (".") following the reserved word end 
	must be identical to the pair of identifiers following the reserved word implementation.
L008	The refines type, subcomponents, connections, calls, flows, modes, and properties 
	subclauses are optional. If they are present and the set of feature or required 
	subcomponent declarations or property associations is empty, none followed by a 
	semi-colon must be present in that subclause. 
L009	The category of the component implementation must match the category of the component 
	type for which the component implementation is declared.
L010	If the component implementation extends another component implementation, 
	the category of both must match.
L011	The refines type subclause must only contain refinement declarations of features 
	in the component type and those refinements are limited to property associations. 
	Specifically, the refines type subclause of a component implementation may not 
	alter the component classifiers of inherited features.
*/

/* 4.5 Subcomponents */
/*
L012	The category of the subcomponent declaration must be identical to the category of 
	its corresponding component classifier reference.
L013	The component type named in the component classifier reference of a subcomponent refinement 
	must be the component type of the subcomponent being refined if the subcomponent being 
	refined has a component type declared.  
L014	The component implementation named in the component classifier reference of a subcomponent 
	refinement must be the component implementation of the subcomponent being refined if the 
	subcomponent being refined has a component type declared.
L015	If the subcomponent declaration contains an in modes statement and any of its property 
	associations also contains an in_modes statement, then the modes named in the property 
	association must be a subset of those named in the subcomponent declaration.  
*/

/* 4.6 Annexe subclauses and annex libraries */
/*
L016	Annex subclauses can only be declared in component types and component implementations.
L017	A component type or component implementation declaration may contain at most one annex 
	subclause for each annex.
L018	Annex libraries can only be declared in packages.
L019	A package declaration may contain at most one annex library declaration for each annex.
*/

/* 5.1 Data */
/*
L020	A data type declaration can contain subprogram declarations, provides data access declarations 
	as well as property associations. 
L021	A data type declaration must not contain a flow specification.
L022	A data implementation can contain data subcomponents, a modes subclause, access connections, 
	and data property associations.
L023	A data implementation must not contain a flow implementation or an end-to-end flow specification. 
L024	Each requires data access reference may have its own Required_Access property value.  
	This property value must not conflict with the Provided_Access property value associated with 
	the data component or the corresponding provides access declaration.
L025	The data classifier references of two data ports, event data ports, data access, or parameters 
	to be identical. When a data type has zero ore one implementation, then the referenced data 
	types must match. When a data type has more than one implementation, both the data type and 
	the appropriate implementation must be present in the data classifier reference of a data port, 
	event data port, data access, or parameter declaration.
*/

/* 5.2 Subprograms and subprogram calls */
/*
L026	A subprogram type declaration can contain parameter, out event port, out event data port, and 
	port group declarations as well as required data access declarations.  It can also contain 
	a flow specification subclause as well as property associations.
L027	A subprogram implementation can contain a connections subclause, a subprogram calls subclause, 
	a flows subclause, a modes subclause and subprogram property associations.
L028	A subprogram implementation must not contain a subcomponent subclause.
L029	Only one subprogram call sequence can apply to a given mode. In other words, a mode identifier 
	can be specified in the in_modes subclause of at most one subprogram call sequence.
*/

/*  5.3 Threads */
/*--------------------------------------------------------------------------------------------------
L030	A thread type declaration can contain port, port group, server subprogram declarations as 
	well as requires and provides data access declarations. It can also contain flow 
	specifications and property associations.
----------------------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------------------
L031	A thread component implementation can contain data declarations, a calls subclause, a flows 
	subclause, a modes subclause, and thread property associations.
----------------------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------------------
L032	A thread or any of its features must not contain an explicit Dispatch in event or event data 
	port declaration, nor a Complete or Error out event or event data port declaration.
----------------------------------------------------------------------------------------------------*/

errL032(X,T,E) :- isTerm(X), ( T='ACTIVE'; T='CYCLIC'; T='SPORADIC' ), 
	isProvided(E,K,X), ( 
	( K='OPERATION', memezTra(E,'Dispatch') );
	( K='EXCEPTION', memezTra(E,'Complete') );
	( K='EXCEPTION', memezTra(E,'Error') ) ).

checkL :- hoodObject(X,T,_), errL032(X,T,E),
	write('ERROR : Thread '), write(X), 
	write(' should not contain such a port named : '), write(E), 
	write(' (L032)'), nl.
checkL :- printlf('---> rule L032 checked...').

/*--------------------------------------------------------------------------------------------------
L033	The Compute_Entrypoint property must have a value that indicates the source code to execute 
	after a thread has been dispatched. Other entrypoint properties are optional, i.e., if a 
	property value is not defined, then the entrypoint is not called. 
----------------------------------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------------------------------
L034	The Period property must have a value if the Dispatch_Protocol property value is periodic or 
	sporadic.
----------------------------------------------------------------------------------------------------*/

errL034(X,'CYCLIC') :- not(hrtPeriod(X,_,_)).
errL034(X,'SPORADIC') :- not(hrtMinArrivalTime(X,_,_)).

checkL :- hoodObject(X,T,_), errL034(X,T),
	write('ERROR : A Period property value should be specified for Thread '), write(X), 
	write(' (L034)'), nl.
checkL :- printlf('---> rule L034 checked...').


/* 5.4 Thread groups */
/*
L035	A thread group component type can contain provides and requires data access, as well as port, 
	port group, and server subprogram declarations. It can also contain flows specifications and 
	property associations.
L036	A thread group component implementation can contain data, thread, and thread group declarations.
L037	An instantiable thread group component implementation must contain at least one thread 
	subcomponent or one thread group subcomponent.  
L038	A thread group implementation can contain a connections subclause, a flows subclause, 
	a modes subclause, and properties subclause. 
L039	A thread group must not contain a subprogram calls subclause.
*/

/* 5.5 Processes */
/*
L040	A process component type can contain port, port group, provides and requires data access, 
	and server subprogram declarations. It can also contain flow specifications and property 
	associations.
L041	A process component implementation can contain data, thread, and thread group declarations.
L042	A complete process component implementation must contain at least one thread subcomponent 
	or one thread group subcomponent. 
L043	A process implementation can contain a connections subclause, a flows subclause, 
	a modes subclause, and a properties subclause.  
L044	The complete source text associated with a process component must form a complete and legal 
	program as defined in the applicable source language standard. This source text shall include 
	the source text that corresponds to the complete set of subcomponents in the processs 
	containment hierarchy along with the data and subprograms that are referenced by required 
	subcomponent declarations. 
*/

/* 6.1 Processors */
/*
L045	A processor component type can contain port, port group, server subprogram, and required bus 
	access declarations. It may contain flow specifications as well as property associations.
L046	A processor component implementation can contain declarations of memory subcomponents.  
L047	A processor implementation can contain a modes subclause, flows subclause, and a properties 
	subclause.  
L048	A processor implementation must not contain a connection subclause, or a subprogram calls 
	subclause.
L049	A processor component must contain at least one memory component or require at least one 
	bus access.
*/

/* 6.2 Memory */
/*
L050	A memory type can contain requires bus access declarations and property associations. It must 
	not contain flow specifications.
L051	A memory implementation can contain memory subcomponent declarations.  
L052	A memory implementation can contain a modes subclause and property associations.
L053	A memory implementation must not contain a connections subclause, flows subclause, or 
	subprogram calls subclause.
*/

/* 6.3 Buses */
/*
L054	A bus type can have requires bus access declarations and property associations.
L055	A bus type must not contain any flow specifications.
L056	A bus implementation must not contain any subcomponent declarations.  
L057	A bus implementation can contain a modes subclause and property associations.  
L058	A bus implementation must not contain a connections subclause, flows subclause, 
	or subprogram calls subclause.
*/

/* 6.4 Devices */
/*
L059	A device type can contain port, port group, server subprogram, and requires bus access 
	declarations, flow specifications, as well as property associations.
L060	A device component implementation must not contain a subcomponents subclause, connections 
	subclause, or subprogram calls subclause.  
L061	A device implementation can contain a modes subclause, a flows subclause, and property 
	associations.  
*/

/* 7.1 Systems */
/*
L062	A system component type can contain provides and requires data and bus access declarations, 
	port, port group, and server subprogram declarations. It can also contain flow specifications 
	as well as property associations.
L063	A system component implementation can contain data, process, and system subcomponent 
	declarations as well as execution platform components, i.e., processor, memory, bus, and device.
L064	A system implementation can contain a modes subclause, a connections subclause, a flows 
	subclause, and property associations.  
*/

/* 8. Features and shared access */
/*
L065	The feature classifier reference and the port direction in a refined feature declaration must 
	be identical to the feature classifier reference and the port direction in the refined declaration.  
L066	In the case of data and event data ports, the refined feature declaration in a component type 
	extension can complete an incomplete data classifier reference.  
L067	Feature refinements can associate new property values.
L068	Each feature can be refined at most once in the same component implementation or type extension.
*/

/* 8.1 Ports */
/*
L069	Ports can be declared in subprogram, thread, thread group, process, system, processor, 
	and device component types.
L070	Data and event data ports may be incompletely defined by not specifying the data component 
	classifier reference or data component implementation identifier of a data component classifier 
	reference. The port definition can be completed using refinement.
L071	Data and event data ports may be refined by adding a property association. The data component 
	classifier declared as part of the data or event data port declaration being refined does not 
	need to be included in this refinement.
L072	The property names Overflow_Handling_Protocol, Queue_Processing_Protocol, Dequeue_Protocol and 
	Queue_Size may only appear in property associations for in event ports and in event data ports.
*/

/* 8.2 Port groups and port group types */
/*
L073	A port group type may contain zero or more elements, i.e., ports or port groups. If it contains 
	zero elements, then the port group type may be declared to be the inverse of another port group 
	type. Otherwise, it is considered to be incompletely specified. 
L074	A port group type can be declared to be the inverse of another port group type, as indicated by 
	the reserved words inverse of and the name of a port group type. Any port group type named in 
	an inverse of statement cannot itself contain an inverse of statement. This means that several 
	port groups can be declared to be the inverse of one port group. However, chaining of inverses 
	such as B inverse of A and C inverse of B is not permitted.  
L075	A port group type that is an extension of another port group type cannot contain an inverse of 
	statement. The port group type being extended cannot contain an inverse of statement.
L076	Two port group types are considered to complement each other if the following holds: 
	- The number of ports or port groups contained in the port group and its complement must 
	  be identical.
	- Each of the declared ports or port groups in a port group must be a pair-wise complement 
	  with that in the port group complement, with pairs determined by declaration order.  
L077	If both port group types have zero features, then they are considered to complement each other.  
L078	In the case of port group type extensions, the port and port group declarations in the extension 
	are considered to be after the declarations in the port group type being extended.  
L079	Ports are pair-wise complementary if they have complementary direction ( out / in, in / out, 
	in out / in out), and are of the same port type. In the case of event data or data ports, 
	the data component classifier reference must be identical.
L080	A port group declaration that does not specify a port group type reference is incomplete. 
	Such a reference can be added in a port group refinement declaration.
L081	A port group declaration may be refined by adding a property association. Inclusion of the 
	port group type reference is optional.
L082	If the Aggregate_Data_Port property of a port group has the value true, all ports contained 
	in its port group type or the port group type of any contained port group must be data ports 
	and they must all have the same port direction.
*/

/* 8.3 Subprograms as features */
/*
L083	Data subprogram features can be declared in data components and must not have the reserved 
	word server.
L084	Server subprogram features can be declared in thread, thread group, process, processor, and 
	system component types. They must have the reserved word server.
L085	A server subprogram feature declaration must only refer to a subprogram classifier or to a 
	subprogram feature in a data component type.
L086	If several subprogram declarations refer to the same subprogram type or via the Source_Name 
	property to the same subprogram in the source text, then their parameter signatures in the 
	source text and the property associations must be consistent with each other.
L087	When a server subprogram declaration appears in a thread component type, the scheduling 
	protocol property value for all thread implementations, subcomponents, and instances having 
	that component type must be Aperiodic or Sporadic.
L088	A subprogram refinement can specify a subprogram type reference and declare property associations.
L089	For calls to server subprograms, the subprogram classifier or subprogram feature reference 
	of the subprogram call and the subprogram classifier or subprogram feature reference of the 
	server subprogram must be the same.  
*/

/* 8.4 Subprogram parameters */
/*
L090	Parameters can be declared for subprogram component types.
L091	If a parameter refinement includes a data classifier reference, then the classifier reference 
	must be the same as that of the parameter being refined.
L092	If the parameter being refined has an incomplete data classifier reference, then the parameter 
	refinement may complete an imcompletely specified data classifier reference.
L093	A parameter refinement cannot redefine the direction of a parameter.
*/

/* 8.5 Subcomponent access */
/*
L094	The category of the subcomponent access declaration must be identical to the category of the 
	component type (and of the component implementation) in the referenced subcomponent classifier.
*/

/* 9.1.1 Port connections */
/*
L095	The ultimate source of a semantic port connection must be a feature of a thread, processor, 
	or device.  
L096	The source feature referenced in a port connection declaration must be a feature of a thread, 
	thread group, process, processor, device, or system component.  
L097	The ultimate destination of a semantic port connection must be a port of a thread, a processor, 
	a device.
L098	If the ultimate destination is the result of a mode transition, the mode change is indicated 
	by the mode subclause of the respective thread, thread group, process, system, device, bus, 
	memory, or processor naming an in event port in one of its mode transitions.  
L099	The destination feature referenced in a port connection declaration must be a feature of 
	a thread, thread group, process, processor, device, or system component. 
L100	One end of the connection must be a thread. The other end may be a processor, or a device, 
	or a thread.
L101	If the ultimate destination of a semantic port connection is the result of a mode transition, 
	then the ultimate source must be an out event port.
L102	If a semantic port connection is declared to apply to a particular mode, then the ultimate 
	source and ultimate destination components must be part of that mode.  
L103	If a semantic port connection is declared to apply to a particular mode transition, then the 
	ultimate source component must be part of a system mode that includes the old mode identifier 
	and the ultimate destination component must be part of a system mode that includes the new 
	mode identifier.
L104	The category of the port connection declaration must match the source and destination features 
	as described in the following paragraphs. This implies that all connection declarations of a 
	semantic connection must be of the same category.
L105	The direction declared for the destination feature of a port connection declaration must be 
	compatible with the direction declared for the source feature(s) as defined by the following rules.:  
	- If the port connection declaration represents a connection between sibling components, then 
	  the source must be an out or an in out port and the destination must be an in or an in out port, 
	  or in the case of port group connections the source and destination port groups must be 
	  complements of each other.
	- If the port connection declaration represents a connection between elements of two port groups 
	  in the component type, then source must be an in or an in out port and the destination must be 
	  an out or an in out port, or in the case of port group connections the source and destination 
	  port groups must be complements of each other.
	- If the port connection declaration represents a connection up or down the containment hierarchy, 
	  then the source and destination must both be an out or an in out port, or both an in or in out 
	  port, or in the case of port group connections the port groups of the same port group type.  
L106	A data port cannot be the destination feature reference of more than one port connection 
	declaration unless each port connection declaration is (are) contained in a different mode. 
	In this case, the restriction applies for each mode. 
L107	The ultimate source and ultimate destination of a delayed port connection must be periodic threads. 
L108	For data and event data port connections, the data classifier of the source port must be identical 
	to the data type of the destination port.  
L109	If more than one port connection declaration in a semantic port connection has a property 
	association for a given connection property, then the resulting property values must be identical.
L110	For port group connections the following must hold.:
	- If the connection declaration represents a component connection between sibling components, 
	  the port group types must be complements as indicated with the inverse of statement in one 
	  of the two port group types.  
	- If the connection declaration represents a connection up or down the containment hierarchy, 
	  the port group types must be identical. 
*/

/* 9.1.2. Parameter connections */
/*
L111	Parameter connections must adhere to the following rule regarding their source and destination:
	- the source must be an in data or event data port or in out data or event data port of the 
	  containing thread and the destination a subprogram call in or in out parameter, 
	- the source must be an in parameter or in out parameter of the containing subprogram and the 
	  destination a subprogram call in or in out parameter, 
	- the source must be a subprogram call out or in out parameter and the destination a out or 
	  in out parameter of the containing subprogram, 
	- the source must be a subprogram call out or in out parameter and the destination a subprogram 
	  call in or in out parameter or an out or in out data or event data port of the containing thread.
L112	If the parameter connection declaration represents a parameter connection between sibling 
	components, then the source must be an out or an in out parameter and the destination must be an 
	in or an in out parameter. Furthermore, the source must be a parameter of a preceding subprogram 
	call in the call sequence, and the destination must be a parameter of a succeeding subprogram 
	call in the call sequence.
L113	If a parameter connection is declared to apply to a particular mode, then the source and 
	destination must be part of that mode.  
L114	A parameter cannot be the destination feature reference of more than one parameter connection 
	declaration unless the source feature reference(s) of each parameter connection declaration is 
	(are) contained in a different mode. In this case, the restriction applies for each mode. 
L115	The data classifier of the source port or parameter must be identical to the data type of the 
	destination port or parameter.  
*/

/* 9.1.3 Access connections */
/*
L116	All access declarations forming a semantic data access connection must be data access declarations.  
L117	All access declarations forming a semantic bus access connection must be bus access declarations.  
L118	The ultimate source of a semantic access connection must be data or bus subcomponent.  
L119	The ultimate destination of a semantic data access connection must be a requires data access 
	feature of a thread or a subprogram call without a containing subprogram call requiring the same 
	data access.  
L120	The ultimate destination of a semantic bus access connection must be a requires bus access 
	feature of a processor, memory, or device subcomponent.
L121	If a semantic access connection is declared to apply to a particular mode, then the ultimate 
	source and ultimate destination must be part of that mode.  
L122	For access connections between access features, the direction declared for the destination 
	feature must be compatible with the direction declared for the source feature(s) as defined 
	by the following rules:.  
	- If the access connection declaration represents an access connection between access features 
	  of sibling components, then the source must be a provides access or a data or bus component 
	  and the destination must be a requires access.  
	- If the access connection declaration represents a feature mapping up or down the containment 
	  hierarchy, then the source and destination must both be a requires access, both be a provides 
	  access, or the source a data or bus subcomponent and the destination a provides access.  
L123	A requires access cannot be the destination feature reference of more than one access connection 
	declaration unless the source feature reference(s) of each access connection declaration is (are) 
	contained in a different mode. In this case, the restriction applies for each mode. 
L124	For data access connections the data type of the provides access must be identical to the data 
	type of the requires access.  
L125	If more than one access feature in a semantic access connection has an access Required_Access 
	or Provided_Access property association, then the resulting property values must be compatible.  
	This means that the provider must provide read-only or read-write access if the requirer 
	specifies read-only. Similarly, the provider must provide write-only or read-write access if 
	the requirer specifies write-only. The provider must provide read-write access if the requirer 
	specifies read-write. Finally, the provider must provide by-method access if the requirer 
	specifies by-method access.
*/

/* 9.2.1 Flow specifications */
/*
L126	The direction declared for the destination of a flow path specification declaration must be 
	compatible with the direction declared for the source as defined by the following rules:
	- If the source is a port or parameter, its direction must be must be an in or an in out. 
	- If the destination is a port or parameter, its direction must be an out or an in out.
L127	The direction declared for the destination port or parameter of a flow source specification 
	declaration must be out or in out.
L128	The direction declared for the source port or parameter of a flow source specification 
	declaration must be in or in out.
*/

/* 9.2.2 Flow implementations */
/*
L129	The source of a connection named in a flow implementation declaration must be the same as the 
	source flow feature of the flow implementation or as the destination of the directly preceding 
	subcomponent flow specification.  
L130	The destination of a connection named in a flow implementation declaration must be the same as 
	the destination flow feature of the flow implementation or as the source of the directly 
	succeeding subcomponent flow specification.  
L131	If the component implementation provides mode-specific flow implementations, as indicated by the 
	in modes statement, then there must be a flow implementation for each of the modes.
L132	In case of a mode-specific flow implementation, the named connections and the subcomponents of 
	the named flow specifications must be declared for the modes listed in the in modes statement.
L133	In a complete specification, if a system, process, or thread group component implementation 
	contains a flow implementation declaration, then the flow implementation must include flow 
	specifications through named thread, processor, or device subcomponents.
L134	If the category of the component type containing a flow specification declaration is thread or 
	subprogram, and a component implementation of the component type does not contain subprogram 
	calls, then the flow specification represents its implementation and an explicit flow 
	implementation declaration is not required
*/

/* 9.2.3 End to end flows */
/*
L135	The flow specifications identified by the flow_path_subcomponent_flow_identifier must be flow 
	paths.  
L136	The start_subcomponent_flow_identifier must refer to a flow path or a flow source.  
L137	The end_subcomponent_flow_identifier must refer to a flow path or a flow sink.  
L138	In case of a mode-specific end-to-end flow declarations, the named connections and the 
	subcomponents of the named flow specifications must be declared for the modes listed in 
	the in modes statement.
*/

/* 10.1.1 Property types */
/*
L139	The value of the first numeric literal that appears in a range of a number_type must not be 
	greater than the value of the second numeric literal including the value's unit.
L140	Range values should always be declared with unit literals if the property requires a unit literal.
L141	The unique property constant identifier in an integer range must represent an integer constant.
L142	A boundless range type may be declared such that the actual range declarations have no limit on 
	the upper and lower bound.
L143	The unique property constant identifier in a real range must represent a real constant.
L144	If the property requires a unit, then the unit must be specified for lower and upper bound.
*/

/* 10.1.2 Property names */
/*
L145	The reserved word access is only permitted for property name declarations whose applies to 
	property category list contains categories of subcomponents that can be required or provided 
	subcomponents. These categories are data and bus.
*/

/* 10.1.3 Property constants */
/*
L146	If a property constant declaration has more than one property expression, it must contain the 
	reserved words list of.
L147	The property type of the property constants declaration must match the property type of the 
	constant  property value.
L148	If the constant property value is an integer or real value with a unit identifier, then the 
	property type specification of the property constant must include a units identifier.
*/

/* 10.2 Predeclared property sets */
/*
L149	The AADL_Properties property set cannot be modified.  
L150	Existing property type and property constant declarations in the AADL_Project property set can 
	be modified. New declarations must not be added to the AADL_Project property set, but can be 
	introduced through a separate property set declaration.
*/

/* 10.3 Property associations */
/*
L151	The property named by a property association must list the category of the component type, 
	component implementation, subcomponent, feature, connection, flow, or mode the property 
	association is declared for in its Property_Owner_Category list.  
L152	If a property association is declared within a package, the property value applies to all 
	component classifier declarations contained in the package for which the property is valid. 
L153	If a property expression list consists of a list of two or more property expressions, all of 
	those property expressions must be of the same property type.  
L154	If the property declaration for the associated property name does not contain the reserved words 
	list of, the property value must be a single property value. If the property declaration for 
	the associated property name contains the reserved words list of, the property value can be a 
	single property value, which is interpreted to be a list of one value.  
L155	The property association operator +=> must only be used if the property declaration for the 
	associated property name contains the reserved words list of. Furthermore, the property 
	association may not have an in modes or in binding statement.
L156	The property association operator +=> may not be used in contained property associations.
L157	In a property association, the type of the evaluated property expression must match the property 
	type of the named property.
L158	A property value declared by a property association with the reserved word constant cannot be 
	changed.
L159	The semantics rules below for determining the value of a property impose a precedence on the 
	property associations for a property. A property association with the reserved word constant 
	must be the highest priority association.
L160	The reserved word access is only permitted and is required in property associations declared in 
	required and provided access subcomponent declarations and refinements.
L161	The unique component type identifiers in the in binding statement must refer to component types 
	of the categories processor, memory, or bus.
L162	Property associations declared as part of a component type declaration, port group type 
	declaration, a feature or feature refinement declaration in a component type or port group type, 
	or as part of a feature refinement in the refines type clause of a component implementation are 
	not permitted to have an in modes statement as the scope of modes is limited to component 
	implementations. A feature refinement in the refines type clause of a component implementation 
	must not inherit a modal property association from its component implementation.
*/

/* 10.4 Property expressions */
/*
L163	If the base type of a property number type or range type is integer, then the numeric literals 
	must be integers.
L164	The type of a property named in a property term must match the type of the property name in 
	the property association.
L165	The type of a property constant named in a property constant term must match the type of the 
	property name in the property association.
L166	Property name references in property expressions cannot be circular. If a property has a property 
	expression that refers to a property name, then that propertys expression evaluation cannot 
	directly or indirectly depend on the value of the original property.
*/

/* 11.1 Modes */
/*
L167	A mode can be declared in data, thread, thread group, process, system, processor, bus, memory, 
	and device component implementations.
L168	If a component implementation contains mode declarations, one of those modes must be declared 
	with the reserved word initial. If the component implementation extends another component 
	implementation, the initial mode may have been declared in one of the ancestor component 
	implementations.
L169	The set of transitions declared within a single component implementation must define a 
	deterministic transition function. For each mode, there must exist exactly one transition 
	associated with a single event arrival, which can cause transition to another mode.  
L170	The unique port identifier must be either an in or in out event port identifier in the interface 
	namespace of the associated component type or an out or in out event port in the interface 
	namespace of the component type associated with the named subcomponent.  
*/

/* 12.1 System instances */
/*
L171	The system type specified for a system instance must not contain any required subcomponents. 
L172	A complete system instance must not contain incompletely specified subcomponents, ports, and 
	subprograms. All processes in a completely instantiable system instance must contain at least 
	one thread.
L173	In a complete system instance, the ports of all threads, devices, and processors must be the 
	ultimate source or destination of semantic connections. The Required_Connection property may be 
	used to indicate that a port connection is optional. In the case of the predeclared Error and 
	Complete ports, connections are optional.  
L174	In a completely instantiable system, the subprogram calls of all threads must either be local 
	calls or be bound to a server subprogram whose thread is part of the same mode.  
L175	In a completely intantiable system, for every mode that is the source of mode transitions, there 
	must be at least one mode transition that is the ultimate destination of a semantic connection 
	whose ultimate source is part of the mode.  
L176	In a complete system instance, aperiodic and sporadic threads that are part of a given mode 
	must have at least one connection to one of their in event ports or in event data ports or their 
	predeclared Dispatch port. The predeclared Dispatch port must not be connected if the thread has 
	a Dispatch_Protocol property value of periodic or background.
L177	For instantiable systems, all threads must be bindable to processors and all components 
	representing source text must be bindable to memory.
L178	The source text associated with all contained components of the system instance must be compliant 
	with the specified component type, component implementation, and property associations. 
L179	A system instance must contain at least one thread, one processor and one memory component in 
	its containment hierarchy to represent an application system that is executable on an execution 
	platform, i.e., a processor with memory containing the application code and data.
*/

/* 12.2 System binding */
/*
L180	Every mode-specific configuration of a system instance must have a binding of every process 
	component to a (set of) memory component(s), and a binding of every thread component to a 
	(set of) processor(s). In the case of dynamic process loading, the actual binding may change 
	at runtime. In the case of tightly coupled multi-processor configurations, the actual thread 
	binding may change between members of an actual binding set of processors as these processors 
	service a common set of thread ready queues.  
L181	Multiple software components may be bound to a single memory component. 
L182	A software component may be bound to multiple memory components.  
L183	A thread must be bound to a single processor. 
L184	Multiple threads can be bound to a single processor.
L185	All software components for a process must be bound to memory components that are all accessible 
	from every processor to which any thread contained in the process is bound. That is, every thread 
	is able to access every memory component into which the process containing that thread is loaded.
L186	A shared data component must be bound to memory accessible by all processors to which the 
	processes sharing the data component are bound.
L187	For all threads in a process, all processors to which those threads are bound must have identical 
	component types and component implementations. That is, all threads that are contained in a given 
	process must all be executing on the same kind of processor, as indicated by the processor 
	classifier reference value of the Allowed_Processor_Binding_Class property associated with the 
	process. Furthermore, all those processors must be able to access the memory to which the process 
	is bound.
L188	The complete set of software components making up the kernel address space of a processor must be 
	bound to memory that is accessible by that processor.
L189	Each thread must be bound to a processor satisfying the Allowed_Processor_Binding_Class and 
	Allowed_Processor_Binding property values of the thread. The Allowed_Processor_Binding property 
	may specify a single processor, thus specifying the exact processor binding. It may also specify 
	a list of processor components or system components containing processor components, indicating 
	that the thread is bindable to any of those processor components.
L190	Each process must be bound to a memory satisfying the Allowed_Memory_Binding_Class and 
	Allowed_Memory_Binding property values of the process. The Allowed_Memory_Binding property may 
	specify a single memory component, thus specifying the exact memory binding. It may also specify 
	a list of memory components or system components containing memory components, indicating that 
	the process is bindable to any of those memory components.
L191	The memory requirements of the binary images and the runtime memory requirements of threads and 
	processes bound to a memory component must not exceed that memorys capacity. 
	The execution time requirements of all threads bound to a processor must not exceed the 
	schedulable cycles required to insure that all thread timing requirements are met.  
	These two constraints may be checked statically or dynamically. 
	Runtime detection of such a memory capacity or timing requirements violation results in an error 
	that the application system can choose to recover from.
L192	The memory requirements of ports and data components are specified as property values of their 
	data types. Those property associations can have binding-specific values.
*/

/* 13.3 Identifiers */
/*
L193	An identifier must be distinct from the reserved words of the AADL.
*/

/* 13.4.2 Based literals */
/*
L194	The base (the numeric value of the decimal numeral preceding the first #) shall be at least 
	two and at most sixteen. The extended_digits A through F represent the digits ten through 
	fifteen respectively. The value of each extended_digit of a based_literal shall be less than 
	the base. 
*/