/*``changes``Wed Oct 23 15:57:10 MET 2002``*/
/* version */
/*------------------------------------------------
	STOOD v5.0
	HOOD 4 rules checker 
	Consistency rules: HRM4 p97-99

	Copyright TNI Europe 2004-2005
	Pierre DISSAUX (pierre.dissaux@tni-world.com)
------------------------------------------------*/
/**** changes ****/
/*------------------------------------------------
HOOD001 (17.May.2001) in C6, fixing pb with implemented_by sub classes
HOOD003 (17.May.2001) in C18, no more errors for abstract and inherited ops
HOOD005 (17.May.2001) in C5, removing C5_2 and same change as C6
HOOD007 (25.Jun.2001) in C5 and C6, ignoring environments
HOOD008 (23.Oct.2002) in C3, handling actual parameters of instances
HOOD009 (23.Oct.2002) in C19, fixing bug in detecting code for OBCS
HOOD010 (23.Oct.2002) in C26 and C27, only an error if terminal
HOOD020 (23.Dec.2004) in C15, allowing HRT constraints renaming
HOOD021 (23.Dec.2004) in C20, allowing internal Op Sets
HOOD022 (11.Feb.2005) in C15, adding the case of non constrained ops !
------------------------------------------------*/
/**** consistency rules ****/
/*-----------------------------------------------*/
 
/*``Consistency``Wed Mar 17 14:48:41 MET 1999``*/
/* HRM 4 rule C1: 
Information defined in HOOD diagrams shall be consistent with 
their textual representation in the relevant ODS fields:
- one ODS for each MODULE
- provided TYPEs, OPERATIONs and EXCEPTIONs
- required MODULE for each uncle MODULE
- internal MODULE for each child MODULE
- DATAFLOWS for each dataflow drawn
- EXCEPTION_FLOWS for each exception flow drawn
- CONSTRAINED OPERATIONS for each operation with a trigger arrow
- protocol constrained operation for each operation with a labelled trigger arrow
- OBCS reduced to OSTD for PASSIVE MODULEs
- full OBCS for ACTIVE MODULEs
- OPCS for each OPERATION of a terminal MODULE.*/

/* Implementation note:
Consistency between STOOD graphical editor and ods textual editor, 
is managed automatically, except for REQUIRED INTERFACE which is deduced 
from entered code or pseudo-code . */

checkC :- printlf('---> rule C1 is implicit.').
 
/*``Consistency``Thu Jul 13 18:26:40 MET DST 2000``*/
/* HRM 4 rule C2: 
If MODULE A requires an OPERATION of MODULE B, 
then MODULE A shall OP_USE B.*/

/* implementation note:
This rule is opposite to U1.
last modified: 13 Jul 2000 */

ruleC2(X,Y) :- uses(X,Y,'OPERATION',_).
ruleC2(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'ada'), memezTra('STANDARD',Y).
ruleC2(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'c'), Y='std'.
ruleC2(X,Y) :- hoodObject(Y,_,'SYSTEM_CONFIGURATION').

checkC :- requires(U,E,X,V,'OPERATION',Y,W), X \= Y, W \= 'OpImpBy', 
	implem(U,E,X,_,Z), Z \= Y, not(ruleC2(Z,Y)), 
	( W='InstPars' ->
		( write('ERROR : actual parameters') );
		( write('ERROR : '), write(E), write(' : '), write(U) ) 
	), 
	write(' of Module : '), write(X), 
	write(' requires OPERATION : '), write(V), 
	write(' of Module : '), write(Y), 
	write(' but there is no relevant Op_Use relationship ! (C2)'), nl.
checkC :- printlf('---> rule C2 checked...').
 
/*``Consistency``Wed Oct 23 16:26:05 MET 2002``*/
/* HRM 4 rule C3: 
If MODULE A requires a TYPE of MODULE B, 
then MODULE A shall TYPE_USE B.*/

/* implementation note:
Considering also rule C4, we can suppose that C3 
regards non terminal Modules only.
last modified: 23.10.02 */

ruleC3(X,Y) :- uses(X,Y,'TYPE',1).
ruleC3(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'ada'), memezTra('STANDARD',Y).
ruleC3(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'c'), Y='std'.
ruleC3(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'cpp'), Y='std'.
ruleC3(X,Y) :- hoodObject(Y,_,'SYSTEM_CONFIGURATION').

checkC :- requires(U,E,X,V,'TYPE',Y,W), X \= Y, W \= 'TyImpBy', 
	isParent(Y), implem(U,E,X,_,Z), Z \= Y, not(ruleC3(Z,Y)), 
	( W='InstPars' -> /*(cf.HOOD008)*/
		( write('ERROR : actual parameters') ); 
		( write('ERROR : '), write(E), write(' : '), write(U) ) 
	), 
	write(' of Module : '), write(X), 
	write(' requires TYPE : '), write(V), 
	write(' of Module : '), write(Y), 
	write(' but there is no relevant Type_Use relationship ! (C3)'), nl.
checkC :- printlf('---> rule C3 checked...').
 
/*``Consistency``Thu Jul 13 18:34:52 MET DST 2000``*/
/* HRM 4 rule C4: 
If MODULE A requires a TYPE of CHILD B, 
then MODULE A shall TYPE_USE or INHERIT from 
or ATTRIBUTE B.*/

/* implementation note:
Considering also rule C3, we can suppose that C4 
regards terminal Modules only.
last modified: 13 Jul 2000 */

ruleC4(X,Y) :- uses(X,Y,'TYPE',_).
ruleC4(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'ada'), memezTra('STANDARD',Y).
ruleC4(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'c'), Y='std'.
ruleC4(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'cpp'), Y='std'.
ruleC4(X,Y) :- hoodObject(Y,_,'SYSTEM_CONFIGURATION').

checkC :- requires(U,E,X,V,'TYPE',Y,W), X \= Y, W \= 'TyImpBy', 
	isTerm(Y), implem(U,E,X,_,Z), Z \= Y, not(ruleC4(Z,Y)), 
	( W='InstPars' ->
		( write('ERROR : actual parameters') );
		( write('ERROR : '), write(E), write(' : '), write(U) ) 
	), 
	write(' of Module : '), write(X), 
	write(' requires TYPE : '), write(V), 
	write(' of Module : '), write(Y), 
	write(' but there is no relevant Type_Use relationship ! (C4)'), nl.
checkC :- printlf('---> rule C4 checked...').
 
/*``Consistency``Mon Jun 25 08:58:32 MET DST 2001``*/
/* HRM 4 rule C5: 
Information defined in the ATTRIBUTES field of the ODS of a CLASS 
shall be consistent with ATTRIBUTE relationships in HOOD diagrams.*/

/* implementation note:
Same rule may be applied for any structured type
(not only Classes).
last modified: 17 May 2001*/

errC5_1('').
ruleC5_3(X,Y) :- uses(X,Z,'TYPE',3), same(Y,Z). /*(cf.HOOD005)*/
ruleC5_3(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'ada'), memezTra('STANDARD',Y).
ruleC5_3(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'c'), Y='std'.
ruleC5_3(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'cpp'), Y='std'.
ruleC5_3(X,Y) :- hoodObject(Z,_,'SYSTEM_CONFIGURATION'), same(Y,Z).

checkC :- attributes(U,X,A,V,_,_), errC5_1(V), 
	write('WARNING : TYPE : '), write(U), 
	write(' of Module : '), write(X), 
	write(' has an Attribute : '), write(A), 
	write(' which TYPE is not specified ! (C5_1)'), nl.
checkC :- attributes(U,X,A,V,Y,_), isNotEnv(X), /*(cf.HOOD007)*/
	V\='', implem(U,'TYPE',X,_,Z), 
	( Y \= 'NIL' -> C=Y; C=X ), implem(V,'TYPE',C,_,B), 
	diff(Z,B), not(ruleC5_3(Z,B)), 
	write('ERROR : TYPE : '), write(U), 
	write(' of Module : '), write(X), 
	write(' has an Attribute : '), write(A), 
	write(' of  TYPE : '), write(V), 
	write(' of Module : '), write(C), 
	write(' but there is no relevant Attributes relationship ! (C5_3)'), nl.
checkC :- printlf('---> rule C5 checked...').
 
/*``Consistency``Mon Jun 25 08:58:21 MET DST 2001``*/
/* HRM 4 rule C6: 
Information defined in the INHERITANCE field of the ODS of a CLASS 
shall be consistent with INHERITANCE relationships in HOOD diagrams.*/

/* implementation note:
List of superclasses in Inheritance field of a Class is automatically 
filled in when an Inheritance relationship is drawn, but may be 
changed inconsistently later.
last modified: 17 May 2001*/

ruleC6(X,Y) :- uses(X,Z,'TYPE',2), same(Y,Z). /*(cf.HOOD001)*/
ruleC6(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'ada'), memezTra('STANDARD',Y).
ruleC6(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'c'), Y='std'.
ruleC6(X,Y) :- isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,'cpp'), Y='std'.
ruleC6(X,Y) :- hoodObject(Z,_,'SYSTEM_CONFIGURATION'), same(Y,Z).

checkC :- inherits(U,X,V,Y), isNotEnv(X), /*(cf.HOOD007)*/
	V\='', implem(U,'TYPE',X,_,Z), 
	( Y \= 'NIL' ->  C=Y; C=X ), implem(V,'TYPE',C,_,B), 
	diff(Z,B), not(ruleC6(Z,B)), 
	write('ERROR : CLASS : '), write(U), 
	write(' of Module : '), write(X), 
	write(' inherits from CLASS : '), write(V), 
	write(' of Module : '), write(C), 
	write(' but there is no relevant Inheritance relationship ! (C6)'), nl.
checkC :- printlf('---> rule C6 checked...').
 
/*``Consistency``Tue Mar  2 09:26:25 MET 1999``*/
/* HRM 4 rule C7: 
Information defined in the USED_OPERATION field of the ODS of a MODULE 
shall be consistent with OP_USE relationships in HOOD diagrams.*/

/* implementation note:
In Stood, Used_Operations field is automatically updated from source 
code (or pseudo_code) dependencies, like other Required_Interface fields.
So, C7 and C2 are synonyms for Stood.*/

checkC :- printlf('---> rule C7 is implicit.').
 
/*``Consistency``Tue Mar  2 16:48:24 MET 1999``*/
/* HRM 4 rule C8: 
Information defined in the PROPAGATED_EXCEPTION field of the ODS of a MODULE 
shall be consistent with PROVIDED_INTERFACE of server MODULEs.*/

/* implementation note:
In Stood, Propagated_Exceptions field is automatically updated from 
Exception declarations, and exception "raised_by" field consistency is 
checked at input in graphical or ods text editors (procedure 33).*/

checkC :- printlf('---> rule C8 not implemented.').
 
/* 
errC8(V,Y) :- not(isProvided(V,'EXCEPTION',Y)).

checkC :- raisedException(U,'OPERATION,X,V,Y), errC8(V,Y), 
	write('ERROR : OPERATION : '), write(U), 
	write(' of Module : '), write(X), 
	write(' raises or propagates EXCEPTION : '), write(V),
	write(' which should be but is not provided by Module : '), write(Y), 
	write(' ! (C8)'), nl.
checkC :- printlf('---> rule C8 checked...'). 
*/
 
/*``Consistency``Tue Mar  2 11:22:37 MET 1999``*/
/* HRM 4 rule C9: 
Information defined in the DATAFLOW field of the ODS shall be consistent with 
relevant OP_USE and TYPE_USE relationships in HOOD diagrams.*/

/* implementation note:
In Stood, DataFlows field of the ODS is automatically deduced from 
information entered in graphical descriptions, but from OP_USE 
relationships only (procedure 4).*/

checkC :- printlf('---> rule C9 is implicit.').
 
/*``Consistency``Tue Mar  2 16:40:56 MET 1999``*/
/* HRM 4 rule C10: 
Information defined in the EXCEPTION_FLOW field of the ODS shall be consistent with 
relevant OP_USE relationships in HOOD diagrams.*/

/* implementation note:
In Stood, Exception_Flows field of the ODS is automatically deduced from 
information entered in graphical descriptions (procedure 5).*/

checkC :- printlf('---> rule C10 is implicit.').
 
/*``Consistency``Fri Mar 12 12:36:51 MET 1999``*/
/* HRM 4 rule C11: 
Only EXCEPTIONs listed in the PROVIDED_INTERFACE EXCEPTIONS field 
of a MODULE shall be propagated by OPERATIONs provided by the MODULE.*/

errC11(X,E) :- not(isProvided(E,'EXCEPTION',X)).

checkC :- raisedException(O,'OPERATION',X,E), errC11(X,E), 
	write('ERROR : OPERATION : '), write(O), 
	write(' of Module : '), write(X), 
	write(' raises an EXCEPTION : '), write(E), 
	write(' not provided by the same Module ! (C11)'), nl.
checkC :- printlf('---> rule C11 checked...').
 
/*``Consistency``Mon Mar  8 17:13:21 MET 1999``*/
/* HRM 4 rule C12: 
Parameters of an INSTANCE_OF shall be consistent with 
formal parameters of relevant GENERIC.*/

errC12(U,K,I) :- not( isInstance(I,_,G), formalParameters(U,K,G) ).

checkC :- actualParameter(U,K,I,_), errC12(U,K,I), 
	write('ERROR : '), write(K), write(' : '), write(U), 
	write(' of Instance_Of : '), write(I), 
	write(' is not consistent with formal parameters ! (C12)'), nl.
checkC :- printlf('---> rule C12 checked...').
 
/*``Consistency``Thu Dec  7 17:21:15 MET 2000``*/
/* HRM 4 rule C13: 
CHILD OPERATION declaration shall match 
PARENT OPERATION declaration.*/

/* implementation note:
In Stood, consistency along implemented_by link
is managed automatically, except for parameters types.
last modified: 07 Dec 2000*/

errC13_1(U,X,V,Y) :- arg(U,X,P,_,_,_), not(arg(V,Y,P,_,_,_)), !.

errC13_2(X,V,Y,P,A,B,Q) :- arg(V,Y,P,C,D,R), 
	( ( diff(A,X), A\='NIL' ) -> 	/* if type of parent parameter is not provided by same module as the op... */
		( diff(C,A) ; diff(D,B) ; R\=Q ); 	/* then type of child parameter should be the same as parent one... */
		( isImplementedBy(B,'TYPE',X,E,F) ->  /* else this type should be implemented by... */
			( ( C\='NIL', diff(C,F) ) ; diff(D,E) ; R\=Q );	/* if yes, then type of child parameter should be the implemented_by one... */
			fail ) ), !.	/* if no, it may be simply use of a standard type */

checkC :- isImplementedBy(U,'OPERATION',X,V,Y), errC13_1(U,X,V,Y), 
	write('ERROR : parameter names of OPERATION : '), write(U), 
	write(' of parent Module : '), write(X), 
	write(' are not consistent with parameters of OPERATION : '), write(V), 
	write(' of child Module : '), write(Y), write(' ! (C13_1)'), nl.
checkC :- isImplementedBy(U,'OPERATION',X,V,Y), 
	arg(U,X,P,A,B,Q), errC13_2(X,V,Y,P,A,B,Q), 
	write('ERROR : '), write(P), 
	write(' parameter of OPERATION : '), write(U),
	write(' of parent Module : '), write(X), 
	write(' has not a consistent TYPE with the same parameter of OPERATION : '), write(V), 
	write(' of child Module : '), write(Y),
	write(' ! (C13_2)'), nl.
checkC :- printlf('---> rule C13 checked...').
 
/*``Consistency``Mon Mar  8 18:32:59 MET 1999``*/
/* HRM 4 rule C14: 
CHILD OPERATION_SET contents shall match 
PARENT OPERATION_SET contents.*/

errC14(U,X,V,Y) :- isMemberOf(A,B,X,U),
	( isImplementedBy(A,B,X,C,Y) -> 
	not(isMemberOf(C,B,Y,V)); true ), !.

checkC :- isImplementedBy(U,'OPERATIONSET',X,V,Y), errC14(U,X,V,Y), 
	write('ERROR : contents of OPERATION_SET : '), write(U), 
	write(' of parent Module : '), write(X), 
	write(' is not consistent with contents of OPERATION_SET : '), write(V), 
	write(' of child Module : '), write(Y), write(' ! (C14)'), nl.
checkC :- printlf('---> rule C14 checked...').
 
/*``Consistency``Tue Mar  9 09:24:34 MET 1999``*/
/* HRM 4 rule C15: 
CHILD OPERATION constraints shall match 
PARENT OPERATION constraints.
last modified: 11 Feb 2005*/

ruleC15(U,X,V,Y) :- not(isConstrained(U,'OPERATION',X,_,_)), not(isConstrained(V,'OPERATION',Y,_,_)).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,A,B), isConstrained(V,'OPERATION',Y,A,B).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,'ASER',B), isConstrained(V,'OPERATION',Y,'ASATC',B).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,'ASER',B), isConstrained(V,'OPERATION',Y,'PSER',B).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,'ASER',B), isConstrained(V,'OPERATION',Y,'PAER',B).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,'LSER',B), isConstrained(V,'OPERATION',Y,'PSER',B).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,'LSER',B), isConstrained(V,'OPERATION',Y,'LSATC',B).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,'HSER',B), isConstrained(V,'OPERATION',Y,'PSER',B).
ruleC15(U,X,V,Y) :- isConstrained(U,'OPERATION',X,'HSER',B), isConstrained(V,'OPERATION',Y,'HSATC',B).

checkC :- isImplementedBy(U,'OPERATION',X,V,Y), not(ruleC15(U,X,V,Y)), 
	write('ERROR : constraints of OPERATION : '), write(U), 
	write(' of parent Module : '), write(X), 
	write(' are not consistent with constraints of OPERATION : '), write(V), 
	write(' of child Module : '), write(Y), write(' ! (C15)'), nl.
checkC :- printlf('---> rule C15 checked...').
 
/*``Consistency``Tue Mar  9 09:30:49 MET 1999``*/
/* HRM 4 rule C16: 
An operation listed in the USED_OPERATION field of an OPCS 
shall be either a required operation, an internal or a provided one.*/

/* implementation note:
In Stood, Used_Operations field of the ODS is automatically deduced from 
entered source code or pseudo-code via the cross-references (procedure 23).*/

checkC :- printlf('---> rule C16 is implicit.').
 
/*``Consistency``Tue Mar  9 09:35:04 MET 1999``*/
/* HRM 4 rule C17: 
The PROPAGATED_EXCEPTION list of an OPCS shall be a subset 
of the provided exception list in the PROVIDED_INTERFACE.*/

/* implementation note:
Same rule as C11.*/

checkC :- printlf('---> rule C17 is the same as C11.').
 
/*``Consistency``Thu May 17 15:40:54 MET DST 2001``*/
/* HRM 4 rule C18: 
Each provided and internal OPERATION of a terminal MODULE 
shall have an OPCS.
last changed: 17 May 2001*/

errC18(U,X) :- not( file(U,'OPERATION',X,_,'OpDef') ). /*(cf.HOOD003)*/

checkC :- isTerm(X), isNotInst(X), 
	not(isObject(X,_,'SYSTEM_CONFIGURATION')), 
	( isProvided(U,'OPERATION',X) ; isInternal(U,'OPERATION',X) ), 
	not(isInherited(U,X)), not(isAbstract(U,'OPERATION',X)), 
	errC18(U,X), write('WARNING : OPERATION : '), write(U), 
	write(' of terminal Module : '), write(X), 
	write(' should have an OPCS ! (C18)'), nl.
checkC :- printlf('---> rule C18 checked...').
 
/*``Consistency``Wed Oct 23 16:25:43 MET 2002``*/
/* HRM 4 rule C19: 
Each TERMINAL MODULE shall have one OBCS.*/

/* implementation note:
With Stood, only terminal Modules providing constrained 
operations, should have an OBCS.
If a STD has been defined, OBCS can be generated automatically.
last modified: 23.10.02 */

errC19(X) :- isConstrained(U,'OPERATION',X,_,_), 
	not( isState(X,_,_) ), not( file('NIL','NIL',X,_,'obcs') ), !. /*(cf.HOOD009)*/

checkC :- isTerm(X), not(isObject(X,_,'SYSTEM_CONFIGURATION')), 
	errC19(X), 
	write('WARNING : Terminal Module : '), write(X), 
	write(' should have an OBCS ! (C19)'), nl.
checkC :- printlf('---> rule C19 checked...').
 
/*``Consistency``Wed Mar 17 16:23:36 MET 1999``*/
/* HRM 4 rule C20: 
Each OPERATION, MEMBER_OF and the relevant OPERATION_SET 
shall be declared in the same MODULE.
last modified: 23 Dec 2004*/

errC20(X,S) :- not(isProvided(S,'OPERATIONSET',X)), not(isInternal(S,'OPERATIONSET',X)).

checkC :- isMemberOf(O,'OPERATION',X,S), errC20(X,S), 
	write('ERROR : OPERATION_SET: '), write(S), 
	write(' should be in the same Module : '), write(X), 
	write(' as its member_of OPERATION : '), write(O), 
	write(' ! (C20)'), nl.
checkC :- printlf('---> rule C20 checked...').
 
/*``Consistency``Tue Mar  9 10:21:43 MET 1999``*/
/* HRM 4 rule C21: 
A PARENT has no internal operations.*/

errC21(X) :- isInternal(_,'OPERATION',X), !.

checkC :- isNotTerm(X), errC21(X), 
	write('ERROR : Parent Module: '), write(X), 
	write(' should have no internal OPERATION ! (C21)'), nl.
checkC :- printlf('---> rule C21 checked...').
 
/*``Consistency``Tue Mar  9 10:26:20 MET 1999``*/
/* HRM 4 rule C22: 
A PARENT has no OPCS.*/

errC22(X) :- file(_,'OPERATION',X,_,'OpDef'), !.

checkC :- isNotTerm(X), errC22(X), 
	write('ERROR : Parent Module: '), write(X), 
	write(' should have no OPCS ! (C22)'), nl.
checkC :- printlf('---> rule C22 checked...').
 
/*``Consistency``Fri Mar 12 12:39:47 MET 1999``*/
/* HRM 4 rule C23: 
A PARENT has no DATA.*/

/* implementation note:
This rule has been extended to internal Constants and Types.*/

errC23(X) :- isInternal(_,K,X), 
	( K='DATA' ; K='CONSTANT' ; K='TYPE' ), !.

checkC :- isNotTerm(X), errC23(X), 
	write('ERROR : Parent Module: '), write(X), 
	write(' should have no DATA - and no internal CONSTANT nor TYPE - ! (C23)'), nl.
checkC :- printlf('---> rule C23 checked...').
 
/*``Consistency``Tue Mar  9 10:39:05 MET 1999``*/
/* HRM 4 rule C24: 
An OP_CONTROL has no PROVIDED_INTERFACE.*/

/* implementation note:
In Stood, an Op_Set has only one provided Operation.*/

errC24(X) :- isProvided(O,'OPERATION',X), 
	isProvided(U,K,X), ( K='OPERATION' -> U\=O; true ), !.

checkC :- hoodObject(X,'OP_CONTROL',_), errC24(X), 
	write('ERROR : Op_Control Module: '), write(X), 
	write(' should have no provided interface ! (C24)'), nl.
checkC :- printlf('---> rule C24 checked...').
 
/*``Consistency``Fri Mar 12 12:40:11 MET 1999``*/
/* HRM 4 rule C25: 
An OP_CONTROL has no DATA.*/

/* implementation note:
This rule has been extended to all internals.*/

errC25(X) :- isInternal(_,_,X), !.

checkC :- hoodObject(X,'OP_CONTROL',_), errC25(X), 
	write('ERROR : Op_Control Module: '), write(X), 
	write(' should have no DATA - nor other internal elements - ! (C25)'), nl.
checkC :- printlf('---> rule C25 checked...').
 
/*``Consistency``Thu Dec 19 15:52:51 MET 2002``*/
/* HRM 4 rule C26: 
The OSTD of a TERMINAL (not VN) MODULE shall describe 
the possible transitions associated only to provided OPERATIONs. */

/* implementation note:
last modified: 19.12.02 */

errC26(X,O) :- not( isProvided(O,'OPERATION',X) ).

checkC :- isTransition(X,T,_,_,O), isTerm(X), errC26(X,O), /*(cf.HOOD010)*/
	write('WARNING : TRANSITION : '), write(T), 
	write(' in STD of Module : '), write(X), 
	write(' is not associated to a provided Operation of this Module ! (C26)'), nl.
checkC :- printlf('---> rule C26 checked...').
 
/*``Consistency``Thu Dec 19 14:42:37 MET 2002``*/
/* HRM 4 rule C27: 
The OBCS of a TERMINAL PASSIVE (not VN) MODULE shall be reduced to 
the OSTD and OSTM fields, with each constrained operation OPCS comprising 
an associated OPCS_HEADER and OPCS_FOOTER field.*/

/* implementation note:
In Stood, implementation of the STD is slightly different from that.
Related ODS fields are:
- For each STATE: assignment and test fields.
- For each TRANSITION: condition and exception fields.
First ones are mandatory to generate code.
last modified: 19.12.02 */

errC27(X,S) :- 
	not( file(S,'STATE',X,_,'SetState') );
	not( file(S,'STATE',X,_,'GetState') ), !.

checkC :- isState(X,S,_), isTerm(X), errC27(X,S), /*(cf.HOOD010)*/
	write('WARNING : missing assignment and test code for STATE : '), write(S), 
	write(' in STD of terminal Module : '), write(X), 
	write('  ! (C27)'), nl.
checkC :- printlf('---> rule C27 checked...').
 
/*``Consistency``Wed Jun 21 18:03:42 MET DST 2000``*/
/* HRM 4 rule C28: 
The reference to an entity in the REQUIRED_INTERFACE of 
a client MODULE shall be consistent with the declaration of 
that entity in the PROVIDED_INTERFACE of the server MODULE.*/

/* last modified: 21 Jun 2000 */

errC28(V,F,Y) :- not( isProvided(V,F,Y) ).

checkC :- requires(U,E,X,V,F,Y,W), 
	W \= 'TyImpBy', W \= 'CoImpBy', W \= 'OpImpBy', W \= 'ExImpBy', 
	X \= Y, Y \= '?', errC28(V,F,Y), 
	write('ERROR : reference to '), write(F), write(' : '), write(V), 
	write(' of Module : '), write(Y), 
	write(' in '), write(E), write(' : '), write(U), 
	write(' of Module : '), write(X),  
	write('  is not consistent with its declaration ! (C28)'), nl.
checkC :- printlf('---> rule C28 checked...').
 
