/*``changes``Wed Oct 23 16:34:11 MET 2002``*/
/* version */
/*-------------------------------------------------
	STOOD v5.0
	HOOD 4 rules checker 
	General rules: HRM4 p91-92

	Copyright TNI Europe 2004
	Pierre DISSAUX (pierre.dissaux@tni-world.com)
------------------------------------------------*/
/**** changes ****/
/*------------------------------------------------
HOOD006 (25.Jun.2001) changing G6 for non case sensitive use of environments (cf.RASTO/01.045)
HOOD011 (23.Oct.2002) avoiding multiple message in G6 when same environment is used many times
HOOD012 (23.Oct.2002) fixing a bug in G13 and extending to HRT-HOOD constraints
HOOD019 (23.Dec.2004) removing obsolete G1_2 and G1_3 rules. G1_1 renamed into G1
------------------------------------------------*/
/**** general rules ****/
/*-----------------------------------------------*/
 
/*``General``Tue Jan 11 17:36:01 MET 2000``*/
/* HRM 4 rule G1:
A MODULE is of one and only one of the following types:
	OBJECT, ACTIVE or PASSIVE;
	CLASS, ACTIVE or PASSIVE;
	OP_CONTROL;
	INSTANCE_OF GENERIC;
	(object) ENVIRONMENT, ACTIVE or PASSIVE;
	CLASS ENVIRONMENT, ACTIVE or PASSIVE;
	GENERIC ENVIRONMENT, ACTIVE or PASSIVE;
	(generic class environment active or passive);
	VIRTUAL_NODE. */

/* Implementation notes: 
1- STOOD v4 doesn't reference explicitly a "GENERIC", 
but identifies it by the presence of a "FORMAL_PARAMETERS" 
module. 
2- STOOD v4 doesn't reference explicitly a "CLASS", 
but identifies it by the presence of a class TYPE. 
last modified: 23 Dec 2004 */

errG1_1(X,T) :- 
	T \= 'ENVIRONMENT',
	T \= 'ACTIVE', T \= 'PASSIVE', 
	T \= 'OP_CONTROL', T \= 'INSTANCE_OF',
	T \= 'VIRTUAL_NODE', 
	T \= 'FORMAL_PARAMETERS',
	T \= 'PROTECTED', T \='CYCLIC', T\= 'SPORADIC'. 

checkG :- hoodObject(X,T,_), errG1_1(X,T),
	write('ERROR : Module : '), write(X), 
	write(' is declared with an illegal type : '), write(T),
	write(' ! (G1)'), nl.
checkG :- printlf('---> rule G1 checked...').
 
/*``General``Fri Jan 22 15:09:01 MET 1999``*/
/* HRM 4 rule G2:
A MODULE which does not have a parent is called a ROOT. */

checkG :- printlf('---> rule G2 is a definition.').
 
/*``General``Thu Oct 28 10:36:02 MET 1999``*/
/* HRM 4 rule G3:
a SYSTEM_CONFIGURATION is a set of ROOTs which defines the scope and visibility
of a given ROOT considered as the "SYSTEM_TO_DESIGN". */

/* implementation note:
last modified: 28 Oct 1999 */

ruleG3_1(X) :- isRootObject(Y,_,_), memezTra(X,Y).

errG3_2(X,P) :- hoodObject(X,_,P), P \= 'SYSTEM_CONFIGURATION'.

checkG :- isCurrentRoot(X), not(ruleG3_1(X)), 
	write('ERROR : current Root Module : '), write(X),
	write(' is not declared as a member of System_Configuration ! (G3_1) '), nl.
checkG :- isCurrentRoot(X), errG3_2(X,P), 
	write('WARNING : current Root Module : '), write(X),
	write(' has for parent : '), write(P),
	write(' ! (G3_2) '), nl.
checkG :- printlf('---> rule G3 checked...').
 
/*``General``Wed Mar 17 15:29:50 MET 1999``*/
/* HRM 4 rule G4:
A ROOT may be:
	The current "SYSTEM_TO_DESIGN";
	An ENVIRONMENT, application or library;
	A GENERIC (OBJECT or CLASS);
	A VIRTUAL_NODE. */

/* Implementation notes:
1- In STOOD v4, current "SYSTEM_TO_DESIGN" and "ENVIRONMENT" 
modules are referenced by "DESIGN" keyword.
2- HOOD4 "GENERICs" replace HOOD3.x "CLASSes". */

errG4_1(X,T) :- 
	T \= 'DESIGN',
	T \= 'GENERIC',
	T \= 'VIRTUAL_NODE'.

errG4_2(X) :- isRootObject(X,'CLASS',_).

checkG :- isRootObject(X,T,_), errG4_1(X,T),
	write('ERROR : Root Module : '), write(X),
	write(' is declared with an illegal type : '), write(T),
	write(' ! (G4_1)'), nl. 
checkG :- errG4_2(X),
	write('WARNING : for Root Module : '), write(X),
	write(' : HOOD3.x classes become HOOD4 generics ! (G4_2) '), nl.
checkG :- printlf('---> rule G4 checked...').
 
/*``General``Fri Jan 22 15:08:52 MET 1999``*/
/* HRM 4 rule G5:
A TERMINAL MODULE is a MODULE which has no CHILD MODULES. */

checkG :- printlf('---> rule G5 is a definition.').
 
/*``General``Wed Oct 23 17:09:43 MET 2002``*/
/* HRM 4 rule G6:
A MODULE may require one or more ENVIRONMENTs, each of which defines 
the interfaces of a ROOT declared in the SYSTEM_CONFIGURATION. 

/* implementation note:
last changed: 23.10.02 */

usedEnv(X) :- uses(_,X,_,_), not(hoodObject(X,_,_)).

ruleG6_1(E) :- isRootObject(F,'DESIGN',_), same(E,F). /*(cf.HOOD006)*/

errG6_2(E) :- hoodObject(E,'ENVIRONMENT',P), P \= 'SYSTEM_CONFIGURATION'.

checkG :-  
	( hoodObject(X,'ENVIRONMENT',_); usedEnv(X) ), /*(cf.HOOD011)*/
	not(ruleG6_1(X)), 
	write('WARNING : Environment Module : '), write(X),
	write(' is not a root module of current system  ! (G6_1) '), nl.
checkG :- errG6_2(X),
	write('ERROR : Environment Module : '), write(X),
	write(' is not declared in the System_Configuration ! (G6_2) '), nl.
checkG :- printlf('---> rule G6 checked...').
 
/*``General``Fri Mar 12 14:46:02 MET 1999``*/
/* HRM 4 rule G7:
An OP_CONTROL is a TERMINAL MODULE without internal DATA, 
dedicated to the implementation of one and only one OPERATION. */

errG7_1(X) :- hoodObject(Y,_,X), !.

errG7_2(X) :- isInternal(_,'DATA',X), !.

errG7_3(X) :- isProvided(O,'OPERATION',X), isProvided(U,E,X),
	( U \= O; E \= 'OPERATION' ), !.

checkG :- hoodObject(X,'OP_CONTROL',_), errG7_1(X), 
	write('ERROR : Op_Control : '), write(X), 
	write(' should be a Terminal Module ! (G7_1)'), nl.
checkG :- hoodObject(X,'OP_CONTROL',_), errG7_2(X), 
	write('ERROR : Op_Control : '), write(X), 
	write(' should not contain Data ! (G7_2)'), nl.
checkG :- hoodObject(X,'OP_CONTROL',_), errG7_3(X), 
	write('ERROR : Op_Control : '), write(X), 
	write(' should not provide more than a single Operation ! (G7_3)'), nl.
checkG :- printlf('---> rule G7 checked...').
 
/*``General``Thu Dec 19 16:10:41 MET 2002``*/
/* HRM 4 rule G8:
A CLASS is a TERMINAL MODULE which provides one and only one main 
(inheritable) TYPE, and OPERATIONS whose receiver is of this TYPE. */

/* Implementation note:
last modified: 19 Dec 2002*/

errG8_1(X) :- isClass(T,X), isClass(U,X), U \= T, !.

errG8_2(X) :- isClass(C,X), 
	not(argument(_,'OPERATION',X,_,'me','NIL',T,_,_), memezTra(C,T)),
	not(return(_,'OPERATION',X,'NIL',T,_), memezTra(C,T)).

checkG :- hoodObject(X,_,_), isTerm(X), not(isRootObject(X,_,_)), errG8_1(X), 
	write('ERROR : Class Module : '), write(X), 
	write(' should provide only one main Type  ! (G8_1) '), nl.
checkG :- hoodObject(X,_,_), isTerm(X), errG8_2(X), 
	write('WARNING : Class Module : '), write(X), 
	write(' provides no Operation whose receiver is main Type  ! (G8_2) '), nl.
checkG :- printlf('---> rule G8 checked...').
 
/*``General``Fri Jan 22 16:36:04 MET 1999``*/
/* HRM 4 rule G9:
A state constrained operation is a CONSTRAINED OPERATION with a 
CONSTRAINED_BY STATE clause.*/

checkG :- printlf('---> rule G9 is a definition.').
 
/*``General``Fri Jan 22 16:35:48 MET 1999``*/
/* HRM 4 rule G10:
A concurrency constrained operation is a CONSTRAINED OPERATION with a 
CONSTRAINED_BY clause containing at least one of the constraint attributes: 
MTEX, RWER or ROER.*/

checkG :- printlf('---> rule G10 is a definition.').
 
/*``General``Fri Jan 22 16:35:51 MET 1999``*/
/* HRM 4 rule G11:
A protocol constrained operation is a CONSTRAINED OPERATION with a 
CONSTRAINED_BY clause containing at least one of the constraint attributes: 
ASER, HSER, LSER, RASER or RLSER.*/

checkG :- printlf('---> rule G11 is a definition.').
 
/*``General``Fri Jan 22 16:36:01 MET 1999``*/
/* HRM 4 rule G12:
A time out constrained operation is a CONSTRAINED OPERATION with a 
CONSTRAINED_BY clause containing at least the constraint attribute TO.*/

checkG :- printlf('---> rule G12 is a definition.').
 
/*``General``Wed Oct 23 16:32:40 MET 2002``*/
/* HRM 4 rule G13: 
TO constraint attribute shall only be combined with protocol constraints. */

/* implementation note:
last changed: 23.10.02 */

ruleG13(U,X) :- isConstrained(U,'OPERATION',X,C,_),
	( C = 'ASER'; C = 'HSER'; C = 'LSER'; C = 'RASER'; C = 'RLSER';
	  C = 'HSATC'; C = 'LSATC' ).

checkG :- isConstrained(U,'OPERATION',X,'TO',_), not(ruleG13(U,X)), /*(cf.HOOD012)*/
	write('ERROR : TO constrained Operation : '), write(U), write(' of Module : '), write(X),
	write(' should not have another constraint except a protocol constraint  ! (G13) '), nl.
checkG :- printlf('---> rule G13 checked....').
 
/*``General``Thu Feb 11 14:44:32 MET 1999``*/
/* HRM 4 rule G14: 
Protocol constraint attributes may be combined with state and/or 
concurrency constraints.*/

/* Implementation note:
This is implicitly allowed by STOODv4 Syntax. */

checkG :- printlf('---> rule G14 is implicit.').
 
/*``General``Thu Feb 11 14:45:07 MET 1999``*/
/* HRM 4 rule G15: 
Concurrency constraint attributes may be combined with 
state constraints.*/

/* Implementation note:
This is implicitly allowed by STOODv4 Syntax. */

checkG :- printlf('---> rule G15 is implicit.').
 
/*``General``Thu Feb 11 14:46:33 MET 1999``*/
/* HRM 4 rule G16: 
A PASSIVE MODULE is a MODULE that provides non constrained operations 
and/or state constraints. */

/* Implementation notes:
1- It also includes the case of concurrency only constraints.
2- I prefer following definitions: 
A PASSIVE MODULE is a MODULE that doesn't include 
its own execution threads.
-or:
A PASSIVE MODULE is a MODULE that doesn't provide 
any protocol constrained operation.*/

errG16(X) :- isConstrained(_,'OPERATION',X,C,_), 
	( C='ASER'; C='HSER', C='LSER', C='RASER', C='RLSER' ), !.

checkG :- hoodObject(X,'PASSIVE',_), errG16(X),
	write('WARNING : Passive Module : '), write(X), 
	write(' should not provide protocole constrained Operations ! (G16) '), nl.
checkG :- printlf('---> rule G16 checked....').
 
/*``General``Thu Oct 28 10:40:20 MET 1999``*/
/* HRM 4 rule G17: 
An ACTIVE MODULE is a MODULE that is not PASSIVE. */

/* Implementation note:
I prefer following definitions: 
An ACTIVE MODULE is a MODULE that includes 
its own execution threads.
-or:
An ACTIVE MODULE is a MODULE that provides 
at least one protocol constrained operation.
In addition "ASER_BY_IT" should be taken into accounts
for HOOD3.1 compatibility.
last modified: 28 Oct 1999 */

errG17(X) :- 
	not(isConstrained(_,'OPERATION',X,'ASER',_)), 
	not(isConstrained(_,'OPERATION',X,'ASER_BY_IT',_)),
	not(isConstrained(_,'OPERATION',X,'HSER',_)),
	not(isConstrained(_,'OPERATION',X,'LSER',_)),
	not(isConstrained(_,'OPERATION',X,'RASER',_)),
	not(isConstrained(_,'OPERATION',X,'RLSER',_)).

checkG :- hoodObject(X,'ACTIVE',_), errG17(X),
	write('WARNING : Active Module : '), write(X), 
	write(' should provide at least one protocol constrained operation ! (G17) '), nl.
checkG :- printlf('---> rule G17 checked...').
 
/*``General``Thu Jan 28 18:12:21 MET 1999``*/
/* HRM 4 rule G18: 
A TERMINAL MODULE built by instanciation of a GENERIC CLASS 
(resp. OBJECT) is a CLASS (resp. an OBJECT).*/

errG18_1(X,G) :- isClass(T,G), not(isClass(T,X)), !.

errG18_2(X,G) :- isClass(T,X), not(isClass(T,G)), !.

checkG :- isInstance(X,_,G), errG18_1(X,G),
	write('Module : '), write(X), 
	write(' Instance_Of Generic Class : '), write(G), 
	write(' should be a Class ! (G18_1) '), nl.
checkG :- isInstance(X,_,G), errG18_2(X,G),
	write('Module : '), write(X), 
	write(' Instance_Of Generic Object : '), write(G), 
	write(' should be an Object ! (G18_2) '), nl.
checkG :- printlf('---> rule G18 checked...').
 
/*``General``Fri Mar 12 12:29:26 MET 1999``*/
/* HRM 4 rule G19: 
A TERMINAL VIRTUAL_NODE is defined by allocation of (non VN) 
MODULEs of the same SYSTEM_CONFIGURATION.*/

errG19_1(X) :- not(hoodObject(_,_,X)), not(allocatedObject(X,_)).

errG19_2(X) :- not(hoodObject(_,_,X)), allocatedObject(X,M), 
	getRoot(M,R), not(isRootObject(R,'DESIGN',_)), !.

checkG :- hoodObject(X,'VIRTUAL_NODE',_), errG19_1(X), 
	write('WARNING : Terminal Virtual_Node : '), write(X), 
	write(' has no allocated module ! (G19_1) '), nl.
checkG :- hoodObject(X,'VIRTUAL_NODE',_), errG19_2(X), 
	write('WARNING : Terminal Virtual_Node : '), write(X), 
	write(' has an illegal allocated module ! (G19_2) '), nl.
checkG :- printlf('---> rule G19 checked...').
 
/*``General``Mon Jan 25 18:17:28 MET 1999``*/
/* HRM 4 rule G20:
A VIRTUAL_NODE shall provide at least one predefined OPERATION "Message_In", 
specifying the communication protocol with remote client VNs. */

checkG :- printlf('---> rule G20 not implemented').
 
/*``General``Thu Jan 28 18:12:51 MET 1999``*/
/* HRM 4 rule G21:
A PASSIVE (not VIRTUAL_NODE) MODULE shall be allocated to the same VN 
as all its clients. */

errG21(V,X) :- hoodObject(X,'PASSIVE',_), uses(Y,X,'OPERATION',_), 
	not(allocatedObject(V,Y)).

checkG :- allocatedObject(V,X), errG21(V,X), 
	write('WARNING : All clients of Passive Module : '), write(X), 
	write(' should be allocated on the same Virtual_Node ! (G21) '), nl.
checkG :- printlf('---> rule G21 checked...').
 