/*``changes``Wed Feb 12 17:22:17 MET 2003``*/
/* version */
/*-------------------------------------------------
	STOOD v4
	HOOD 4 rules checker 
	Operation rules: HRM4 p95

	Copyright TNI 1999-2003
	Pierre DISSAUX (pierre.dissaux@tni.fr)
	17.Mar.99
	29.Jul.99: O5
	28.Oct.99: O1
	11.Jul.00: O5
------------------------------------------------*/
/**** changes ****/
/*------------------------------------------------
HOOD004 (17.May.2001) restricting O6 to terminal modules
HOOD015 (23.Oct.2002) allowing differences for parameters types in O6
HOOD016 (12.Feb.2003) change in O6_3
HOOD024 (21.Mar.2005) O5 : removing false errors in environment class libs
------------------------------------------------*/
/**** operation rules ****/
/*-----------------------------------------------*/
 
/*``Operation``Thu Oct 28 10:38:29 MET 1999``*/
/* HRM 4 rule O1: 
If an OPERATION is protocol constrained, then it shall be provided by 
an ACTIVE MODULE.*/

/* Implementation note:
no errors generated HRT_HOOD modules.
last modified: 28.Oct.99 */

errO1(X,U,C) :- ( C='HSER'; C='LSER'; C='ASER'; C='ASER_BY_IT' ), 
	not(hoodObject(X,'ACTIVE',_)),
	not(hoodObject(X,'PROTECTED',_)),
	not(hoodObject(X,'CYCLIC',_)),
	not(hoodObject(X,'SPORADIC',_)).

checkO :- isConstrained(U,'OPERATION',X,C,_), errO1(X,U,C),
	write('ERROR : Module : '), write(X),
	write(' provides protocol constrained Operation : '), write(U),
	write(' and should be Active ! (O1)'), nl.
checkO :- printlf('---> rule O1 checked...').
 
/*``Operation``Thu Feb 25 11:09:06 MET 1999``*/
/* HRM 4 rule O2: 
If an OPERATION_SET is provided by a MODULE, then all its MEMBER_OF 
elements shall be provided by the same MODULE.*/

errO2(X,S,O) :- isMemberOf(O,'OPERATION',X,S),
	not(isProvided(O,'OPERATION',X)).

checkO :- isProvided(S,'OPERATIONSET',X), errO2(X,S,O),
	write('ERROR : Operation : '), write(O), 
	write(' member of Operation_Set : '), write(S), 
	write(' of Module : '), write(X), 
	write(' should be provided by the same Module ! (O2)'), nl.
checkO :- printlf('---> rule O2 checked...').
 
/*``Operation``Thu Feb 25 09:40:46 MET 1999``*/
/* HRM 4 rule O3: 
An element, MEMBER_OF an OPERATION_SET shall not be 
MEMBER_OF another OPERATION_SET.*/

errO3(X,S,O) :- isMemberOf(O,'OPERATION',X,T),
	T \= S.

checkO :- isMemberOf(O,'OPERATION',X,S), errO3(X,S,O),
	write('ERROR : Operation : '), write(O), 
	write(' should not be member of several Operation_Sets of Module : '), write(X), 
	write(' ! (O3)'), nl.
checkO :- printlf('---> rule O3 checked...').
 
/*``Operation``Thu Feb 25 11:09:39 MET 1999``*/
/* HRM 4 rule O4: 
An OPERATION_SET of a PARENT can only have OPERATIONs and/or 
OPERATION_SETs as MEMBER_OF elements.*/

errO4(X,S,O) :- isMemberOf(O,_,X,S),
	not(isProvided(O,'OPERATION',X)),
	not(isProvided(O,'OPERATIONSET',X)).

checkO :- isParent(X), isProvided(S,'OPERATIONSET',X), errO4(X,S,O),
	write('ERROR : Element : '), write(O), 
	write(' member of Operation_Set : '), write(S), 
	write(' of Module : '), write(X), 
	write(' should be an Operation or an Operation_Set ! (O4)'), nl.
checkO :- printlf('---> rule O4 checked...').
 
/*``Operation``Tue Jul 11 17:45:55 MET DST 2000``*/
/* HRM 4 rule O5: 
If an OPERATION is provided by or internal to a CLASS, 
then its declaration shall specify one parameter, called the receiver, 
of reserved name ME or MYCLASS, 
and being of the main TYPE of the CLASS.*/

/* Implementation note:
receiver may also be returned value. 
last modified: 21 Mar 2005*/

errO5_1(X,T,O) :- ( isProvided(O,'OPERATION',X); isInternal(O,'OPERATION',X) ),
	not(hasArgO5(O,X,'me')),
	not(hasArgO5(O,X,'myclass')),
	not(return(O,'OPERATION',X,'NIL',T,_)).

errO5_2(X,T,O) :- ( isProvided(O,'OPERATION',X); isInternal(O,'OPERATION',X) ),
	argument(O,'OPERATION',X,_,P,M,C,_,_), 
	( memezTra(P,'me'); memezTra(P,'myclass') ),
	( neoKetMemezTra(C,T) ; ( neoKetMemezTra(M,X), M\= 'NIL') ).

hasArgO5(O,X,V) :- argument(O,'OPERATION',X,_,P,_,_,_,_), memezTra(P,V).

checkO :- isProvided(T,'TYPE',X), isClass(T,X), isTerm(X), isNotEnv(X), errO5_1(X,T,O), 
	write('WARNING : Operation : '), write(O), /*(cf.HOOD024)*/
	write(' provided by or internal to Class : '), write(X), 
	write(' should specify a receiver ! (O5_1)'), nl.
checkO :- isProvided(T,'TYPE',X), isClass(T,X), isTerm(X), isNotEnv(X), errO5_2(X,T,O), 
	write('ERROR : Receiver of Operation : '), write(O), /*(cf.HOOD024)*/
	write(' provided by or internal to Class : '), write(X), 
	write(' should be of Type : '), write(T), 
	write('  ! (O5_2)'), nl.
checkO :- printlf('---> rule O5 checked...').
 
/*``Operation``Wed Feb 12 17:23:16 MET 2003``*/
/* HRM 4 rule O6: 
If an OPERATION is provided by or internal to a CLASS, 
and is INHERITED from an OPERATION provided by or internal to the superclass, 
their names shall be identical and their declaration shall match, except for the receiver. */

/* implementation note:
last changed: 12.02.03*/

errO6_1(X,T,O) :- isInherited(O,X), /*(cf.HOOD004)*/
	not(inherits(T,X,_,_)).

errO6_2(X,T,O) :- isInherited(O,X), inherits(T,X,S,Y), /*(cf.HOOD004)*/
	not(isProvided(O,'OPERATION',Y)),
	not(isInternal(O,'OPERATION',Y)).

errO6_3(X,T,O) :- isInherited(O,X), inherits(T,X,S,Y), /*(cf.HOOD004)*/
	argument(O,'OPERATION',X,M,P,_,_,_,_), /*(cf.HOOD015)(cf.HOOD016)*/
	not(argument(O,'OPERATION',Y,M,P,_,_,_,_)).

errO6_4(X,T,O) :- isInherited(O,X), inherits(T,X,S,Y), /*(cf.HOOD004)*/
	argument(O,'OPERATION',X,M,P,_,T,_,H), 
	memezTra(P,'me'), 
	not(hasArgO6(O,Y,'me',M,S,H)).

hasArgO6(O,X,V,M,T,H) :- argument(O,'OPERATION',X,M,P,_,T,_,H), memezTra(P,V).

checkO :- isTerm(X), isProvided(T,'TYPE',X), isClass(T,X), errO6_1(X,T,O), 
	write('ERROR : Class : '), write(X), 
	write(' has inherited Operation : '), write(O), 
	write(' but no superclass ! (O6_1)'), nl.
checkO :- isTerm(X), isProvided(T,'TYPE',X), isClass(T,X), errO6_2(X,T,O), 
	write('ERROR : superclass of Class : '), write(X), 
	write(' should have an Operation named : '), write(O), 
	write('  ! (O6_2)'), nl.
checkO :- isTerm(X), isProvided(T,'TYPE',X), isClass(T,X), errO6_3(X,T,O), 
	write('WARNING : a parameter name of Operation : '), write(O), /*(cf.HOOD016)*/
	write(' of Class : '), write(X), 
	write(' does not match a parameter name of the inheriting Operation ! (O6_3)'), nl.
checkO :- isTerm(X), isProvided(T,'TYPE',X), isClass(T,X), errO6_4(X,T,O), 
	write('ERROR : Operation : '), write(O), 
	write(' of superclass of Class : '), write(X), 
	write(' has an inconsistent receiver ! (O6_4)'), nl.
checkO :- printlf('---> rule O6 checked...').
 
/*``Operation``Thu Feb 25 11:06:19 MET 1999``*/
/* HRM 4 rule O7: 
An ABSTRACT OPERATION shall only be defined in an ABSTRACT CLASS.*/

errO7(X,O,T) :- not(isAbstract(T,'TYPE',X)).

checkO :- isAbstract(O,'OPERATION',X), errO7(X,O,T),
	write('ERROR : concrete Class : '), write(X), 
	write(' has abstract Operation : '), write(O), 
	write(' ! (O7)'), nl.
checkO :- printlf('---> rule O7 checked...').
 
