/*``changes``Wed Oct 23 17:19:48 MET 2002``*/
/* version */
/*-------------------------------------------------
	STOOD v4
	HOOD 4 rules checker 
	Required_Interface rules: HRM4 p99-100

	Copyright TNI 1999
	Pierre DISSAUX (pierre.dissaux@tni.fr)
	17.Mar.1999
	21.Jun.2000 (R17)
------------------------------------------------*/
/**** changes ****/
/*------------------------------------------------
HOOD014 (23.Oct.2002) R16 message changed to a warning
------------------------------------------------*/
/**** required interface rules ****/
/*-----------------------------------------------*/
 
/*``Required``Wed Mar 17 16:24:26 MET 1999``*/
/* HRM 4 rule R1: 
REQUIRED_MODULE field shall list all of (and only) actual MODULEs 
(excluding environment) i.e. siblings, uncles required by the MODULE.*/

/* implementation notes:
1- In Stood, Required Interface fields are automatically updated from 
cross references table
2- In Stood, Environments may be design as "high level" uncles.*/

checkR :- printlf('---> rule R1 is implicit.').
 
/*``Required``Tue Mar 16 11:18:07 MET 1999``*/
/* HRM 4 rule R2: 
REQUIRED_MODULE field shall list all of environment MODULEs 
required by the MODULE itself and not those required by its descendants.*/

/* implementation note:
In Stood, Required Interface fields are automatically updated from 
cross references table.*/

checkR :- printlf('---> rule R2 is implicit.').
 
/*``Required``Wed Mar 17 16:24:56 MET 1999``*/
/* HRM 4 rule R3: 
For each actual MODULE, the list of all required resources 
(types, classes, constants, operation_sets, operations, exceptions) 
shall be given.*/

/* implementation notes:
1- In Stood, Required Interface fields are automatically updated from 
cross references table.
2- In Stood, there is no Required Classes nor Required Operation_Sets 
fields. Required Classes may be found among required Types.*/

checkR :- printlf('---> rule R3 is implicit.').
 
/*``Required``Tue Mar 16 11:47:07 MET 1999``*/
/* HRM 4 rule R4: 
TYPES field shall list all of (and only) the following types: 
those required from sibling, environment or uncle MODULEs 
and referenced in: 
- the provided interface of this MODULE for type definitions, 
  constant declarations or operation parameter types
- the internals section of this MODULE for instantiations, type definitions,
  data definitions, constant definitions, operation parameter types.
- the required interface of any child.*/

/* implementation note:
In Stood, Required Interface fields are automatically updated from 
cross references table.*/

checkR :- printlf('---> rule R4 is implicit.').
 
/*``Required``Wed Mar 17 16:25:21 MET 1999``*/
/* HRM 4 rule R5: 
CONSTANTS field shall list all of (and only) the following constants: 
those required from sibling, environment or uncle MODULEs 
and referenced in: 
- the provided interface of this MODULE for type definitions  
  or constant declarations.
- the internals section of this MODULE for instantiations, type definitions,
  data definitions or constant definitions.
- the required interface of any child..*/

/* implementation notes:
1- In Stood, Required Interface fields are automatically updated from 
cross references table.
2- Reference to a remote Constant may also occur in a parameter 
default value in a provided or internal  Operation declaration*/

checkR :- printlf('---> rule R5 is implicit.').
 
/*``Required``Tue Mar 16 11:56:37 MET 1999``*/
/* HRM 4 rule R6: 
OPERATION_SETS field shall list all operation_set whose one operation 
at least is used by a MODULE.*/

/* implementation note:
In Stood, Required Interface fields are automatically updated from 
cross references table, and Operation_Sets are never directly 
referenced there.*/

checkR :- printlf('---> rule R6 not implemented.').
 
/*``Required``Tue Mar 16 11:57:44 MET 1999``*/
/* HRM 4 rule R7: 
OPERATIONS field shall list all operations used by a MODULE.*/

/* implementation note:
In Stood, Required Interface fields are automatically updated from 
cross references table.*/

checkR :- printlf('---> rule R7 is implicit.').
 
/*``Required``Tue Mar 16 11:58:53 MET 1999``*/
/* HRM 4 rule R8: 
EXCEPTIONS field shall list all exceptions that might be propagated by 
required operations.*/

/* implementation note:
In Stood, Required Interface fields are automatically updated from 
cross references table.*/

checkR :- printlf('---> rule R8 is implicit.').
 
/*``Required``Wed Mar 17 16:26:17 MET 1999``*/
/* HRM 4 rule R9: 
The FORMAL_PARAMETERS of a GENERIC can only be types or classes, 
constants or operations.*/

errR9(E) :- E\='TYPE', E\='CONSTANT', E\='OPERATION'.

checkR :- isProvided(U,E,'FORMAL_PARAMETERS'), errR9(E),
	write('ERROR : '), write(E), write(' : '), write(U), 
	write(' should not be a Formal_Parameter of current Generic ! (R9)'), nl.
checkR :- printlf('---> rule R9 checked...').
 
/*``Required``Wed Mar 17 16:26:40 MET 1999``*/
/* HRM 4 rule R10: 
FORMAL_PARAMETERS field in a GENERIC definition shall list all 
formal parameters i.e. types, main type of classes, constants and operations 
required by the generic itself or any of its descendant.*/

errR10(U,E) :- not(formalParameter(U,E,_)).

checkR :- requires(_,_,_,U,E,'FORMAL_PARAMETERS',_), errR10(U,E),
	write('ERROR : '), write(E), write(' : '), write(U), 
	write(' should  be a Formal_Parameter of current Generic ! (R10)'), nl.
checkR :- printlf('---> rule R10 checked...').
 
/*``Required``Wed Mar 17 10:57:50 MET 1999``*/
/* HRM 4 rule R11: 
FORMAL_PARAMETERS in the required interface of descendants of a 
generic definition shall be consistent with FORMAL_PARAMETERS 
declaration in the generic definition.*/

errR11(U,E) :- not(requires(_,_,_,U,E,'FORMAL_PARAMETERS',_)).

checkR :- isCurrentRoot(R), isRootObject(G,'GENERIC',_), memezTra(R,G), 
	formalParameter(U,E,_), errR11(U,E),
	write('ERROR : Formal_Parameter '), write(E), write(' : '), write(U), 
	write(' should  be required within current Generic ! (R11)'), nl.
checkR :- printlf('---> rule R11 checked...').
 
/*``Required``Wed Mar 17 10:15:23 MET 1999``*/
/* HRM 4 rule R12: 
The required items listed in an INSTANCE_OF a GENERIC 
shall be complete and compliant, with respect to parameter 
matching rules with the formal parameter declaration 
in the GENERIC definition.*/

errR12(I) :- not(requires(_,_,I,_,_,_,'InstPars')), !.

checkR :- isInstance(I,_,G), errR12(I),
	write('ERROR : Required Interface of Instance_Of  Module : '), write(I), 
	write(' do not match Formal_Parameters of relevant Generic Module : '), write(G), 
	write('  ! (R12)'), nl.
checkR :- printlf('---> rule R12 checked...').
 
/*``Required``Wed Mar 17 10:22:26 MET 1999``*/
/* HRM 4 rule R13: 
A formal type shall be instanciated by an actual type or class.*/

errR13(U,I) :- not(actualParameter(U,'TYPE',I,_)).

checkR :- isInstance(I,_,G), formalParameter(U,'TYPE',G), errR13(U,I),
	write('ERROR : formal TYPE : '), write(U), write(' of Generic  Module : '), write(G), 
	write(' is not instanciated by an actual TYPE of  Instance_Of Module : '), write(I), 
	write('  ! (R13)'), nl.
checkR :- printlf('---> rule R13 checked...').
 
/*``Required``Wed Mar 17 10:24:13 MET 1999``*/
/* HRM 4 rule R14: 
A formal constant shall be instanciated by an actual constant or value.*/

errR14(U,I) :- not(actualParameter(U,'CONSTANT',I,_)).

checkR :- isInstance(I,_,G), formalParameter(U,'CONSTANT',G), errR14(U,I),
	write('ERROR : formal CONSTANT : '), write(U), write(' of Generic  Module : '), write(G), 
	write(' is not instanciated by an actual CONSTANT of  Instance_Of Module : '), write(I), 
	write('  ! (R14)'), nl.
checkR :- printlf('---> rule R14 checked...').
 
/*``Required``Wed Mar 17 10:25:30 MET 1999``*/
/* HRM 4 rule R15: 
A formal operation shall be instanciated by an actual operation.*/

errR15(U,I) :- not(actualParameter(U,'OPERATION',I,_)).

checkR :- isInstance(I,_,G), formalParameter(U,'OPERATION',G), errR15(U,I),
	write('ERROR : formal OPERATION : '), write(U), write(' of Generic  Module : '), write(G), 
	write(' is not instanciated by an actual OPERATION of  Instance_Of Module : '), write(I), 
	write('  ! (R15)'), nl.
checkR :- printlf('---> rule R15 checked...').
 
/*``Required``Wed Oct 23 17:18:10 MET 2002``*/
/* HRM 4 rule R16: 
The parameters of an INSTANCE_OF shall correspond 
to the formal parameters of the GENERIC definition.*/

/* implementation note:
can't be checked if not "within syst conf"
last changed: 23.10.02 */

errR16(I,G) :- actualParameter(U,E,I,_), not(formalParameter(U,E,G)), !.

checkR :- isInstance(I,_,G), errR16(I,G),
	write('WARNING : Actual_Parameters of Instance_Of  Module : '), write(I), 
	write(' may not match Formal_Parameters of relevant Generic Module : '), write(G), 
	write(' - use "check within syst conf"  ! (R16)'), nl. /*(cf.HOOD014)*/
checkR :- printlf('---> rule R16 checked...').
 
/*``Required``Wed Jun 21 17:34:08 MET DST 2000``*/
/* STOOD rule R17: 
If an entity is required in a pseudo_code section, 
then it should be also required for target language.*/

/* implementation note: 
hood/config file must contain:
XLangs	("pseudo" ). 
last modified: 21 Jun 2000 */

errR17(U,E,X,V,F,Y) :- not(requires(U,E,X,V,F,Y,_)).

checkR :- requires('pseudo',U,E,X,V,F,Y,_), 
	not(isImplementedBy(U,E,X,V,Y,_)), errR17(U,E,X,V,F,Y), 
	isCurrentRoot(R), pragma_TARGET_LANGUAGE(R,L),
	write('WARNING : '), write(E), write(' : '), write(U),
	write(' of Module : '), write(X),
	write(' should require '), write(F), write(' : '), write(V), 
	write(' of Module : '), write(Y), 
	write(' to comply with pseudo_code ! (R17)'), nl.
checkR :- printlf('---> additional rule R17 checked...').
 
