/*``changes``Wed Oct 23 16:42:27 MET 2002``*/
/* version */
/*-------------------------------------------------
	STOOD v4.1
	HOOD 4 rules checker 
	Use rules: HRM4 p94-95

	Copyright TNI 1999-2001
	Pierre DISSAUX (pierre.dissaux@tni.fr)
	17.Mar.1999
	29.Jul.1999: U1, U2
	11.Jan.2000: U1, U2
	21.Jun.2000: U17
------------------------------------------------*/
/**** changes ****/
/*------------------------------------------------
HOOD002 (17.May.2001) U17 is implicitly checked by code generation
HOOD013 (23.Oct.2002) U14 case sensitive check
HOOD017 (30.Apr.2004) U2 accepts simple names for attributes and inherits
HOOD018 (30.Apr.2004) U13_3 fixing bad error message
------------------------------------------------*/
/**** use relationship rules ****/
/*-----------------------------------------------*/
 
/*``Use``Tue Jan 11 16:49:32 MET 2000``*/
/* HRM 4 rule U1: 
A MODULE OP_USES another MODULE when it requires at least 
one of its provided OPERATIONs. */

/* Implementation note:
last modified: 29 Jul 1999*/

ruleU1(X,Y) :- requires(_,_,X,_,'OPERATION',Z,_), memezTra(Y,Z).

checkU :- isTerm(X), not(isRootObject(X,_,_)), 
	uses(X,Y,'OPERATION',_), not(ruleU1(X,Y)), 
	write('ERROR : Module : '), write(X), 
	write(' should require at least one OPERATION from Module : '), write(Y), 
	write(' ! (U1)'), nl.
checkU :- printlf('---> rule U1 checked...').
 
/*``Use``Tue Jan 11 16:57:37 MET 2000``*/
/* HRM 4 rule U2: 
A MODULE TYPE_USES another MODULE when it requires at least 
one of its provided TYPEs. */

/* Implementation note:
last modified: 30 Apr 2004*/

ruleU2(X,Y,_) :- requires(_,_,X,_,'TYPE',Z,_), memezTra(Y,Z).
ruleU2(X,Y,2) :- inherits(_,X,_,Z), memezTra(Y,Z).
ruleU2(X,Y,2) :- inherits(_,X,Z,_), memezTra(Y,Z).
ruleU2(X,Y,3) :- attributes(_,X,_,_,Z,_), memezTra(Y,Z).
ruleU2(X,Y,3) :- attributes(_,X,_,Z,_,_), memezTra(Y,Z).

checkU :- isTerm(X), not(isRootObject(X,_,_)), 
	uses(X,Y,'TYPE',K), not(ruleU2(X,Y,K)), 
	write('ERROR : Module : '), write(X), 
	write(' should require at least one TYPE from Module : '), write(Y), 
	write(' ! (U2)'), nl.
checkU :- printlf('---> rule U2 checked...').
 
/*``Use``Thu Feb 11 15:11:26 MET 1999``*/
/* HRM 4 rule U3: 
An UNCLE of a CHILD MODULE is a MODULE used by its PARENT. */

checkU :- printlf('---> rule U3 is a definition.').
 
/*``Use``Thu Feb 11 15:28:08 MET 1999``*/
/* HRM 4 rule U4: 
If a child MODULE OP_USES an UNCLE, then it (this uncle) shall also be 
OP_USED by its PARENT. This rule is implicit for ENVIRONMENTs. */

errU4(X,P,O) :- uses(X,O,'OPERATION',_),
	not(isRootObject(O,_,_)),
	not(hoodObject(O,'ENVIRONMENT',_)),
	not(uses(P,O,'OPERATION',_)).

checkU :- hoodObject(X,_,P), brothers(P,O), errU4(X,P,O),
	write('ERROR : Module : '), write(O),
	write(' should be Op_Used by Module : '), write(P),
	write(' ! (U4)'), nl.
checkU :- printlf('---> rule U4 checked...').
 
/*``Use``Thu Feb 11 15:47:05 MET 1999``*/
/* HRM 4 rule U5: 
If a PARENT OP_USES another MODULE, then at least one of its children shall 
also OP_USE it (the other module).*/

errU5(P,O) :- not( hoodObject(C,_,P), uses(C,O,'OPERATION',_) ).

checkU :- isNotTerm(P), uses(P,O,'OPERATION',_),errU5(P,O),
	write('ERROR : At least one child of Module : '), write(P),
	write(' should Op_Use Module : '), write(O),
	write(' ! (U5)'), nl.
checkU :- printlf('---> rule U5 checked...').
 
/*``Use``Wed Mar 17 19:06:24 MET 1999``*/
/* HRM 4 rule U6: 
If a child MODULE TYPE_USES an UNCLE, then it (this uncle) shall also be 
TYPE_USED by its PARENT. This rule is implicit for ENVIRONMENTs. */

errU6(X,P,O) :- uses(X,O,'TYPE',_),
	not(isRootObject(O,_,_)),
	not(hoodObject(O,'TYPE',_)),
	not(uses(P,O,'TYPE',_)), !.

checkU :- hoodObject(X,_,P), brothers(P,O), errU6(X,P,O),
	write('ERROR : Module : '), write(O),
	write(' should be Type_Used by Module : '), write(P),
	write(' ! (U6)'), nl.
checkU :- printlf('---> rule U6 checked...').
 
/*``Use``Thu Feb 25 09:23:15 MET 1999``*/
/* HRM 4 rule U7: 
If a PARENT TYPE_USES another MODULE, then at least one of its children shall 
also TYPE_USE it (the other module).*/

errU7(P,O) :- not( hoodObject(C,_,P), uses(C,O,'TYPE',_) ).

checkU :- isNotTerm(P), uses(P,O,'TYPE',_), errU7(P,O),
	write('ERROR : At least one child of Module : '), write(P),
	write(' should Type_Use Module : '), write(O),
	write(' ! (U7)'), nl.
checkU :- printlf('---> rule U7 checked...').
 
/*``Use``Thu Feb 11 16:15:07 MET 1999``*/
/* HRM 4 rule U8: 
An OP_CONTROL may use any OPERATION, constrained or not. */

/* Implementation note:
This is allowed by STOOD v4 */

checkU :- printlf('---> rule U8 is implicit.').
 
/*``Use``Wed Feb 24 18:26:42 MET 1999``*/
/* HRM 4 rule U9: 
Passive MODULEs should not USE each other in a cycle. */

errU9(X) :- uses(X,Y,'OPERATION',_),
	hoodObject(Y,'PASSIVE',_), loopU9(X,Y,[]), !.

loopU9(X,X,_) :- true, !.
loopU9(X,Y,L1) :- uses(Y,Z,'OPERATION',_), isNewU9(Z,L1,L2), loopU9(X,Z,L2).

isNewU9(E,L,[E|L]) :- hoodObject(E,'PASSIVE',_), 
	not($member(E,L)).

checkU :- hoodObject(X,'PASSIVE',_), errU9(X), 
	write('WARNING : Passive Module : '), write(X), 
	write(' should not use other passive Modules in a cycle ! (U9)'), nl.
checkU :- printlf('---> rule U9 checked...').
 
/*``Use``Wed Feb 24 16:43:37 MET 1999``*/
/* HRM 4 rule U10: 
A VIRTUAL_NODE shall only use other VIRTUAL_NODEs. */

errU10(X) :- uses(X,Y,_,_),
	hoodObject(Y,K,_), K \= 'VIRTUAL_NODE'.

checkU :- hoodObject(X,'VIRTUAL_NODE',_), errU10(X), 
	write('ERROR : Virtual_Node : '), write(X), 
	write(' should use only other Virtual_Nodes ! (U10)'), nl.
checkU :- printlf('---> rule U10 checked...').
 
/*``Use``Wed Feb 24 17:14:06 MET 1999``*/
/* HRM 4 rule U11:
A PASSIVE (not OP_CONTROL) MODULE may OP_USE an ACTIVE MODULE, 
only if it requires non constrained or state constrained OPERATIONs. */

/* Implementation note:
Use of concurrency constrained OPERATIONs should also be allowed */

errU11(X,Y) :- 
	requires(_,'OPERATION',X,O,'OPERATION',Y,_), 
	isConstrained(O,'OPERATION',Y,C,_), 
	( C='HSER'; C='LSER'; C='ASER' ).

checkU :- hoodObject(X,'PASSIVE',_), uses(X,Y,'OPERATION',_), errU11(X,Y), 
	write('ERROR : Passive Module : '), write(X), 
	write(' should not use protocol constrained Operations of Module : '), write(Y), write(' ! (U11)'), nl.
checkU :- printlf('---> rule U11 checked...').
 
/*``Use``Wed Feb 24 17:17:58 MET 1999``*/
/* HRM 4 rule U12:
OP_USE relationships between children of a parent 
should not be cyclic. */

/* Implementation note:
This rule seems too restrictive and redundant with U9 */

checkU :- printlf('---> rule U12 not implemented').
 
/*``Use``Wed Feb 24 18:18:14 MET 1999``*/
/* HRM 4 rule U13:
A CLASS X INHERITS from CLASS Y if main TYPE of X 
is a subclass of main TYPE of Y. */

/* Implementation note:
last modified: 30 Apr 2004*/

errU13_1(X) :- not(isClass(_,X)).
errU13_2(Y) :- not(isClass(_,Y)).
errU13_3(X,Y,U,V) :- isClass(U,X), isClass(V,Y), not(inherits(U,X,V,Y)).

checkU :- uses(X,Y,'TYPE',2), errU13_1(X), 
	write('ERROR : Module : '), write(X), 
	write(' origin of an Inherits relationship, should be a Class ! (U13_1)'), nl.
checkU :- uses(X,Y,'TYPE',2), errU13_2(Y), 
	write('ERROR : Module : '), write(Y), 
	write(' end of an Inherits relationship, should be a Class ! (U13_2)'), nl.
checkU :- uses(X,Y,'TYPE',2), errU13_3(X,Y,U,V), 
	write('ERROR : Type : '), write(U), write(' of Module : '), write(X), 
	write(' should be a subclass of Type : '), write(V), write(' of Module : '), write(Y), 
	write(' ! (U13_3)'), nl.
checkU :- printlf('---> rule U13 checked...').
 
/*``Use``Wed Oct 23 16:41:44 MET 2002``*/
/* HRM 4 rule U14:
A CLASS X ATTRIBUTES CLASS Y if main TYPE of X 
is composed of at least one element defined as an instance 
of main TYPE of Y. */

/* Implementation note:
This rule seems a bit too restrictive : in some cases, 
attributes may be defined between non class modules. 
last changed: 23.10.02*/

errU14(X,Y) :- not(attributes(_,X,_,_,Z,_)), same(Y,Z). /*(cf.HOOD013)*/

checkU :- uses(X,Y,'TYPE',3), errU14(X,Y), 
	write('ERROR : a Type of Module : '), write(X), 
	write(' should have an Attribute of a Type of Module : '), write(Y), 
	write(' ! (U14)'), nl.
checkU :- printlf('---> rule U14 checked...').
 
/*``Use``Wed Feb 24 18:42:18 MET 1999``*/
/* HRM 4 rule U15: 
INHERITS relationships shall not be cyclic. */

errU15(X,Y) :- loopU15(X,Y,[]), !.

loopU15(X,X,_) :- true, !.
loopU15(X,Y,L) :- uses(Y,Z,'TYPE',2), not($member(Z,L)), loopU15(X,Z,[E|L]).

checkU :- uses(X,Y,'TYPE',2), errU15(X,Y), 
	write('ERROR : Inherits relationship should not be cyclic, from Module : '), 
	write(X), write(' ! (U15)'), nl.
checkU :- printlf('---> rule U15 checked...').
 
/*``Use``Wed Mar 17 19:08:57 MET 1999``*/
/* HRM 4 rule U16: 
If a CLASS INHERITs from, or ATTRIBUTEs an UNCLE, the latter shall be 
TYPE_USED by its PARENT. */

errU16(X,P,O) :- uses(X,O,'TYPE',_),
	not(isRootObject(O,_,_)),
	not(hoodObject(O,'ENVIRONMENT',_)),
	not(uses(P,O,'TYPE',_)), !.

checkU :- hoodObject(X,_,P), isClass(_,X), brothers(P,O), 
	errU16(X,P,O),
	write('ERROR : Module : '), write(O),
	write(' should be Type_Used by Module : '), write(P),
	write(' ! (U16)'), nl.
checkU :- printlf('---> rule U16 checked...').
 
/*``Use``Thu May 17 15:42:34 MET DST 2001``*/
/* HRM 4 rule U17: 
A main TYPE of a CLASS shall only INHERIT from another main TYPE 
required by the CLASS. */

/* last changed: 17 May 2000 
this rule is no more checked as inheritance code may be 
automatically generated in Ada or C 

errU17(U,X,V,Y) :- not(requires(U,'TYPE',X,V,'TYPE',Y,_)). (cf.HOOD002)

checkU :- inherits(U,X,V,Y), errU17(U,X,V,Y),
	write('ERROR : Type : '), write(V), write(' of Module : '), write(Y), 
	write(' should be required by its subclass : '), write(U), 
	write(' of Module : '), write(X),
	write(' ! (U17)'), nl. 
*/
checkU :- printlf('---> rule U17 no more checked...').
 
