/*-------------------------------------------------
  STOOD 5.0
  HRT schedulability analyser

  Copyright TNI Europe 2004
  Pierre DISSAUX (pierre.dissaux@tni-world.com)
  v1.1 - 04.May.2004
--------------------------------------------------*/
/*--------------------------------------------------
changes: new layout
--------------------------------------------------*/

rma :- 
	writeSummary, 
	ubTest, 
	checkRMPriority, 
	checkDMPriority, 
	checkUtilization, 
	checkBound, 
	checkDeadlines, 
	writeCaption.

/*--------------------------------------------------
get data from the design model
--------------------------------------------------*/

/* look for HRT tasks in the design model */

task(X) :- hoodObject(X,K,_), ( K='CYCLIC'; K='SPORADIC' ).
 
/* calculate the number of tasks */

nbTasks(N) :- cpt2Rst, task(X), cpt2Inc, fail.
nbTasks(N) :- cpt2Get(N).
 
/* get the Period of a task */

getT(X,T) :- 
	hrtPeriod(X,_,T1) -> T is T1; 
	hrtMinArrivalTime(X,_,T2) -> T is T2;
	T is 0.

/* get the Worst Case Execution Time of a task */

getC(X,C) :- hrtWcet(X,'thread',_,C1) -> C is C1; C is 0.

/* getP(X,P) */
/* get the Priority of a task */

getP(X,P) :- hrtPriority(X,_,P1) -> P is P1; P is 0.

/* get the Deadline for a task */

getD(X,D) :- 
	hrtDeadline(X,_,D1) -> D is D1;
	hrtPeriod(X,_,D2) -> D is D2;
	hrtMinArrivalTime(X,_,D3) -> D is D3;
	D is 0.

/*--------------------------------------------------
calculate schedulability data
--------------------------------------------------*/

/* calculate the Utilization ratio for a task */

getU(T,C,U) :- T \= 0,  K is 100 * C, U is K / T, cpt3Add(U).
getU(0,_,0).
 
/* calculate the Response Time for a task */

getR(X,C,R) :- w(X,C,C,R).

w(X,C,W,'N/A') :- getT(X,T), W > T, !.
w(X,C,W,R) :- cptRst, blocking(X,B), hp(X,W,I), S is C + B + I, 
	( S =\= W ->  w(X,C,S,R); R = S ).

hp(X,W,I) :- cptRst, getP(X,Px), task(Y), Y \= X, getP(Y,Py), Py > Px, 
	getT(Y,T), getC(Y,C), ceiling(W,T,V), A is V * C, cptAdd(A), fail.
hp(X,W,I) :- cptGet(I).

ceiling(W,T,V) :- ( W mod T =:= 0 ) -> ( V is W // T ); ( V is ( W // T ) + 1 ).

blocking(X,B) :- getP(X,Px), uses(X,P,'OPERATION',1), isObject(P,'PROTECTED',_),
	task(Y), getP(Y,Py), Py < Px, uses(Y,P,'OPERATION',1), 
	getC(Y,C), cptAdd(C), fail.
blocking(X,B) :- cptGet(B).
 
/* getS(V) */
/* calculate the total utilization ratio */

getS(V) :- cpt3Get(V).
 
/* getB(V) */
/* calculate the utilization bounds */

getB(V) :- nbTasks(N), N \= 0, $exp(A,2), $exp(B,A), $exp(C,N), D is B-C, $exp(D,E), $exp(E,F), 
	G is 100*N*(F-1), $floor(G,V).
getB(0) :- nbTasks(0).
 
/*--------------------------------------------------
utilities
--------------------------------------------------*/

/* text output utilities */

writeLine(M,C) :- fillIn(0,M,C).

fillIn(L,M,C) :- L >= M, !.
fillIn(L,M,C) :- $put(C), K is L+1, fillIn(K,M,C).

sp :- $put(32).
sep :- $put(124).

writeTrunc(X,M) :- $name(X,[C|L]), trunc(C,L,1,M).

trunc(C,[],_,_) :- $put(C),!.
trunc(C,L,P,M) :- P > M, !.
trunc(C,[D|L],P,M) :- $put(C), Q is P+1, trunc(D,L,Q,M).
 
/* text input utilities */

readFile(V) :- readNum(V,0).

readNum(Vo,Vi) :- $get0(C), ( C =:= -1 ->  
	( Vo is Vi );
	( C =:= 46 -> 
		( Vo is Vi );
		( A is C - 48, B is 10 * Vi, V is A + B, readNum(Vo,V) ) ) ).
 
/*--------------------------------------------------
print schedulability analysis report
--------------------------------------------------*/

/* write HRT parameters summary */

writeSummary :- writeSummaryHeader, cpt3Rst, fail.
writeSummary :- task(X), getT(X,T), getC(X,C), getP(X,P), getU(T,C,U), getD(X,D), getR(X,C,R), 
	writeParams(X,T,C,P,U,D,R), fail.
writeSummary :- writeLine(72,45), nl, fail.
writeSummary.

writeSummaryHeader :- 
	writeLine(72,45), nl,
	write('| tasks:               |   T   |   W   |   P   |   U   |   D   |   R   |'), nl,
	writeLine(72,45), nl.

writeParams(X,T,C,P,U,D,R) :- 
	sep, writeName(X), sep, writeInt(T), sep, writeInt(C), sep, writeInt(P), 
	sep, writePer(U), sep, writeInt(D), sep, writeInt(R), sep, nl.

writeName(X) :- $conlength(X,L), sp, writeTrunc(X,20), ( L > 20 -> sp; fillIn(L,21,32) ).

writeInt('N/A') :- write('  N/A  '), !.
writeInt(X) :-  $conlength(X,L), fillIn(L,6,32), write(X), sp.

writePer('N/A') :- write('  N/A  '), !.
writePer(X) :- $conlength(X,L), fillIn(L,5,32), write(X), $put(37), sp.
 
/* writeCaption */
 	
writeCaption :- 
	writeLine(32,45), nl,
	write('| T: Period                    |'), nl, 
	write('| W: Worst Case Execution Time |'), nl, 
	write('| P: Priority                  |'), nl, 
	write('| U: Utilization               |'), nl, 
	write('| D: Deadline                  |'), nl, 
	write('| R: Response Time             |'), nl,
	writeLine(32,45), nl, fail.
writeCaption.

/*--------------------------------------------------
check schedulability 
--------------------------------------------------*/

/* utilization-based test */

ubTest :- 
	sep, write(' summation:'), fillIn(12,48,32), getS(S), writePer(S), fillIn(56,72,32), sep, nl, 
	sep, write(' bound:'), fillIn(8,48,32), getB(B), writePer(B), fillIn(56,72,32), sep, nl, 
	writeLine(72,45), nl, fail.
ubTest.
 
/* check Rate Monotonic Priority Ordering */

checkRMPriority :- 
	task(X), task(Y), Y \= X, 
	getT(X,T1), getP(X,P1), getT(Y,T2), getP(Y,P2), 
	T1 < T2, P1 =< P2,
	write('WARNING: Rate Monotonic Priority Ordering is not respected'), nl, !.
checkRMPriority :-
	write('INFORMATION: Rate Monotonic Priority Ordering is respected'), nl.
 
/* check Deadline Monotonic Priority Ordering */

checkDMPriority :- 
	task(X), task(Y), Y \= X, 
	getD(X,D1), getP(X,P1), getD(Y,D2), getP(Y,P2), 
	D1 < D2, P1 =< P2,
	write('WARNING: Deadline Monotonic Priority Ordering is not respected'), nl, !.
checkDMPriority :-
	write('INFORMATION: Deadline Monotonic Priority Ordering is respected'), nl.
 
/* check Utilization ratio */

checkUtilization :- 
	getS(S), S > 100, write('WARNING: Total Utilization is greater than 100%'), nl, fail.
checkUtilization.
 
/* check Utilization towards Bound */

checkBound :- 
	getS(S), getB(B), S > B, write('WARNING: Total Utilization is out of Bound'), nl, fail.
checkBound.
 
/* check Deadlines */

checkDeadlines :- 
	task(X), getD(X,D), getC(X,C), getR(X,C,R), R \= 'N/A', R > D, 
	write('WARNING: Missed Dealine for: '), write(X), nl, fail.
checkDeadlines.
