/*-----------------------------------------*/
/*  STOOD v4  */
/*  Ada lexical analyser  */
/*                                      */
/*  Copyright TNI 1992-2000  */
/*  20.sep.2000  */
/*-----------------------------------------*/

Letter		[A-Za-z]
Digit		[0-9]
Extended_Digit	[0-9A-F]
Identifier	{Letter}(_?({Letter}|{Digit}))*
Exponent	E[-+]?{Digit}+
%%

{Identifier}			identifier();
\.				dot();
[ \t\n]+			;		/* separators */
\"[^\"]*\"			operator();	/* quoted operator */

\"([^\"]|\"\")*\"		anything();	/* string literals */
{Digit}+("."{Digit}+)?({Exponent})?		anything();	/* decimal number*/
{Digit}+#{Extended_Digit}+	anything();	/* based literals */
\'.\'				anything();	/* character literals */
\'{Identifier}			attribute();	/* attribute designator */
--[^\n]*			;		/* comments */
.				anything(); 	/* anything else */
"use at"			address();
%%

#define IgnoreCase 1
char *ReservedWordList[] = {
	"abort", "abs", "accept", "access", "all", "and", "array", "at", "begin", "body",
	"case", "constant", "declare", "delay", "delta", "digits", "do", "else", "elsif",
	"end", "entry", "exception", "exit", "for", "function", "generic", "goto", "if",
	"in", "is", "limited", "loop", "mod", "new", "not", "null", "of", "or", "others",
	"out", "package", "pragma", "private", "procedure", "raise", "range",
	"record", "rem", "renames", "return", "reverse", "select", "separate",
	"subtype", "task", "terminate", "then", "type", "use", "when", "while",
	"with", "xor",
	"abstract", "aliased", "protected", "requeue", "tagged", "until",
0};

char *ReservedOperationName[] = {
	"\"and\"", "\"or\"", "\"xor\"", "\"=\"", "\"/=\"", "\"<\"", "\"<=\"", "\">\"",
	"\">=\"", "\"+\"", "\"-\"", "\"&\"", "\"*\"", "\"/\"", "\"mod\"",
	"\"rem\"", "\"**\"", "\"abs\"", "\"not\"",
0};




/**********************************************************
	this code is independant of the langage :
**********************************************************/

int State = 0;
int BoL = 1;

identifier()
{
/*	static int BoL = 1; */

	if(isReserved(yytext)) {
		State = 0;
		return;
	}
	if (State == 2) {
		printf(" ");
	} else {
		if (!BoL) {printf("\n"); BoL = 1;}
	}
	printf("%s", yytext);
	BoL = 0;
	State = (State == 2) ? 0 : 1;
}

operator()
{
	if(isReservedOperator(yytext)) {
		identifier();
	}
}

address()
{
	if (!BoL) {printf("\n"); BoL = 1;}
	printf("system address\n");
	State = 0;
}

attribute()
{
	if(stringCompare(yytext,"'address")) {
		if (!BoL) {printf("\n"); BoL = 1;}
		printf("system address\n");
	}
	State = 0;
}

dot()
{
	State = (State == 1) ? 2 : 0;
}

anything()
{
	State = 0;
}

int isReserved(word)
char *word;
{
	char **p = ReservedWordList;
	while (*p) {
		if (stringCompare(word, *p)) return 1;
		p++;
	}
	return 0;
}

int isReservedOperator(word)
char *word;
{
	char **p = ReservedOperationName;
	while (*p) {
		if (stringCompare(word, *p)) return 1;
		p++;
	}
	return 0;
}
int stringCompare(s1, s2)
char *s1, *s2;
{
	do {
		register char c1 = *(s1++);
		register char c2 = *(s2++);
		if (c1 == 0) return (c2 == 0);
		if (c2 == 0) return (c1 == 0);
#if IgnoreCase == 1
		if ('A' <= c1 && c1 <= 'Z' ) c1 += ('a' - 'A');
		if ('A' <= c2 && c2 <= 'Z' ) c2 += ('a' - 'A');
#endif
		if (c1 != c2) return (0);
	} while (1);
}

int yywrap()
{
	return 1;
}
	
main()
{
	yylex();
}
