/*--------------------------------------------
--		librairie de communication taskCom.h
--		- pthreads POSIX
--
--				X. Mehaut	02/02/98
--				A. Thura	26/02/98
----------------------------------------------*/
#ifndef INCLUDE_taskCom
#define INCLUDE_taskCom

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/resource.h>

/*--------------------------------*/
/*dfinition des macros de debuggage  */
/*--------------------------------*/
#ifndef SPY

#define DEBUG 1
#ifdef DEBUG
#define SPY(TEXT) printf(#TEXT "\n");
#else
#define SPY(TEXT)
#endif

#endif

/*--------------------------------*/
/*dfinition des macros de channel*/
/*--------------------------------*/
#define BUFFER_SIZE 1

#define TCHANNEL(CHANNEL_NAME, CHANNEL_TYPE, CHANNEL_SIZE, DIMS) \
		typedef CHANNEL_TYPE CHANNEL_NAME##_buffer_type##DIMS; \
		typedef struct { \
			int maxEltNum , currentSize, in_cursor, out_cursor; \
			CHANNEL_NAME##_buffer_type buffer[BUFFER_SIZE]DIMS; \
			pthread_mutex_t monitor; \
			pthread_cond_t  isEmpty, isFull; \
			int EOR; \
			int EOS; \
		} CHANNEL_NAME##_type; \
		static CHANNEL_NAME##_type  CHANNEL_NAME;

#define CREATE_TCHANNEL(CHANNEL_NAME, CHANNEL_TYPE, CHANNEL_SIZE) \
	CHANNEL_NAME.maxEltNum = BUFFER_SIZE  ; \
	CHANNEL_NAME.currentSize = 0;  \
	CHANNEL_NAME.in_cursor = 0; \
	CHANNEL_NAME.out_cursor = 0; \
	pthread_mutex_init(&(CHANNEL_NAME.monitor), NULL ); \
	pthread_cond_init(&(CHANNEL_NAME.isEmpty), NULL ); \
	pthread_cond_init(&(CHANNEL_NAME.isFull), NULL ); \
	CHANNEL_NAME.EOR = 0;\
	CHANNEL_NAME.EOS = 0;

#define START_TCHANNEL(CHANNEL_NAME)

/* destroy */
#define CLOSE_TCHANNEL(CHANNEL_NAME) \
	SPY(CHANNEL_NAME closed) \
	pthread_mutex_lock(&(CHANNEL_NAME.monitor));\
	CHANNEL_NAME.EOR = 1;\
	pthread_cond_broadcast(&(CHANNEL_NAME.isFull));\
	FLUSH_CHANNEL(CHANNEL_NAME)\
	pthread_mutex_unlock(&(CHANNEL_NAME.monitor));


#define TSEND(CHANNEL_NAME,MSG, SIZE, FCTTERM) \
	SPY(send SIZE into CHANNEL_NAME)\
	pthread_mutex_lock(&(CHANNEL_NAME.monitor)); \
	while ((CHANNEL_NAME.currentSize == CHANNEL_NAME.maxEltNum) && (!AT_END_OF_RECEIVER(CHANNEL_NAME))) pthread_cond_wait(&(CHANNEL_NAME.isFull),&(CHANNEL_NAME.monitor)); \
	if (AT_END_OF_RECEIVER(CHANNEL_NAME)) {pthread_mutex_unlock(&(CHANNEL_NAME.monitor)); FCTTERM(); } \
	memcpy(&(CHANNEL_NAME.buffer[CHANNEL_NAME.in_cursor]), &(MSG),sizeof(CHANNEL_NAME##_buffer_type)); \
	CHANNEL_NAME.in_cursor = (CHANNEL_NAME.in_cursor + 1) % CHANNEL_NAME.maxEltNum; \
	(CHANNEL_NAME.currentSize)++; \
	pthread_cond_broadcast(&(CHANNEL_NAME.isEmpty)); \
	pthread_mutex_unlock(&(CHANNEL_NAME.monitor)); \
	SPY(send SIZE into CHANNEL_NAME term)

#define TRECEIVE(CHANNEL_NAME, MSG, SIZE, FCTTERM) \
	SPY(wait SIZE from CHANNEL_NAME) \
	pthread_mutex_lock(&(CHANNEL_NAME.monitor)); \
	while ((CHANNEL_NAME.currentSize == 0) && (!AT_END_OF_SENDER(CHANNEL_NAME))) pthread_cond_wait(&(CHANNEL_NAME.isEmpty),&(CHANNEL_NAME.monitor)); \
	if ((CHANNEL_NAME.currentSize == 0) && (AT_END_OF_SENDER(CHANNEL_NAME))) {pthread_mutex_unlock(&(CHANNEL_NAME.monitor)); FCTTERM(); } \
	memcpy(&(MSG),&(CHANNEL_NAME.buffer[CHANNEL_NAME.in_cursor]),sizeof(CHANNEL_NAME##_buffer_type)); \
	CHANNEL_NAME.out_cursor = (CHANNEL_NAME.out_cursor +1) % CHANNEL_NAME.maxEltNum; \
	(CHANNEL_NAME.currentSize)--; \
	pthread_cond_broadcast(&(CHANNEL_NAME.isFull)); \
	pthread_mutex_unlock(&(CHANNEL_NAME.monitor)); \
	SPY(rec SIZE from CHANNEL_NAME term)


#define SEND_EOT(CHANNEL_NAME) \
	SPY(end of task send on CHANNEL_NAME) \
	pthread_mutex_lock(&(CHANNEL_NAME.monitor));\
	CHANNEL_NAME.EOS = 1;\
	pthread_cond_broadcast(&(CHANNEL_NAME.isEmpty));\
	pthread_mutex_unlock(&(CHANNEL_NAME.monitor));

#ifndef AT_END_OF_SENDER
#define AT_END_OF_SENDER(CHANNEL_NAME) CHANNEL_NAME.EOS
#define AT_END_OF_RECEIVER(CHANNEL_NAME) CHANNEL_NAME.EOR
#endif

#define FLUSH_CHANNEL(CHANNEL_NAME) \
	CHANNEL_NAME.currentSize = 0; \
	CHANNEL_NAME.in_cursor = 0; \
	CHANNEL_NAME.out_cursor = 0;

/*------------------------------*/
/*dfinition des macros de tache*/
/*------------------------------*/

#define MAXPRIO 0

#define PROGRAM(ProgramName, Task_Count)  static int TaskCount = Task_Count;

/* declaration du prototype et de l'id du thread*/
#define TASK(taskName) \
	void *exec_##taskName(void *);\
	pthread_t taskName##_id;\
	pthread_attr_t taskName##_attr;


#define INIT(progName) init_##progName()
#define INIT_HEADER() setpriority( PRIO_PROCESS , 0, MAXPRIO);
#define INIT_TRAILER() SPY(init_trailer)

#define CREATE_TASK(taskName, priority)\
	pthread_attr_init(&taskName##_attr);\
	SPY(CREATE_TASK taskName)\
	pthread_attr_setdetachstate(&taskName##_attr,PTHREAD_CREATE_DETACHED); \
/* 	pthread_attr_setscope(&taskName##_attr,PTHREAD_SCOPE_SYSTEM); */\
	pthread_attr_setinheritsched(&taskName##_attr, PTHREAD_EXPLICIT_SCHED); \
	{\
	struct sched_param param;\
	pthread_attr_getschedparam(&taskName##_attr, &param);\
	param.sched_priority = 20 - priority;\
	pthread_attr_setschedparam(&taskName##_attr, &param); }

#define START_TASK(taskName) \
	SPY(start task taskName) \
	pthread_create(&taskName##_id, &taskName##_attr, exec_##taskName, NULL);
#define START_HEADER()
#define START_TRAILER() pthread_exit(0);

/* indique le nom de la tache qui est lancee apres creation */
#define TASK_BODY(taskName) void *exec_##taskName(void *__x)

/* on peut tuer explicitement le thread */
#define KILL_TASK(taskName,termprog) TaskCount--; \
	SPY(task_end)\
	if (TaskCount==0) termprog();\
	pthread_exit(NULL);

#define START() start()

#define FOPENR(ficdesc,fic,mode,fctexit) \
	ficdesc = fopen(fic,mode); \
	if (ficdesc == NULL) {fctexit(); printf("Error : file " #fic " not found \n"); exit(1);}

#define FOPENW(ficdesc,fic,mode,fctexit) ficdesc = fopen(fic,mode)

#endif