/*-----------------------------------------*/
/*  STOOD v4  */
/*  C lexical analyser  */
/*                                      */
/*  Copyright TNI 1992-2000  */
/*  23.feb.2001  */
/*-----------------------------------------*/
Letter		[A-Za-z]
Digit		[0-9]
Extended_Digit	[0-9A-Fa-f]
Identifier	{Letter}(_?({Letter}|{Digit}))*
Exponent	(E|e)[-+]?{Digit}+
Fractional_Constant	({Digit}+".")|(({Digit}+)?"."{Digit}+)
%%

{Identifier}						identifier();
__								dot();
\"([^\"\\]|\\.)*\"				anything();		/* string literals */
\'([^\'\\]|\\.)*\'				anything();		/* character literals */
({Fractional_Constant}{Exponent}?)|({Digit}+{Exponent})		anything();		/* decimal number*/
0(x|X){Extended_Digit}+				anything();		/* hexadecimal number */
"/*"([^\/]|[^\*]\/)*"*/"			anything();		/* comments */
^#{Letter}*					anything();		/* preprocessor directives */
.|\n								anything(); 	/* anything else, including separators */

%%

#define IgnoreCase 0
char *ReservedWordList[] = {
	"void", "auto", "static", "extern", "register", "typedef", "struct", "union", "enum",
	"unsigned", 
	"if", "else", "while", "do", "for", "switch", "break", "continue", "return", "goto",
	"case", "default", "sizeof", "const",
0};




/**********************************************************
	this code is independant of the langage :
**********************************************************/

int State = 0;

identifier()
{
	static int BoL = 1;

	if(isReserved(yytext)) {
		State = 0;
		return;
	}
	if (State == 2) {
		printf(" ");
	} else {
		if (!BoL) {printf("\n"); BoL = 1;}
	}
	printf("%s", yytext);
	BoL = 0;
	State = (State == 2) ? 0 : 1;
}

dot()
{
	State = (State == 1) ? 2 : 0;
}

anything()
{
	State = 0;
}

int isReserved(word)
char *word;
{
	char **p = ReservedWordList;
	while (*p) {
		if (stringCompare(word, *p)) return 1;
		p++;
	}
	return 0;
}

int stringCompare(s1, s2)
char *s1, *s2;
{
	do {
		register char c1 = *(s1++);
		register char c2 = *(s2++);
		if (c1 == 0) return (c2 == 0);
		if (c2 == 0) return (c1 == 0);
#if IgnoreCase == 1
		if ('A' <= c1 && c1 <= 'Z' ) c1 += ('a' - 'A');
		if ('A' <= c2 && c2 <= 'Z' ) c2 += ('a' - 'A');
#endif
		if (c1 != c2) return (0);
	} while (1);
}

int yywrap()
{
	return 1;
}
	
main()
{
	yylex();
}
