%{
#define IgnoreCase 1
char *ReservedWordList[] = {
	"abort", "abs", "accept", "access", "all", "and", "array", "at", "begin", "body",
	"case", "constant", "declare", "delay", "delta", "digits", "do", "else", "elsif",
	"end", "entry", "exception", "exit", "for", "function", "generic", "goto", "if",
	"in", "is", "limited", "loop", "mod", "new", "not", "null", "of", "or", "others",
	"out", "package", "pragma", "private", "procedure", "raise", "range",
	"record", "rem", "renames", "return", "reverse", "select", "separate",
	"subtype", "task", "terminate", "then", "type", "use", "when", "while",
	"with", "xor", "si", "alors", "sinon", "fsi", "selon", "cas", "defaut", "fselon",
	"pour", "faire", "fpour", "tant_que", "ftq", "appel_de", "et", "ou", "ouex",
	"finsi", "tantque", "fintq", "ftantque", "appel",
0};

typedef enum { Code, CodeInComment, CodeInCode, Comment } CodeStateType; 
typedef enum { St0, St1, St2, St3 } StateType;

CodeStateType CodeState = Code;
StateType State = St0;

char FullName[256] = {0};
char CurOpName[256] = {0};
char CurOpParName[256] = {0};
int FlagR = 0;
int FlagW = 0;
int tabState[16];
int* pState = tabState;
int CptNbAccOuv = 0;
int CptNbCrochOuv = 0;

void identifier();
void name();
void dot();
void beginComment();
void beginCode();
void endCode();
void endComment();
void assign();
void colon();
void arrow();
void leftpar();
void rightpar(); 
void anything(); 
%}

Letter				[A-Za-z]
Digit				[0-9]
Extended_Digit			[0-9A-Fa-f]
Identifier			_*{Letter}(_?({Letter}|{Digit}))*
Exponent			E[-+]?{Digit}+

%START CODE COMMENT
%%

<COMMENT>\]					endComment();		/* Fin de commentaire */
<COMMENT>\{					beginCode();		/* Dbut de code */
<COMMENT>\n					;
<COMMENT>.					
;

<CODE>^\[			beginComment();		/*Dbut de commentaire 'dbut de ligne' 'crochet ouvrant' */
<CODE>\[			{ if (pState == tabState) beginComment(); }
<CODE>\}			endCode();		/* Fin code  */
<CODE>{Identifier}		identifier();
<CODE>((\[|\().+(\]|\)))?\.			dot();
<CODE>((\[|\().+(\]|\)))?-> 			dot();
<CODE>[ \t\n]+			;			/* separators */
<CODE>\"([^\"]|\"\")*\"		anything();		/* string literals */
<CODE>\'.\'			anything();		/* character literals */
<CODE>\'{Letter}{Identifier}	anything();		/* attribute designator */
<CODE>{Digit}+("."{Digit}+)?({Exponent})?		anything();	/* decimal number (ajout 16-12-02)*/
<CODE>0(x|X){Extended_Digit}+(u|h)*	anything();	/* hexadecimal number */
--[^\n]*			;			/* comments */

<CODE>((\[|\().+(\]|\))[ ]*)?:=			assign();		/* assignment symbol */
<CODE>: 			colon();		/* : */
<CODE>=>			arrow();		/* parameter assignment symbol */
<CODE>\(			leftpar(); 		/* ( */
<CODE>\)			rightpar(); 		/* ( */
<CODE>.				anything(); 		/* anything else */

\n				;
.				;
%%

void FLUSH() {
	printf("%s", FullName);
	if(FlagR || FlagW || CurOpParName[0]) printf("/");
	if(FlagR) printf("R");
	if(FlagW) printf("W");
	if(CurOpParName[0]) {
		printf("P(%s,%s)", CurOpName, CurOpParName);
		CurOpParName[0] = 0;
	}
	FlagR = FlagW = 0; FullName[0]=0;
	printf("\n");
}
void name() {
	switch(State) {
		case St0: strcat(FullName, yytext); State = St1; break;
		case St1: FLUSH(); strcat(FullName, yytext); break;
		case St2: strcat(FullName, "."); strcat(FullName, yytext); State = St1; break;
		default: State = St0; break;
	}
}
void inout(int m) {
	switch(State) {
		case St0: break;
		case St3: if(m=='i') FlagR = 1; if(m=='o') FlagW = 1; break;
		default: FLUSH(); if(m=='i') FlagR = 1; if(m=='o') FlagW = 1; State = St0; break;
	}
}
void identifier() {
	if(isReserved(yytext)) {
		if(stringCompare(yytext, "in")) inout('i');
		else if(stringCompare(yytext, "out")) inout('o');
		else anything();
	}
	else name();
}

void dot() {
	switch(State) {
		case St0: break;
		case St1: State = St2; break;
		default: FLUSH(); State = St0; break;
	}
}
void assign() {
	switch(State) {
		case St0: break;
		case St1: FlagW = 1; FLUSH(); State = St0; break;
		default: FLUSH(); State = St0; break;
	}
}

void colon() {
	switch(State) {
		case St0: break;
		case St1: State = St3; break;
		default: FLUSH(); State = St0; break;
	}

}
void arrow() {
	CurOpParName[0] = 0; 		/*190204 MHL (rasto 03.086)*/
	switch(State) {
		case St0: break;
		default: strcat(CurOpParName, FullName); FullName[0]=0; State = St0; break;
	}

}
void leftpar() {
	strcpy(CurOpName, FullName);
	anything();
}
void rightpar() {
	anything();
}
void anything() {
	switch(State) {
		case St0: break;
		default: FLUSH(); State = St0; break;
	}
}

void beginComment() {
	*++pState = COMMENT;
	BEGIN COMMENT;
}
void beginCode() {
	*++pState = CODE;
	BEGIN CODE;
}
void endCode() {
	BEGIN (*--pState);
}
void endComment() {
	BEGIN (*--pState);
}
	
int isReserved(char *word) {
	char **p = ReservedWordList;
	while (*p) {
		if (stringCompare(word, *p)) return 1;
		p++;
	}
	return 0;
}
int stringCompare(char *s1, char *s2) {
	do {
		register char c1 = *(s1++);
		register char c2 = *(s2++);
		if (c1 == 0) return (c2 == 0);
		if (c2 == 0) return (c1 == 0);
#		if IgnoreCase == 1
			if ('A' <= c1 && c1 <= 'Z' ) c1 += ('a' - 'A');
			if ('A' <= c2 && c2 <= 'Z' ) c2 += ('a' - 'A');
#		endif
		if (c1 != c2) return (0);
	} while (1);
}
int yywrap() {
	return 1;
}	
main() {
	beginCode();
	yylex();
	anything(); /* End of file */
	endCode();
}
