/*-------------------------------------------------------------------
	Stood v4.1           
	html document generator                

	Copyright TNI 1999-2000
	Pierre DISSAUX (pierre.dissaux@tni.fr)
	Philippe CARVALHO (philippe.carvalho@tni.fr)
	last release : 06/10/00
---------------------------------------------------------------------
HTML001 (06.08.99) ObjectType now properly inserted.
HTML002 (04.04.00) Link to files
HTML003 (06.10.00) Now supports GIF diagrams
HTML004 (06.10.00) &nbsp for space char
HTML005 (10.10.00) Setting applets default parameters
-------------------------------------------------------------------*/

go :- $init_sys, initSuffix, $getenv('PREDICATS',File), $consult(File), genHtml.

genHtml :- beginDoc, fail.
genHtml :- selectedObject(X), openHtmlFile(X), insertObj(X), fail.

beginDoc :- 
	openFrameSet, setFrames, close, 
	openIndex('_ods'), setIndex('_ods','ODS'), close,
	openIndex('_code'), setIndex('_code','Code'), close.

setFrames :-
	write('<HTML>'), nl,
	write('<HEAD>'), nl,
	write('<TITLE>'), write('STOOD v4</TITLE>'), nl,
	write('</HEAD>'), nl,
	write('<FRAMESET COLS="25%,*">'), nl,	
	write('<FRAME SRC="index_ods'), writeSuffix, write('" NAME="index">'), nl,
	write('<FRAME SRC="" NAME="contents">'), nl,
	write('</FRAMESET>'),nl,
	write('</HTML>'), nl.

setIndex(T,W) :- 
	write('<HTML>'), nl, 
	write('<BODY>'), nl,
	write('<H2>'), write(W), write(' Contents</H2>'), nl,
	index('SYSTEM_CONFIGURATION',T),nl,
	write('<HR>'), nl,
	( T='_ods' -> true; ( write('<LI><A HREF="'), write('index_ods'), writeSuffix, write('" TARGET="index">'), write('ods'), write('</A>'), nl ) ),
	( T='_code' -> true; ( write('<LI><A HREF="'), write('index_code'), writeSuffix, write('" TARGET="index">'), write('code'), write('</A>'), nl ) ),
	( T='_check' -> true; ( write('<LI><A HREF="'), write('index_check'), writeSuffix, write('" TARGET="index">'), write('check report'), write('</A>'), nl ) ),
	( T='_tests' -> true; ( write('<LI><A HREF="'), write('index_tests'), writeSuffix, write('" TARGET="index">'), write('tests'), write('</A>'), nl ) ),
	write('</BODY>'), nl,
	write('</HTML>'), nl.

index(X,T) :- beginIndex(X,T), isObject(Y,_,X), index(Y,T), fail.
index(X,T) :- endIndex.

beginIndex(X,T) :- 
	write('<LI><A HREF="'), write(X), write(T), writeSuffix, write('" TARGET="contents">'), write(X), write('</A>'), nl,
	write('<UL>'), nl.

endIndex :- write('</UL>'), nl.

insertObj(X) :- beginObj(X), docSection(N,S,P,L,W,X,T,C), insertSec(N,S,P,L,W,X,T,C), fail.
insertObj(X) :- endObj, close.

beginObj(X) :-
	write('<HTML>'), nl, 
	write('<BODY>'), nl,
	write('<HR>'), nl,
	insertTOC(X,'ObjectName',2), nl,
	write('<HR>'), nl.

endObj :-
	write('</BODY>'), nl,
	write('</HTML>'), nl.

insertTOC(X,P,L) :- write('<UL>'), nl,
	docSection(N,S,P,L,W,X,T,C), 
	( ( N='Text', notEmpty(S,L,X,C) ) ; ( N='File', notEmptyFile(S,L,W,X,F) ) ), 
	write('<LI><A HREF="#'), write(S), write('">'), write(T), write('</A>'), nl, 
	insertTOC(X,S,3), fail.
insertTOC(X,P,L) :- write('</UL>').

insertSec('Text',S,_,L,'TXT',X,T,C) :- notEmpty(S,L,X,C), insertTitle(S,L,T), insertText(C).
insertSec('Text',S,_,L,'TXTEND',X,T,C) :- insertTitle(S,L,T), insertText(C).
insertSec('Text',S,_,L,'CODE',X,T,C) :- notEmpty(S,L,X,C), insertTitle(S,L,T), insertCode(C).
insertSec('File',S,_,L,'TXT',X,T,F) :- notEmptyFile(S,L,'TXT',X,F), insertTitle(S,L,T), insertTextFile(F).
insertSec('File',S,_,L,'CODE',X,T,F) :- notEmptyFile(S,L,'CODE',X,F), insertTitle(S,L,T), insertCodeFile(F).
insertSec('File',S,_,L,'POSTSCRIPT',X,T,F) :- notEmptyFile(S,L,'POSTSCRIPT',X,F), insertTitle(S,L,T), insertDiagram(S,X,F). /*HTML003*/

insertTitle(_,_,'NIL') :- !.
insertTitle(S,5,_) :- pragma_doc_conf('RemoveLowLevelTitles','Y'), !.
insertTitle(S,6,_) :- pragma_doc_conf('RemoveLowLevelTitles','Y'), !.
insertTitle(S,L,T) :-
	write('<H'), write(L), write(' ALIGN=LEFT><A NAME="'), 
	write(S), write('">'), filtre(T), fail.
insertTitle(S,L,T) :- write('</A>'), write('</H'), write(L), write('>'), nl.

insertText('NIL') :- !.
insertText(C) :- write('<P ALIGN=LEFT><BLOCKQUOTE>'), filtre(C), fail.
insertText(C) :- write('</BLOCKQUOTE></P>'), nl.

insertCode('NIL') :- !.
insertCode(C) :- write('<P ALIGN=LEFT><BLOCKQUOTE><CODE>'), filtre(C), fail.
insertCode(C) :- write('</CODE></BLOCKQUOTE></P>'), nl.

insertTextFile(F) :- write('<P ALIGN=LEFT><BLOCKQUOTE>'), ( $see(F) -> readDocFile(F); unknownFile(F) ).
insertTextFile(F) :- write('</BLOCKQUOTE></P>'), nl.

insertCodeFile(F) :- write('<P ALIGN=LEFT><BLOCKQUOTE><CODE>'), ( $see(F) -> readDocFile(F); unknownFile(F) ).
insertCodeFile(F) :- write('</CODE></BLOCKQUOTE></P>'), nl.

insertDiagram(S,X,F) :- /*HTML003*/
	( pragma_doc_conf('RemoveApplets','N'), (S='DOC222'; S='DOC233') ) -> 
		insertApplet(S,X); insertGIF(S,X,F).
	
insertGIF(S,X,F1) :- /*HTML003*/
	write('<P>'), nl,
	write('<IMG SRC="'), shortName(F1,F2), write2(F2), write('">'), nl, 
	write('</P>'), nl.

insertApplet(S,X) :- /*HTML003*/
	write('<P>'), nl, pragma_doc_conf('Scale',Sc), W is 1600 / Sc, H is 1200 / Sc, 
	write('<APPLET CODE="HoodDiagram.class" WIDTH='), write(W), sp, write('HEIGHT='), write(H), write('>'), nl,
	insertOptions,
	write('<PARAM NAME="parent" VALUE="'), write(X), write(' 0 0 1414 1000 '), obType(X), write('">'), nl,
	cptRst, insertChildDiagram(S,X), cptGet(N), 
	write('<PARAM NAME="nbChildren" VALUE="'), write(N), write('">'), nl,
	cpt2Rst, insertProvInt(S,X,0), cpt2Get(P),
	write('<PARAM NAME="nbProvPar" VALUE="'), write(P), write('">'), nl,
	cptRst, insertImpBy(S,X), cptGet(I),
	write('<PARAM NAME="nbImp" VALUE="'), write(I), write('">'), nl,
	cptRst, insertUse(S,X), cptGet(U),
	write('<PARAM NAME="nbUse" VALUE="'), write(U), write('">'), nl,
	write('</APPLET>'), nl,
	write('</P>'), nl.

insertChildDiagram(S,X) :- isObject(Y,_,X), cptInc, cptGet(I), 
	graphicBox(Y,A,B,C,D), 
	write('<PARAM NAME="child'), write(I), 
	write('" VALUE="'), write(Y), sp, write(A), sp, write(B), sp, write(C), sp, write(D), sp, obType(Y), write('">'), nl, 
	cpt2Rst, insertProvInt(S,Y,I), cpt2Get(P),
	write('<PARAM NAME="nbProv'), write(I), write('" VALUE="'), write(P), write('">'), nl, fail.
insertChildDiagram(S,X).

insertProvInt('DOC222',X,N) :- isProvided(U,'TYPE',X), cpt2Inc, cpt2Get(I),
	write('<PARAM NAME="prov'), (N =:= 0 -> true; (write(N), write('#'))), write(I),
	write('" VALUE="'), write(U), write('">'), nl, fail.
insertProvInt('DOC233',X,N) :- isProvided(U,'OPERATION',X), cpt2Inc, cpt2Get(I),
	write('<PARAM NAME="prov'), (N =:= 0 -> true; (write(N), write('#'))), write(I),
	write('" VALUE="'), write(U), write('">'), nl, fail.
insertProvInt(_,X,_).

insertImpBy('DOC222',X) :- 
	graphicImp(X,Tx,Y,Ty,'TYPE',Lx,Ly), 
	cptInc, cptGet(I),
	write('<PARAM NAME="imp'), write(I),
	write('" VALUE="'), write(Tx), sp, write(Ty), sp,
	writeNbPoints(Lx,0), sp, writeCoord(Lx,Ly), 
	write('">'), nl, fail.
insertImpBy('DOC233',X) :- 
	graphicImp(X,Ox,Y,Oy,'OPERATION',Lx,Ly), 
	cptInc, cptGet(I),
	write('<PARAM NAME="imp'), write(I),
	write('" VALUE="'), write(Ox), sp, write(Oy), sp,
	writeNbPoints(Lx,0), sp, writeCoord(Lx,Ly), 
	write('">'), nl, fail.
insertImpBy(_,X).

insertUse('DOC222',X) :- 
	graphicUse(Y,Z,'TYPE',S,Lx,Ly,Ll),
	isObject(Y,_,X),
	cptInc, cptGet(I),
	write('<PARAM NAME="use'), write(I),
	write('" VALUE="'), write(Y), sp, write(Z), sp,
	writeNbPoints(Lx,0), sp, writeCoord(Lx,Ly), write(S), sp, 
	writeLabel(Ll), write('">'), nl, fail.
insertUse('DOC233',X) :- 
	graphicUse(Y,Z,'OPERATION',S,Lx,Ly,Ll),
	isObject(Y,_,X),
	cptInc, cptGet(I),
	write('<PARAM NAME="use'), write(I),
	write('" VALUE="'), write(Y), sp, write(Z), sp,
	writeNbPoints(Lx,0), sp, writeCoord(Lx,Ly), write(S), sp, 
	writeLabel(Ll), write('">'), nl, fail.
insertUse(_,X).

writeNbPoints([ ],Np) :- write(Np).
writeNbPoints([E|L],Np) :- Npp is Np + 1, writeNbPoints(L,Npp).

writeCoord([ ],[ ]).
writeCoord([X|Lx],[Y|Ly]) :- write(X), sp, write(Y), sp, writeCoord(Lx,Ly).

writeLabel([Label|R]) :- write(Label).
writeLabel([ ]) :- write('NONE').

insertOptions :-
	write('<PARAM NAME="options" VALUE="'), /*HTML005*/
	( pragma_doc_conf('XOffset',Xo) -> write(Xo); write(25) ), sp,
	( pragma_doc_conf('YOffset',Yo) -> write(Yo); write(25) ), sp,
	( pragma_doc_conf('Scale',Sc) -> write(Sc); write(2) ), sp,
	( pragma_doc_conf('ParentRadius',Pr) -> write(Pr); write(10) ), sp,
	( pragma_doc_conf('ChildRadius',Cr) -> write(Cr); write(5) ), sp,
	( pragma_doc_conf('ParentLabelHeight',Ph) -> write(Ph); write(20) ), sp,
	( pragma_doc_conf('ChildLabelHeight',Ch) -> write(Ch); write(20) ), sp,
	( pragma_doc_conf('PenColor',Pc) -> writeColor(Pc); write('blue') ), sp,
	( pragma_doc_conf('BgColor',Bc) -> writeColor(Bc); write('white') ), sp,
	( pragma_doc_conf('ImpColor',Ic) -> writeColor(Ic); write('lightGray') ), sp,
	( pragma_doc_conf('UseColor',Uc) -> writeColor(Uc); write('red') ), sp,
	( pragma_doc_conf('InhColor',Hc) -> writeColor(Hc); write('pink') ), sp,
	( pragma_doc_conf('AttColor',Ac) -> writeColor(Ac); write('orange') ), sp,
	write('">'), nl.

notEmpty(S,L,X,C) :- docSection(N,A,S,B,W,X,_,D), ( N='Text' ->
	notEmpty(A,B,X,D);
	notEmptyFile(A,B,W,X,D) ), !.
notEmpty(S,L,X,C) :- ( C='NONE'; C='NONE
' ) ->
	not(pragma_doc_conf('RemoveNONESections','Y'));
	( C \= '', C \= 'NIL' ), !.
notEmpty('ObjectType',_,_,_). /*HTML001*/

notEmptyFile(S,L,W,X,F) :- docSection(_,A,S,B,C,X,_,D), notEmptyFile(A,B,C,X,D), !.
notEmptyFile(S,L,W,X,F) :- not(shortName(F,'MissingFile')). /*HTML003*/

moduleLink(X,S) :- write('<A HREF="'), write(X), write(S), writeSuffix, write('">'), write(X), write('</A>').

obType(X) :- isObject(X,T,_), (
	T = 'ACTIVE' -> write('A');
	T = 'OP_CONTROL' -> write('O');
	T = 'FORMAL_PARAMETERS' -> write('F');
	T = 'VIRTUAL_NODE' -> write('V');
	T = 'ENVIRONMENT' -> write('E');
	/* else */ write('P') ).

/***************************/
/**** I / O   R U L E S ****/
/***************************/

write(T) :- $writename(T).
write2(T) :- $name(T,L), $member(E,L), $put(E), fail. /*HTML003*/
write2(T).
read(C,W,S) :- '_$builtin'(22).
nl :- $put(10).
tab :- $put(9).
sp :- $put(32).

openFrameSet :- 
	thisFile(R,_), $name(R,L1), $name('index',L2), $append(L1,L2,L3),
	suffix(S), $name(S,L4), $append(L3,L4,L), $name(F,L), write(F), nl,
	( $tell(F) -> true; unknownFile(F) ).

openIndex(T) :-
	thisFile(R,_), $name(R,L1), $name('index',L2), $append(L1,L2,L3),
	$name(T,L4), $append(L3,L4,L5),
	suffix(S), $name(S,L6), $append(L5,L6,L), $name(F,L), write(F), nl,
	( $tell(F) -> true; unknownFile(F) ). 

openHtmlFile(X) :- 
	thisFile(R,_), $name(R,L0), $name(X,L1), $append(L0,L1,L2), 
	whichDoc(D), $name(D,L3), $append(L2,[95|L3],L4), 
	suffix(S), $name(S,L5), $append(L4,L5,L), $name(F,L), write(F), nl,
	( $tell(F) -> true; unknownFile(F) ).

close :- $told.

unknownFile(F) :- write('*** Cannot open file : '), write(F), nl.

initSuffix :- ($getenv('SUFFIX',V) -> ($name(V,L), $name(S,L)); S='.html'),
        $globalset(suffix(S)).
writeSuffix :- suffix(S), write(S).

whichDoc(T) :- docType(D),
        ((D='ada'; D='c'; D='cpp') -> T='code';
        ((D='hood'; D='hood4') -> T='check'; T=D)).

shortName(F1,F2) :-
	$name(F1,L1), $reverse(L1,R1), $name(N1,R1),
	$subdelim(1,'/',N2,0,N1,_),
	$name(N2,R2), $reverse(R2,L2), $name(F2,L2).

readFile(X) :- $get0(N), ( N =\= -1 ->
	( $put(N), readFile(X) );
	( $seen ) ).

readDocFile(X) :- $get0(N), ( N =\= -1 ->
	( (N =\= 60 -> ( decode(N) ); ( include(N) ) ), readDocFile(X) );
	( $seen ) ).

filtre(T) :- $name(T,L), $member2(C,L), decode(C).

decode(C) :- (C < 0 -> A is C+256; A is C), ( map(A) -> true; $put(A) ).

map(198) :- write('&Aelig;').
map(215) :- write('&Oelig;').
map(224) :- write('&agrave;').
map(226) :- write('&acirc;').
map(228) :- write('&auml;').
map(229) :- write('&aring;').
map(230) :- write('&aelig;').
map(231) :- write('&ccedil;').
map(232) :- write('&egrave;').
map(233) :- write('&eacute;').
map(234) :- write('&ecirc;').
map(235) :- write('&euml;').
map(238) :- write('&icirc;').
map(239) :- write('&iuml;').
map(244) :- write('&ocirc;').
map(246) :- write('&ouml;').
map(247) :- write('&oelig;').
map(249) :- write('&ugrave;').
map(251) :- write('&ucirc;').
map(252) :- write('&uuml;').

map(9) :- write('&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;').
map(10) :- write('<BR>'), nl.
map(13) :- write('<BR>'), nl.
map(32) :- write('&nbsp;'). /*HTML004*/
map(38) :- write('&amp;').
map(60) :- write('&lt;').
map(62) :- write('&gt;').

/* HTML0002 */
include(A) :- $get0(I), (I =\= 73 -> ( decode(A), decode(I) ); 
		( $get0(N), (N =\= 78 -> ( decode(A), decode(I), decode(N) );
			( $get0(C), (C =\= 67 -> ( decode(A), decode(I), decode(N), decode(C) );
				( $get0(L), (L =\= 76 -> ( decode(A), decode(I), decode(N), decode(C), decode(L) );
				( lire(34,[],_,0), lire(34,[],L1,0), lire(62,[],_,0),
				convertSlash(L1,[],L2),
				write('<a href="file://'),
				$name(F,L2), write(F),
				write('">'), write(F),
				write('</a>') ) ) ) ) ) ) ) ).

/* HTML0002 */
lire(A,L,B,D) :- $get0(C), (C =\= -1 -> 
	(C =\= A -> ( lire(A,[C|L],B,D) );
	( D =\= 0 -> ( $reverse([C|L],B) ); ( $reverse(L,B) ) ) ) ).
	
 
/* HTML0002 */
convertSlash([],L2,L) :- $reverse(L2,L), !.
convertSlash([C|L1],L2,L) :- (C =\= 92 -> ( convertSlash(L1,[C|L2],L) ); ( convertSlash(L1,[47|L2],L) ) ).
						
$member2(X,L) :- nonvar(L), L = [Y|Z], (X = Y ; $member2(X,Z)).

cptRst :- $globalset(cpt(0)).
cptInc :- cpt(I), J is I + 1, $globalset(cpt(J)).
cptGet(I) :- cpt(I).
cpt2Rst :- $globalset(cpt2(0)).
cpt2Inc :- cpt2(I), J is I + 1, $globalset(cpt2(J)).
cpt2Get(I) :- cpt2(I).

writeColor('black') :- write('0').
writeColor('white') :- write('1').
writeColor('lightGray') :- write('2').
writeColor('gray') :- write('3').
writeColor('darkGray') :- write('4').
writeColor('red') :- write('5').
writeColor('pink') :- write('6').
writeColor('orange') :- write('7').
writeColor('yellow') :- write('8').
writeColor('green') :- write('9').
writeColor('magenta') :- write('10').
writeColor('blue') :- write('11').
writeColor('cyan') :- write('12').
 
