/*-------------------------------------------------------------------
	Stood v4           
	MIF document generator                                                                                                        
	Copyright TNI 1992,..,1999
	Pierre DISSAUX (pierre.dissaux@tni.fr)
	Philippe Carvalho (philippe.carvalho@tni.fr)
	last release : 04/00
---------------------------------------------------------------------
MIF001 (06.08.99) ObjectType now properly inserted.
MIF002 (04/04/00) Link to files
-------------------------------------------------------------------*/

go :- 
	$init_sys, $getenv('PREDICATS',F), $consult(F),
	storeTagName,
	openMIFFile(F), genMif, closeMIFFile.

genMif :- beginAFrames, fail.
genMif :- selectedObject(X), insertAFrames(X), fail.
genMif :- beginDoc, fail.
genMif :- selectedObject(X), insertObj(X), fail.
genMif :- endDoc.

beginAFrames :-
	resetFrameID,
	write('<AFrames'),nl.

beginDoc :-
	write('> # end of AFrames'), nl,
	resetFrameID,
	write('<TextFlow'), nl,
	write(' <TFTag `A''>'),nl,
	write(' <TFPostScript No>'),nl,
	write(' <TFAutoConnect Yes >'),nl.

endDoc :- 
	write('> # end of TextFlow'), nl,
	$told.

insertAFrames(X) :-	docSection('File',_,_,_,'POSTSCRIPT',X,_,F), insertFrame(F).

insertObj(X) :- docSection(N,S,P,L,W,X,T,C), insertSec(N,S,P,L,W,X,T,C).

insertSec('Text',S,_,L,'TXT',X,T,C) :-  notEmpty(S,L,X,C), beginSec(S,L), insertTitle(S,L,T), insertText(C), endSec(S,L).
insertSec('Text',S,_,L,'TXTEND',X,T,C) :-  beginSec(S,0), insertTitle(S,L,T), endSec(S,0).
insertSec('Text',S,_,L,'CODE',X,T,C) :-  notEmpty(S,L,X,C), beginSec(S,L), insertTitle(S,L,T), insertCode(C), endSec(S,L).
insertSec('File',S,_,L,'TXT',X,T,F) :-  notEmptyFile(S,L,'TXT',X,F), beginSec(S,L), insertTitle(S,L,T), insertTextFile(F), endSec(S,L).
insertSec('File',S,_,L,'CODE',X,T,F) :-  notEmptyFile(S,L,'CODE',X,F), beginSec(S,L), insertTitle(S,L,T), insertCodeFile(F), endSec(S,L).
insertSec('File',S,_,L,'POSTSCRIPT',X,T,F) :- notEmptyFile(S,L,'POSTSCRIPT',X,F), beginSec(S,L), insertTitle(S,L,T), insertEPSFile(F), endSec(S,L).

beginSec(S,L) :- 
	write(' <Para'), nl,
	write('  <PgfTag `'), writeTag(S,L), write('''>'), nl.

endSec(S,L) :- 
	write(' > # end of '), writeTag(S,L), nl.

insertTitle(_,_,'NIL') :- !.
insertTitle(_,5,_) :- pragma_doc_conf('RemoveLowLevelTitles','Y'), !.
insertTitle(_,6,_) :- pragma_doc_conf('RemoveLowLevelTitles','Y'), !.
insertTitle(S,L,T) :- 
	write('  <ParaLine'), nl,
	write('   <String `'), writeString(T), write('''>'), nl,
	write('   <Char HardReturn>'),nl,
	write('  >'), nl.

insertText('NIL') :- !.
insertText(C) :-
	write('  <ParaLine'), nl,
	write('   textFont'), nl,
	write('   <String `'), writeString(C), write('''>'), nl,
	write('   <Char HardReturn>'), nl,
	write('  >'), nl.

insertCode('NIL') :- !.
insertCode(C) :-
	write('  <ParaLine'), nl,
	write('   codeFont'), nl,
	write('   <String `'), writeString(C), write('''>'), nl,
	write('   <Char HardReturn>'), nl,
	write('  >'), nl.

insertTextFile(F) :- 
	write('  <ParaLine'), nl,
	write('   textFont'), nl,
	write('   <String `'),	
	( $see(F) -> readDocFile(F); unknownFile(F) ),
	write('''>'), nl, 
	write('  >'), nl.

insertCodeFile(F) :- 
	write('  <ParaLine'), nl,
	write('   codeFont'), nl,
	write('   <String `'),	
	( $see(F) -> readDocFile(F); unknownFile(F) ),
	write('''>'), nl, 
	write('  >'), nl.

insertEPSFile(F) :- 
	write('  <ParaLine'), nl,
	write('   <AFrame '), writeFrameID, write('>'), nl,
	write('  >'), nl.

insertFrame(F) :- 
	write(' <Frame '), nl,
	write('  <ID '), writeFrameID, write('>'), nl,
	write('  <Pen 15>'), nl,
	write('  <Fill 15>'), nl,
	write('  <PenWidth  1.0 pt>'), nl,
	write('  <Separation 0>'), nl,
	$see(F), analyse(L1,B1,R1,T1), $seen,
	int(L1,L2), int(B1,B2), int(R1,R2), int(T1,T2),
	W is R2 - L2, H is T2 - B2, We is W + 12, He is H + 12,
	write('  <BRect  0 pt 0 pt '), write(We), write(' pt '), write(He), write(' pt>'), nl,
	write('  <FrameType Below>'), nl,
	write('  <Float No>'), nl,
	write('  <NSOffset  0.0 cm>'), nl,
	write('  <BLOffset  0.0 cm>'), nl,
	write('  <AnchorAlign Center>'), nl,
	write('  <Cropped Yes>'), nl,
	write('  <ImportObject '), nl,
	write('=EPSI'),nl,
	write('&%v'),nl,
	write('&'), ( $see(F) -> readPSFile(F); unknownEPSFile(F) ), 
	write('\r'), nl,
	write('=EndInset'), nl,
	write('  <ImportObFile `2.0 internal inset''>'), nl,
	write('   <BRect 6 pt 6 pt '), write(W), write(' pt '), write(H), write(' pt>'), nl,
	write('   <BitMapDpi 0>'), nl,
	write('   <Angle 0>'), nl,
	write('   <FlipLR No>'), nl,
	write('  > # end of ImportObject'), nl,
	write(' > # end of Frame'), nl.

notEmpty(S,L,X,C) :- docSection(N,A,S,B,W,X,_,D), ( N='Text' ->
	notEmpty(A,B,X,D);
	notEmptyFile(A,B,W,X,D) ), !.
notEmpty(S,L,X,C) :- ( C='NONE'; C='NONE
' ) ->
	not(pragma_doc_conf('RemoveNONESections','Y'));
	( C \= '', C \= 'NIL' ), !.
notEmpty('ObjectType',_,_,_). /* MIF001*/

notEmptyFile(S,L,W,X,F) :- docSection(_,A,S,B,C,X,_,D), notEmptyFile(A,B,C,X,D), !.
notEmptyFile(S,L,W,X,F) :- not(shortName(F,'MissingFile')).


/***************************/
/**** I / O   R U L E S ****/
/***************************/

write(T) :- $writename(T).
read(C,W,S) :- '_$builtin'(22).
nl :- $put(10).
tab :- $put(9).
sp :- $put(32).

storeTagName :- 
	$getenv('NOMTAG',T1), $name(T1,L), $name(T2,L), $globalset(tagName(T2)).

int(A,B) :- $name(A,[32|L1]), $reverse(L1,L2), convert(B,L2).

convert(0,[]).
convert(V,[D|L]) :- N is D - 48, convert(W,L), V is ( 10 * W) + N.

writeTag(S,N) :- pragma_doc_conf('OneTagPerSection','Y') -> 
	write(S);
	( tagName(T), write(T), (N > 6 -> M=6 ; M is N), write(M) ).

openMIFFile(F) :- 
	$tell(F) -> true; ( write('*** Cannot open output file : '), write(F) ).

closeMIFFile :- $told.

resetFrameID :- $globalset(id(0)).
writeFrameID :- id(I), J is I + 1, write(J), $globalset(id(J)).

unknownFile(F) :- write('*** Cannot open file : '), write(F).

unknownEPSFile(F) :- write('('), unknownFile(F), write(') S').

shortName(F1,F2) :-
	$name(F1,L1), $reverse(L1,R1), $name(N1,R1),
	$subdelim(1,'/',N2,0,N1,_),
	$name(N2,R2), $reverse(R2,L2), $name(F2,L2).

analyse(L,B,R,T) :- find(37) ->
	( read(32,A,_), ( A=' BoundingBox'  ->
		( read(_,_,_), read(32,L,_), read(32,B,_), read(32,R,_), read(32,T,_) );
		( analyse(L,B,R,T) ) ) );
	( write('% *** error : not an EPS file ! ') ).

find(C) :- $get0(N), N \= -1, ( N ?= C ->
	( $get0(M), M \= -1, ( M ?= C -> true; find(C) ) );
	( find(C) ) ).

readFile(F) :- $get0(N), ( N =\= -1 ->
	( $put(N), readFile(X) );
	( $seen ) ).

readDocFile(X) :- $get0(N), ( N =\= -1 ->
	( (N =\= 60 -> 
		( decode(N) ); 
		( include(N) ) ), readDocFile(X) );
	( $seen ) ).

readPSFile(F) :- $get0(N), ( N =\= -1 ->
	( encodePS(N), readPSFile(F) );
	( $seen ) ).
	

writeString(T) :- T\='', $name(T,L), $member2(C,L), decode(C), fail.
writeString(T).

decode(C) :- ( C < 0 -> A is C+256; A is C ), ( map(A) -> true; $put(A) ).

map(198) :- write('\xae ').
map(215).
map(224) :- write('\x88 ').
map(226) :- write('\x89 ').
map(228) :- write('\x8a ').
map(229) :- write('\xe5 ').
map(230) :- write('\xbe ').
map(231) :- write('\x8d ').
map(232) :- write('\x8f ').
map(233) :- write('\x8e ').
map(234) :- write('\x90 ').
map(235) :- write('\x91 ').
map(238) :- write('\x94 ').
map(239) :- write('\x95 ').
map(244) :- write('\x99 ').
map(246) :- write('\x9a ').
map(247) :- write('\xcf ').
map(249) :- write('\x9d ').
map(251) :- write('\x9e ').
map(252) :- write('\x9f ').
map(9) :- write('\t').
map(13) :- map(10).
map(10) :-
	write('''>'), nl, 
	write('   <Char HardReturn>'),nl,
	write('  >'),nl,
	write('  <ParaLine'),nl,
	write('   <String `').
map(39) :- write('\q').  /* ' */
map(96) :- write('\Q').  /* ` */
map(62) :- write('\>').  /* > */
map(92) :- write('\\').

/* MIF0002 */
include(A) :- $get0(I), (I =\= 73 -> ( decode(A), decode(I) ); 
		( $get0(N), (N =\= 78 -> ( decode(A), decode(I), decode(N) );
			( $get0(C), (C =\= 67 -> ( decode(A), decode(I), decode(N), decode(C) );
				( $get0(L), (L =\= 76 -> ( decode(A), decode(I), decode(N), decode(C), decode(L) );
				( lire(34,[],_,0), lire(34,[],L1,0), lire(62,[],_,0),
				convertSlash(L1,[],L2),
				write(''' include ('),
				$name(F,L2), write(F),
				write(')><String `') ) ) ) ) ) ) ) ).

/* MIF0002 */
lire(A,L,B,D) :- $get0(C), (C =\= -1 -> 
	(C =\= A -> ( lire(A,[C|L],B,D) );
	( D =\= 0 -> ( $reverse([C|L],B) ); ( $reverse(L,B) ) ) ) ).
	
 
/* MIF0002 */
convertSlash([],L2,L) :- $reverse(L2,L), !.
convertSlash([C|L1],L2,L) :- (C =\= 92 -> ( convertSlash(L1,[C|L2],L) ); ( convertSlash(L1,[47|L2],L) ) ).

encodePS(C) :- ( C < 0 -> A is C+256; A is C ), ( mapPS(A) -> true; $put(A) ).

mapPS(10) :- mapPS(13).
mapPS(13) :- write('\r'), nl, write('&').
mapPS(92) :- write('\\').

$member2(X,L) :- nonvar(L), L = [Y|Z], (X = Y ; $member2(X,Z)).
