/*-------------------------------------------------------------------
	Stood v4           
	PostScript document generator                                                                                                        
	Copyright TNI 1992,..,1999
	Pierre DISSAUX (pierre.dissaux@tni.fr)
	last release : 08/99
---------------------------------------------------------------------
PS001 (06.08.99) ObjectType now properly inserted.
PS002 (06.08.99) EndObj now properly inserted.
PS003 (06.10.00) Larger left margin for diagrams
-------------------------------------------------------------------*/

go :- 
	$init_sys, $getenv('PREDICATS',File), $consult(File), 
	openPSFile(File), genPS, closePSFile.

genPS :- beginDoc, fail.
genPS :- selectedObject(X), insertObj(X), fail.
genPS :- endDoc.

beginDoc :- 
	docConf('TopLeft','Text'), 
	docConf('TopCenter','Text'), 
	docConf('TopRight','Text'),
	docConf('BottomLeft','Text'), 
	docConf('BottomCenter','Text'), 
	docConf('BottomRight','Text'),
	docConf('pageNumber','Number').

docConf(W,'Text') :- 
	write('/'), write(W), write(' ('), pragma_doc_conf(W,T), 
	write(T), write(') def'), nl.
docConf(W,'Number') :- 
	write('/'), write(W), sp, ( pragma_doc_conf(W,T) -> 
	write(T); write('1') ), write(' def'), nl.

endDoc :- write('1 INDENT'), nl,
	write('beginOfPage'), nl,
	write('tableOfContents'), nl,
	write('beginOfPage').

insertObj(X) :- beginObj(X).
insertObj(X) :- docSection(N,S,P,L,W,X,T,C), insertSec(N,S,P,L,W,X,T,C).
insertObj(X) :- endObj(X).

beginObj(X) :- write('/pageTitle ('), write(X), write(') def'), nl,
	write('beginOfPage'), nl.

endObj(X) :- write('() TOC'), nl.

insertSec('Text',S,_,L,'TXT',X,T,C) :-  notEmpty(S,L,X,C), insertText(S,L,T,C).
insertSec('Text',S,_,L,'TXTEND',X,T,C) :-  insertTextEnd(S,L,T,C). /*PS002*/
insertSec('Text',S,_,L,'CODE',X,T,C) :-  notEmpty(S,L,X,C), insertCode(S,L,T,C).
insertSec('File',S,_,L,'TXT',X,T,F) :-  notEmptyFile(S,L,'TXT',X,F), insertTextFile(S,L,T,F).
insertSec('File',S,_,L,'CODE',X,T,F) :-  notEmptyFile(S,L,'CODE',X,F), insertCodeFile(S,L,T,F).
insertSec('File',S,_,L,'POSTSCRIPT',X,T,F) :- notEmptyFile(S,L,'POSTSCRIPT',X,F), insertEPSFile(S,L,T,F). 

insertTitle(_,_,'NIL') :- !.
insertTitle(_,5,_) :- pragma_doc_conf('RemoveLowLevelTitles','Y'), !.
insertTitle(_,6,_) :- pragma_doc_conf('RemoveLowLevelTitles','Y'), !.
insertTitle(S,L,T) :- 
	write('('), writePSString(T), write(') '),
	(L < 3 -> write('TOC '); true), write('S'), nl.

insertText(S,L,T,C) :-  
	write('2 F'), nl,
	write(L), write(' INDENT'), nl,
	write(L), write(' ROOM'), nl,
	write('R'), nl, 
	insertTitle(S,L,T),
	C \= 'NIL', write('1 F'), nl,
	write('('), writePSString(C), write(') S'), nl.

insertCode(S,L,T,C) :- 
	write('2 F'), nl,
	write(L), write(' INDENT'), nl,
	write(L), write(' ROOM'), nl,
	write('R'), nl, 
	insertTitle(S,L,T),
	C \= 'NIL', write('3 F'), nl,
	write('('), writePSString(C), write(') S'), nl.

insertTextEnd(S,L,T,C) :-  
	write('2 F'), nl,
	write(L), write(' INDENT'), nl,
	write('R'), nl, 
	insertTitle(S,L,T),
	C \= 'NIL', write('1 F'), nl,
	write('('), writePSString(C), write(') S'), nl.

insertTextFile(S,L,T,F) :- 
	write('2 F'), nl,
	write(L), write(' INDENT'), nl,
	write(L), write(' ROOM'), nl,
	write('R'), nl, 
	insertTitle(S,L,T),
	write('1 F'), nl, write('('), 
	( $see(F) -> readDocFile(F); unknownFile(F) ),
	write(') S'), nl.

insertCodeFile(S,L,T,F) :- 
	write('2 F'), nl,
	write(L), write(' INDENT'), nl,
	write(L), write(' ROOM'), nl,
	write('R'), nl, 
	insertTitle(S,L,T),
	write('3 F'), nl, write('('), 
	( $see(F) -> readDocFile(F); unknownFile(F) ),
	write(') S'), nl.

insertEPSFile(S,L,T,F) :- 
	write('2 F'), nl,
	write(L), write(' INDENT'), nl,
	write('R'), nl, 
	( pragma_doc_conf('FullPageDiagrams','Y'), notStd(S) ) ->
		fullSizeEPSFile(S,L,T,F);
		halfSizeEPSFile(S,L,T,F).
		
notStd(S) :- S \= 'DOC144', S \= 'DOC242', S \= 'Std'.

halfSizeEPSFile(S,L,T,F) :- /*PS003*/
	$see(F), analyse(B,U), $seen,
	write(U), write(B), write(' sub RM'), nl,
	insertTitle(S,L,T),
	write('currenty'), nl,
	write('BeginEpsFile'), nl,
	write('100 exch'), write(U), write(' sub translate'), nl,
	( $see(F) -> readFile(F); unknownEPSFile(F) ), nl,
	write('EndEpsFile'), nl,
	write('100'), write(B), write(U), write(' sub rmoveto'), nl,
	write('( ) S'), nl.

fullSizeEPSFile(S,L,T,F) :- 
	write('625 RM'), nl,
	insertTitle(S,L,T),
	write('currenty'), nl,
	write('BeginEpsFile'), nl,
	write('90 rotate'), nl,
	write('1.4 dup scale'), nl,
	write('0 100 translate'), nl,
	( $see(F) -> readFile(F); unknownEPSFile(F) ), nl, 
	write('EndEpsFile'), nl,
	write('0 -625 rmoveto'), nl.

notEmpty(S,L,X,C) :- docSection(N,A,S,B,W,X,_,D), ( N='Text' ->
	notEmpty(A,B,X,D);
	notEmptyFile(A,B,W,X,D) ), !.
notEmpty(S,L,X,C) :- ( C='NONE'; C='NONE
' ) ->
	not(pragma_doc_conf('RemoveNONESections','Y'));
	( C \= '', C \= 'NIL' ), !.
notEmpty('ObjectType',_,_,_). /*PS001*/

notEmptyFile(S,L,W,X,F) :- docSection(_,A,S,B,C,X,_,D), notEmptyFile(A,B,C,X,D), !.
notEmptyFile(S,L,W,X,F) :- not(shortName(F,'MissingFile')).

/***************************/
/**** I / O   R U L E S ****/
/***************************/

write(T) :- $writename(T).
read(C,W,S) :- '_$builtin'(22).
nl :- $put(10).
tab :- $put(9).
sp :- $put(32).

openPSFile(F) :- $getenv('PSFILE',PsFile), ($tell(PsFile) ->
	true; (write('*** Cannot open PostScript file : '), write(PsFile))).

closePSFile :- $told.

unknownFile(F) :- write('*** Cannot open file : '), write(F).
unknownEPSFile(F) :- write('('), unknownFile(F), write(') S').

shortName(F1,F2) :-
	$name(F1,L1), $reverse(L1,R1), $name(N1,R1),
	$subdelim(1,'/',N2,0,N1,_),
	$name(N2,R2), $reverse(R2,L2), $name(F2,L2).

readFile(X) :- $get0(N), ( N =\= -1 ->
	( $put(N), readFile(X) );
	( $seen ) ).

readDocFile(X) :- $get0(N), ( N =\= -1 ->
	( decode(N), readDocFile(X) );
	( $seen ) ).

analyse(B,T) :- find(37) ->
	( read(32,A,_), ( A=' BoundingBox'  ->
	 ( read(_,_,_), read(_,_,_), read(32,B,_), read(_,_,_), read(32,T,_) );
	 ( analyse(B,T) ) ) );
	( write('% *** error : not an EPS file ! ') ).

find(C) :- $get0(N), N \= -1, ( N ?= C ->
	( $get0(M), M \= -1, ( M ?= C -> true; find(C) ) );
	( find(C) ) ).

writePSString(T) :- T\='', $name(T,L), $member2(C,L), decode(C), fail.
writePSString(T).

decode(C) :- (C < 0 -> A is C+256; A is C), ( map(A) -> true; $put(A) ).

map(198) :- write('\306').
map(215) :- write('\236').
map(224) :- write('\340').
map(226) :- write('\342').
map(228) :- write('\344').
map(229) :- write('\345').
map(230) :- write('\346').
map(231) :- write('\347').
map(232) :- write('\350').
map(233) :- write('\351').
map(234) :- write('\352').
map(235) :- write('\353').
map(238) :- write('\356').
map(239) :- write('\357').
map(244) :- write('\364').
map(246) :- write('\366').
map(247) :- write('\237').
map(249) :- write('\371').
map(251) :- write('\373').
map(252) :- write('\374').

map(10) :- write(') S'), nl, write('(').
map(40) :- write('\(').
map(41) :- write('\)').
map(92) :- write('\\').


$member2(X,L) :- nonvar(L), L = [Y|Z], (X = Y ; $member2(X,Z)).
