/*-----------------------------------------------
   AADL2HOOD v1.0
   AADL code reverse engineering
   SIF 4 generator

   Copyright TNI Europe & Ellidiss Technologies 2004-2005
   Pierre DISSAUX (pierre.dissaux@tni-world.com)
   09.Feb.2005
------------------------------------------------
last changes:
------------------------------------------------*/
 
/*----*/
/* go */
/*----*/

go :- $init_sys,
	$getenv('PREDICATS',In),
	( $consult(In) -> true; ( write('Cannot open input file. (error 1)'), nl ) ),
	$getenv('SIFFILE',Sif), ( Sif \= '' -> 
		( Out = Sif ); 
		( packageSpec(Root,_) -> 
			( splitName(Root,F,_),  addSuffix(F,'.sif',Out) ); 
			( write('Cannot find root module. (error 3)'), Out='user', nl ) ) ), 
	( $tell(Out) -> true; ( write('Cannot open output file. (error 2)'), nl ) ), 
	$getenv('EXTERNAL_FACTS',Ext), 
	( $exists(Ext) -> $consult(Ext); true ), 
	$getenv('ROOT_UNIT',R), $name(R,L), $name(M,L), 
	genSif(M).

genSif(M) :- 
	sifHeader, 
	createSystem(M), fail.
genSif(M) :-
	isRoot(M,E,I,K), createModule('NIL',E,I,K,'NIL','NIL','NIL'), fail.
genSif(M) :-
	isModule(X,T,I,K,P,R,J), not(isRoot(M,T,I,K)), createModule(X,T,I,K,P,R,J), fail.	
genSif(M) :- 
	$told.
	
/*-----------*/
/* sifHeader */
/*-----------*/

sifHeader :- 
	write('SIF_VERSION --|4.0|--;'), nl,
	write('TOOL_NAME --|AADL2HOOD|--;'), nl,
	write('TOOL_VERSION --|1.00|--;'), nl,
	write('TARGET_LANGUAGE --|Ada|--;'), nl.
 
/*--------------*/
/* hood mapping */
/*--------------*/

isRoot(M,P,'NIL','PACKAGE') :- ( isPackage(P,'PUBLIC'); isPackage(P,'BOTH') ), hasDataType(P), isChildPackage('NIL',P).
isRoot(M,R,I,K) :- isComponent('NIL',R,I,K), not(isComponentInstance(_,R,I,K)), K \= 'SUBPROGRAM', findChildren('NIL',R,I).

findChildren(X,T,I) :- isSubcomponent(_,T,I,S,K,Z), isComponent(_,R,J,K), K \= 'SUBPROGRAM', 
	splitReference(Z,A,B,C), memezTra(B,R), memezTra(C,J), 
	( isChildModule(S,R,J,K,X,T,I) -> true; $assert(isChildModule(S,R,J,K,X,T,I)) ), findChildren(S,R,J), fail.
findChildren(X,T,I).

isModule('NIL',C,'NIL','PACKAGE',P,'NIL','NIL') :- ( isPackage(C,'PUBLIC'); isPackage(C,'BOTH') ), hasDataType(C), isChildPackage(P,C), P \='NIL'.
isModule('NIL',T,I,K,P,'NIL','NIL') :- isComponent(P,T,I,K), P \= 'NIL', K \= 'SUBPROGRAM', T \= 'void'.
isModule(X,T,I,K,P,R,J) :- isChildModule(X,T,I,K,P,R,J).

isChild('NIL',P,'NIL',N) :- isComponent(P,T,I,K), P \= 'NIL', K \= 'SUBPROGRAM', T \= 'void', subPackName(T,I,P,N).
isChild('NIL',P,'NIL',Q) :- ( isPackage(Q,'PUBLIC'); isPackage(Q,'BOTH') ), hasDataType(Q), isChildPackage(P,Q).
isChild(X,T,I,Y) :- isSubcomponent(_,T,I,S,K,Z), isComponent(_,R,J,K), K \= 'SUBPROGRAM', 
	splitReference(Z,A,B,C), memezTra(B,R), memezTra(C,J), 
	subCompName(X,T,I,S,Y). 

isTerm(X,T,I,_) :- not(isChild(X,T,I,_)).
isTerm(_,_,_,'DATA').
isTerm(_,_,_,'SUBPROGRAM').

isParent(X,T,I,K) :- not(isTerm(X,T,I,K)).

moduleKind(_,_,_,'SYSTEM','ACTIVE') :- !.
moduleKind(_,_,_,'PROCESS','ACTIVE') :- !.
moduleKind(_,_,_,'THREAD GROUP','ACTIVE') :- !.
moduleKind(_,T,I,'THREAD','CYCLIC') :- isProperty(_,T,I,_,_,P,V), memezTra(P,'Dispatch_Protocol'), memezTra(V,'Periodic'), !.
moduleKind(_,T,I,'THREAD','SPORADIC') :- isProperty(_,T,I,_,_,P,V), memezTra(P,'Dispatch_Protocol'), memezTra(V,'Sporadic'), !.
moduleKind(_,_,_,'THREAD','ACTIVE').

hasRealTimeAttributes(X,T,I) :- isProperty(_,T,I,_,_,P,_), 
	( memezTra(P,'Period'); memezTra(P,'Deadline') ).

hasOpcs(X,T,I,_,O) :- isProvOp2(X,T,I,T,O), isWcet(X,T,I,O,_).
hasOpcs(X,T,I,_,O) :- isProvOp2(X,T,I,T,O), isDataSubprogram(P,T1,O,C), memezTra(T1,T), 
	splitReference(C,P,R,J), isCall(P,R,J,_,_).
hasOpcs(X,T,I,_,O) :- isProvOp2(X,T,I,T,O), isServerSubprogram(P,T1,O,C), memezTra(T1,T), 
	splitReference(C,P,R,J), isCall(P,R,J,_,_).
hasOpcs(X,T,I,'THREAD','thread').
hasOpcs(X,T,I,_,O) :- isPort(_,T1,O,_), isPortType(_,T1,O,K,_), memezTra(T1,T), 
	( K = 'data'; K = 'event data' ).

isWcet(X,T,I,O,W) :- isProperty(_,T,I,O,_,P,V), memezTra(P,'Compute_Deadline'), toNum(V,W), !.
isWcet(X,T,I,O,W) :- isServerSubprogram(_,T1,O,C), memezTra(T1,T), splitReference(C,Y,S,J), 
	J \= 'NIL', isProperty(Y,S,J,_,_,P,V), memezTra(P,'Compute_Deadline'), toNum(V,W), !.
isWcet(X,T,I,O,W) :- isServerSubprogram(_,T1,O,C), memezTra(T1,T), splitReference(C,Y,S,'NIL'), 
	isProperty(Y,S,'NIL',_,_,P,V), memezTra(P,'Compute_Deadline'), toNum(V,W), !.
	
hasDataType(P) :- ( isComponentType(P,_,'DATA'); isComponentImplementation(P,_,_,'DATA') ), !.

isComponent(P,T,I,K) :- isComponentImplementation(P,T,I,K).
isComponent(P,T,'NIL',K) :- isComponentType(P,T,K), not(hasComponentImplementation(P,T)).

isComponentInstance(S,T,I,K) :- isSubcomponent(_,_,_,S,K,Y), memezTra(Y,T,'.',I).
isComponentInstance(S,T,'NIL',K) :- isSubcomponent(_,_,_,S,K,Y), memezTra(Y,T).

hasComponentImplementation(P,T1) :- isComponentImplementation(P,T2,I,_), memezTra(T1,T2).

isImplementation(I) :- not(dummyImp(I)).
dummyImp(I) :- I = 'NIL'; memezTra(I,'OTHERS'); memezTra(I,'DEFAULT').

/*----- names -----*/

moduleName('NIL',P,'NIL','PACKAGE',_,_,_,N) :- packageName(P,N).
moduleName('NIL',T,I,K,P,_,_,N) :- K \= 'PACKAGE', subPackName(T,I,P,N).
moduleName(X,T,I,K,P,R,J,N) :- X \= 'NIL', K \= 'PACKAGE', subCompName(P,R,J,X,N). 

subPackName(T,I,P,N) :-  ( ( isComponentType(Q,T,_), Q \= P ) -> 
	( leafPackage(P,L), concat(L,'_',T,N1) ); N1 = T ),
	( isImplementation(I) -> concat(N1,'_',I,N); N = N1 ).

subCompName('NIL',_,_,X,X).
subCompName(X,T,I,S,Y) :- X \= 'NIL', ( isImplementation(I) -> concat(T,'.',I,U); U = T ), 
	( ( ( isSubcomponent(_,_,_,S1,_,U), isSubcomponent(_,_,_,S2,_,U), S1 \= S2 ); 
	  ( isSubcomponent(_,V,J,S,_,_), ( V \= T; J \= I ) ) ) -> 
	  concat(X,'_',S,Y); Y = S ), !.

/*----- types -----*/

isProvType('NIL',T,I,T,I,P) :- isComponent(P,T,I,'DATA').
isProvType('NIL',P,'NIL',T,I,_) :- isComponent(P,T,I,'DATA'), T \= 'void'.

isTypeComment(X,U,'NIL',D) :- isComponentType(_,U1,'DATA'), isCompTypeComment(_,U1,D), memezTra(U1,U).
isTypeComment(X,U,I,D) :- isComponentImplementation(_,U1,I,'DATA'), isCompImpComment(_,U1,I,D), memezTra(U1,U).
isTypeComment(X,U,I,D) :- isComponentImplementation(_,U1,I,'DATA'), isSubcompComment(_,U1,I,_,D), memezTra(U1,U).

isAttribute(X,U,I,A,T) :- isSubcomponent(_,U,I,A,'DATA',T).

isSuperClass(X,T,I,S) :- isComponentImplementation(_,T1,I,'DATA'), isCompImpExtension(_,T1,I,A,B), 
	memezTra(T1,T), concat(A,'_',B,S).

isTypeImpBy('NIL',P,'NIL',T,I,T) :- isProvType('NIL',P,'NIL',T,I,P). 

/*----- constants -----*/

isProvConst(X,T,I,C,V,'NIL','NIL') :- isProperty(P,T1,I,X,_,C,V), memezTra(T1,T), memezTra(C,'Source_Data_Size').
isProvConst('NIL',P,'NIL',C,V,T,I) :- isProperty(P,T,I,_,_,C,V), memezTra(C,'Source_Data_Size').

printName(T,I,C) :- T = 'NIL' -> printName(C);
	 ( ( isImplementation(I) -> concat(T,'_',I,'_',C,N); concat(T,'_',C,N) ), printName(N) ).
	 
isConstComment(X,T,I,C,D) :- fail.

isConstImpBy('NIL',P,'NIL',T,I,C) :- isProvConst('NIL',P,'NIL',C,_,T,I). 

/*----- data -----*/

isIntData(X,T,I,K,D,R) :- isPort(_,T1,D,R), isPortType(_,T1,D,K,_), memezTra(T1,T), 
	( K = 'data'; K = 'event data' ).
	
/*----- exceptions -----*/

isProvEx(X,T,I,P,O,E) :- isDataSubprogram(_,T1,_,R), memezTra(T1,T), 
	splitReference(R,P,O,_), isPort(P,O,E,'OUT'), isPortType(P,O,E,'event',_).
	
isExComment(X,T,I,P,O,E,D) :- isPortComment(P,O,E,D).

/*----- operation sets -----*/

isProvOpSet1('NIL',P,'NIL',P,G,S) :- isComponentType(P,T,_), isPortGroup(P,T,S,R),
	splitReference(R,P1,G1,_), isPortGroupType(P,G), memezTra(P1,P), memezTra(G1,G).
isProvOpSet2(X,T,I,P,G,S) :- isPortGroup(_,T1,S,R), memezTra(T1,T), 
	splitReference(R,P1,G1,_), isPortGroupType(P,G), memezTra(P1,P), memezTra(G1,G).
	
isOpSetComment(_,T,_,_,S,D) :- isPortGroupComment(_,T,S,D).
isOpSetComment(_,_,P,G,_,D) :- isPortGroupTypeComment(P,G,D).
isOpSetComment(_,_,P,G,S,D) :- concat(S,' is instance of port group ',G,D).
isOpSetComment(_,_,P,G,_,D) :- isInverseOf(P,G,I), concat(G,' is inverse of ',I,D).

isOpSetImpBy1('NIL',P,'NIL',T,I,K,S) :- isComponent(P,T,I,K), isProvOpSet1('NIL',P,'NIL',P,G,S).
isOpSetImpBy2(X,T,I,P,G,S,C) :- isProvOpSet2(X,T,I,P,G,S), isConnection(_,T,I,_,S,_,F), 
	splitName(F,A,D), subCompName(X,T,I,D,B), concat(B,'.',A,C).

/*----- operations -----*/

isProvOp1('NIL',P,'NIL',T,O) :- isComponentType(P,T,_), isPort(P,T,O,_).
isProvOp1('NIL',P,'NIL',T,O) :- isComponentType(P,T,_), isProvOpSet1('NIL',P,'NIL',P,G,_), isPort(P,G,O,_).
isProvOp1('NIL',P,'NIL',T,O) :- isComponentType(P,T,_), isDataSubprogram(P,T,O,_).
isProvOp1('NIL',P,'NIL',T,O) :- isComponentType(P,T,_), isServerSubprogram(P,T,O,_).

isProvOp2(X,T,I,T,O) :- isPort(_,T1,O,_), memezTra(T1,T).
isProvOp2(X,T,I,G,O) :- isProvOpSet2(X,T,I,P,G,_), isPort(P,G,O,_).
isProvOp2(X,T,I,G,O) :- isProvOpSet2(X,T,I,P,G,_), not(isPort(P,G,_,_)), isInverseOf(P,G,H), isPort(P,H,O,_).
isProvOp2(X,T,I,T,O) :- isDataSubprogram(_,T1,O,_), memezTra(T1,T).
isProvOp2(X,T,I,T,O) :- isServerSubprogram(_,T1,O,_), memezTra(T1,T).

isOpComment(X,T,U,D) :- isPortComment(X,T1,U,D), memezTra(T1,T).
isOpComment(X,T,U,D) :- isDataSubpgComment(X,T1,U,D), memezTra(T1,T).
isOpComment(X,T,U,D) :- isServerSubpgComment(X,T1,U,D), memezTra(T1,T).

isOpImpBy1('NIL',P,'NIL',T,I,K,O) :- isComponent(P,T,I,K), isPort(P,T,O,_).
isOpImpBy1('NIL',P,'NIL',T,I,K,O) :- isComponent(P,T,I,K), isProvOpSet1('NIL',P,'NIL',P,G,_), isPort(P,G,O,_).
isOpImpBy1('NIL',P,'NIL',T,I,'DATA',O) :- isComponent(P,T,I,'DATA'), isDataSubprogram(P,T,O,_).
isOpImpBy1('NIL',P,'NIL',T,I,K,O) :- isComponent(P,T,I,K), isServerSubprogram(P,T,O,_).

isOpImpBy(X,T,I,O,C) :- isPort(P,T1,O,'IN'), memezTra(T1,T), isConnection(P,T,I,_,O,_,F), splitName(F,A,S), subCompName(X,T,I,S,B), concat(B,'.',A,C).
isOpImpBy(X,T,I,O,C) :- isPort(P,T1,O,'OUT'), memezTra(T1,T), isConnection(P,T,I,_,F,_,O), splitName(F,A,S), subCompName(X,T,I,S,B), concat(B,'.',A,C).

isOpImpBy2(X,T,I,P,G,O,C) :- isProvOpSet2(X,T,I,P,G,S), isConnection(_,T,I,_,U,_,F), splitName(U,O,S), isPort(P,G,O,_), 
	splitName(F,A,D), subCompName(X,T,I,D,B), concat(B,'.',A,C).
isOpImpBy2(X,T,I,P,G,O,C) :- isProvOpSet2(X,T,I,P,G,S), isConnection(_,T,I,_,F,_,U), splitName(U,O,S), isPort(P,G,O,_), 
	splitName(F,A,D), subCompName(X,T,I,D,B), concat(B,'.',A,C).
isOpImpBy2(X,T,I,P,G,O,C) :- isProvOpSet2(X,T,I,P,G,S), isConnection(_,T,I,_,U,_,F), splitName(U,O,S), 
	not(isPort(P,G,_,_)), isInverseOf(P,G,H), isPort(P,H,O,_), 
	splitName(F,A,D), subCompName(X,T,I,D,B), concat(B,'.',A,C).
isOpImpBy2(X,T,I,P,G,O,C) :- isProvOpSet2(X,T,I,P,G,S), isConnection(_,T,I,_,F,_,U), splitName(U,O,S), 
	not(isPort(P,G,_,_)), isInverseOf(P,G,H), isPort(P,H,O,_), 
	splitName(F,A,D), subCompName(X,T,I,D,B), concat(B,'.',A,C).

/*----------------------*/
/* System Configuration */
/*----------------------*/

createSystem(M) :- 
	write('OBJECT SYSTEM_CONFIGURATION IS PASSIVE'), nl, 
	write('  INTERNALS'), nl,
	write('    OBJECTS'), nl, 
	( isRoot(M,_,_,_) ->
		( isRoot(M,X,I,K), write('      '), 
		( K = 'PACKAGE' -> packageName(X,Y); Y = X ), printName(Y), 
		( isImplementation(I) -> ( write('_'), write(I) ); true ), nl ); 
		none ), fail.
createSystem(M) :-
	write('END SYSTEM_CONFIGURATION'), nl.

/*-----------------------------*/
/* Object Description Skeleton */
/*-----------------------------*/

createModule(X,T,I,K,P,R,J) :- 
	( K='DATA' -> write('CLASS '); write('OBJECT ') ), 
	moduleName(X,T,I,K,P,R,J,N), printName(N),  write(' IS '), 
	( moduleKind(X,T,I,K,A) -> write(A); write('PASSIVE') ), nl, 
	createPragmas(X,T,I), 
	createDescription(X,T,I), 
	createRealTimeAttributes(X,T,I), 
	createProvidedInterface(X,T,I,P), 
	createVisibleObcs(X,T,I), 
	createRequiredInterface(X,T,I,P), 
	createDataFlows(X,T,I,P), 
	( isTerm(X,T,I,K) -> createTermInternals(X,T,I,K,P); createNotTermInternals(X,T,I,K,P) ), 
	write('END '), printName(N), nl. 
	
/*---------*/
/* Pragmas */
/*---------*/

createPragmas(X,T,I).
 
/*-------------*/
/* Description */
/*-------------*/

createDescription(X,T,I) :- 
	write('  DESCRIPTION'), nl, 
	( isCompTypeComment(_,T,C) -> 
		( write('    PRAGMA HCS(chapter => --|Statement of the Problem (text)|--)'), nl, fileOrString(C) ); true ),
	( isCompImpComment(_,T,I,D) -> 
		( write('    PRAGMA HCS(chapter => --|General Strategy (text)|--)'), nl, fileOrString(D) ); true ), fail.
createDescription(X,T,I) :- X \= 'NIL', 
	write('    PRAGMA HCS(chapter => --|General Strategy (text)|--)'), nl, 
	write('--|'), write(X), write(' is an instance of component '), write(T), 
	( I \= 'NIL' -> ( dot, write(I)); true ), dot, write('|--'), nl, fail.
createDescription(X,T,I).	
 
/*----------------------*/
/* Real Time Attributes */
/*----------------------*/

createRealTimeAttributes(X,T,I) :- ( hasRealTimeAttributes(X,T,I) ->  
	write('  REAL_TIME_ATTRIBUTES'), nl, isProperty(_,T,I,_,_,P,V), toNum(V,N), 
	( memezTra(P,'Period') -> ( write('    PERIOD '), write(N), nl ); true ),
	( memezTra(P,'Deadline') -> ( write('    DEADLINE '), write(N), nl ); true ) ), fail.
createRealTimeAttributes(X,T,I).
 
/*--------------------*/
/* Provided Interface */
/*--------------------*/

createProvidedInterface(X,T,I,P) :- 
	write('  PROVIDED_INTERFACE'), nl,
	write('    TYPES'), nl, 
		( isProvType(X,T,I,_,_,P) -> createProvTypes(X,T,I,P); none ),
	write('    CONSTANTS'), nl, 
		( isProvConst(X,T,I,_,_,_,_) -> createProvConstants(X,T,I); none ),
	write('    OPERATION_SETS'), nl, 
		( ( isProvOpSet1(X,T,I,_,_,_); isProvOpSet2(X,T,I,_,_,_) ) -> createProvOpSets(X,T,I); none ), 
	write('    OPERATIONS'), nl, 
		( ( isProvOp1(X,T,I,_,_); isProvOp2(X,T,I,_,_) ) -> createProvOperations(X,T,I,P); none ), 
	write('    EXCEPTIONS'), nl, 
		( isProvEx(X,T,I,_,_,_) -> createProvExceptions(X,T,I); none ).
 
/*--------------*/
/* Visible OBCS */
/*--------------*/

createVisibleObcs(X,T,I) :- write('  OBJECT_CONTROL_STRUCTURE'), nl, none.
 
/*-----------*/
/* DataFlows */
/*-----------*/

createDataFlows(X,T,I,P) :- 
	( ( isConnection(_,B,_,_,O,_,D), splitName(O,U,S), splitName(D,_,X), 
	    isVisible(B,U,S,P) );
	  ( isConnection(_,_,_,_,O,_,D), splitName(O,_,R), splitName(D,A,X), 
	    R = '', isAccess(_,T,A,'REQUIRES DATA',_) ) ) -> 
	write('  DATAFLOWS'), nl, fail.
createDataFlows(X,T,I,P) :- 
	isConnection(_,B,C,_,O,_,D), splitName(O,U,S), splitName(D,_,X), S \= '', not(isPortGroup(_,B,S,_)),  
	write('    '), write(U), write(' <= '), subCompName(P,B,C,S,Y), printName(Y), nl, fail.
createDataFlows(X,T,I,P) :- 
	isConnection(_,B,C,_,O,_,D), splitName(O,U,S), splitName(D,_,X), S \= '', isPortGroup(_,B,S,_), concat(P,'.',S,G), 
	isConnection(_,_,_,_,R,_,G), splitName(R,_,E), 
	write('    '), write(U), write(' <= '), printName(E), nl, fail.
createDataFlows(X,T,I,P) :- 
	isConnection(_,B,C,_,O,_,D), splitName(O,U,S), splitName(D,_,X), S = '', isPort(_,B,U,_), 
	isConnection(Q,M,N,_,R,_,U), splitName(R,_,E), not(isSubcomponent(Q,M,N,E,_,_)), 
	write('    '), write(U), write(' <= '), printName(E), nl, fail.	
createDataFlows(X,T,I,P) :- 
	isConnection(_,B,C,_,O,_,D), splitName(O,S,R), splitName(D,A,X), R = '', isAccess(_,T,A,'REQUIRES DATA',_), 
	write('    '), write(A), write(' <= '), subCompName(P,B,C,S,Y), printName(Y), nl, fail.
createDataFlows(X,T,I,P).

isVisible(B,U,S,P) :- S \= '', not(isPortGroup(_,B,S,_)).
isVisible(B,U,S,P) :- S \= '', isPortGroup(_,B,S,_), concat(P,'.',S,G), isConnection(_,_,_,_,_,_,G).
isVisible(B,U,'',P) :- isPort(_,B,U,_), isConnection(Q,M,N,_,R,_,U), splitName(R,_,E), not(isSubcomponent(Q,M,N,E,_,_)).
 
/*-----------*/
/* Internals */
/*-----------*/

createTermInternals(X,T,I,K,P) :- 
	write('  INTERNALS'), nl,
	write('    TYPES'), nl, none, 
	write('    CONSTANTS'), nl, none, 
	write('    OPERATIONS'), nl, none, 
	write('    EXCEPTIONS'), nl, none, 
	write('    DATA'), nl, 
		( isIntData(X,T,I,_,_,_) -> createIntData(X,T,I); none ), 
	write('  OBJECT_CONTROL_STRUCTURE'), nl, none,
	write('  OPERATION_CONTROL_STRUCTURES'), nl, 
		( hasOpcs(X,T,I,K,_) -> createOpcs(X,T,I,K); none ).
 
createNotTermInternals(X,T,I,K,P) :- 
	write('  INTERNALS'), nl,
	write('    OBJECTS'), nl, 
		( isChild(X,T,I,_) -> createChildren(X,T,I); none ),
	write('    TYPES'), nl, 
		( isTypeImpBy(X,T,I,_,_,_) -> createTyImpBy(X,T,I); none ), 
	write('    CONSTANTS'), nl, 
		( isConstImpBy(X,T,I,_,_,_) -> createCoImpBy(X,T,I); none ), 
	write('    OPERATION_SETS'), nl, 
		( ( isOpSetImpBy1(X,T,I,_,_,_,_); isOpSetImpBy2(X,T,I,_,_,_,_) ) -> createOpSetImpBy(X,T,I); none ), 		
	write('    OPERATIONS'), nl, 
		( ( isOpImpBy1(X,T,I,_,_,_,_); isOpImpBy2(X,T,I,_,_,_,_); isOpImpBy(X,T,I,_,_) ) -> createOpImpBy(X,T,I); none ), 
	write('    EXCEPTIONS'), nl, none, 
	write('    DATA'), nl, none, 
	write('  OBJECT_CONTROL_STRUCTURE'), nl, none.
 
/*---------------*/
/* Child Objects */
/*---------------*/

createChildren(X,T,I) :- isChild(X,T,I,Y), write('      '), 
	( X='NIL' -> packageName(Y,C); C=Y ), printName(C), nl, fail.
createChildren(X,T,I).

/*-------*/
/* Types */
/*-------*/

createProvTypes(X,T,I,P) :- isProvType(X,T,I,U,J,P), 
	write('      '), printName(U), 
	( isImplementation(J) -> ( write('_'), write(J) ); true ), nl, 
	classInheritance(X,U,J), 
	typeAttributes(X,U,J), 
	( isTypeComment(X,U,J,D) -> ( write(';'), nl, fileOrString(D) ); true ), fail.
createProvTypes(X,T,I,P).
  
createTyImpBy(X,T,I) :- isTypeImpBy(X,T,I,Y,J,U), /* Y \= T, */
	write('      '), printName(U),  
	( isImplementation(J) -> ( write('_'), write(J) ); true ), nl, 
	write('      IMPLEMENTED_BY '), moduleName('NIL',Y,J,'DATA',T,_,_,N), printName(N), 
	dot, printName(U),
	( isImplementation(J) -> ( write('_'), write(J) ); true ), nl, fail.
createTyImpBy(X,T,I).
 
classInheritance(X,T,I) :- write('        INHERITANCE '), nl, 
	( isSuperClass(X,T,I,S) -> ( printTypeName(S), dot, printTypeName(S), nl ); none ), fail.
classInheritance(X,T,I).
 
typeAttributes(X,U,I) :- isAttribute(X,U,I,_,_) ->  write('        ATTRIBUTES'), nl, fail.
typeAttributes(X,U,I) :- firstAttribute(X,U,I,A,T), createAttributes(X,U,I,A,T), fail.
typeAttributes(X,U,I).

firstAttribute(X,U,I,A,T) :- isAttribute(X,U,I,A,T), !.

createAttributes(X,U,I,A1,T1) :- 
	printAttribute(A1,T1,''), 
	isAttribute(X,U,I,A2,T2), A1 \= A2, write(','), nl, 
	printAttribute(A2,T2,'').
createAttributes(X,U,I,A,T) :- nl.

printAttribute(A,T,V) :- write('          '), printName(A), write(' : '), printReference(T), 
	( V \= '' -> ( write(' := --|'), write(V), write('|--') ); true ).
 
/*-----------*/
/* Constants */
/*-----------*/

createProvConstants(X,T,I) :- isProvConst(X,T,I,C,V,Y,J), 
	write('      '), printName(Y,J,C), nl, fileOrString(V), 
	( isConstComment(X,T,I,C,D) -> ( write(';'), nl, fileOrString(D) ); true ), fail.
createProvConstants(X,T,I).

createCoImpBy(X,T,I) :- isConstImpBy(X,T,I,Y,J,C), 
	write('      '), printName(Y,J,C), nl, 
	write('      IMPLEMENTED_BY '), printName(Y), 
	( isImplementation(J) -> ( write('_'), write(J) ); true ), 
	dot, printName(C), nl, fail.
createCoImpBy(X,T,I).

/*------*/
/* Data */
/*------*/
 
 createIntData(X,T,I) :- isIntData(X,T,I,K,D,_), 
 	write('      '), write(D), write('_port'), 
 	write('; --| '), write(K), write(' port |--'), nl, fail.
 createIntData(X,T,I).

/*------------*/
/* Exceptions */
/*------------*/

createProvExceptions(X,T,I) :- isProvEx(X,T,I,P,O,E), 
	write('      '), printName(E), nl, 
	write('        RAISED_BY '), createOpDecl(T,O,P), write(';'), nl, 
	( isExComment(X,T,I,P,O,E,D) -> ( write(';'), nl, fileOrString(D) ); true ), fail.
createProvExceptions(X,T,I).

/*----------------*/
/* Operation Sets */
/*----------------*/

insertMemberClause(S) :- write('        MEMBER_OF '), write(S), nl.

createProvOpSets('NIL',P,'NIL') :- isProvOpSet1('NIL',P,'NIL',P,G,S), write('      '), write(S), nl, 
	( isOpSetComment('NIL',P,P,G,S,_) -> 
		write(';'), isOpSetComment('NIL',P,P,G,S,D), nl, fileOrString(D) ), 
	fail.
createProvOpSets(X,T,I) :- isProvOpSet2(X,T,I,P,G,S), write('      '), write(S), nl, 
	( isOpSetComment(X,T,P,G,S,_) -> 
		write(';'), isOpSetComment(X,T,P,G,S,D), nl, fileOrString(D) ), 
	fail.
createProvOpSets(X,T,I).

createOpSetImpBy('NIL',P,'NIL') :- isOpSetImpBy1('NIL',P,'NIL',T,I,K,S), 
	write('      '), write(S), nl, 
	write('      IMPLEMENTED_BY '), moduleName('NIL',T,I,K,P,_,_,N), printName(N), 
	dot, printName(S), nl, fail.
createOpSetImpBy(X,T,I) :- isOpSetImpBy2(X,T,I,P,G,S,C), 
	write('      '), write(S), nl, 
	write('      IMPLEMENTED_BY '), printName(C), nl, fail.
createOpSetImpBy(X,T,I).

/*------------*/
/* Operations */
/*------------*/

createProvOperations('NIL',P,'NIL',_) :- isProvOp1('NIL',P,'NIL',T,O), 
	write('      '), createOpDecl(T,O,P), 
	( isProvOpSet1('NIL',P,'NIL',P,T,S) -> insertMemberClause(S); true ), 
	( isOpComment(P,T,O,_) -> 
		write(';'), isOpComment(P,T,O,D), nl, fileOrString(D) ), 
	fail. 
createProvOperations(X,T,I,P) :- isProvOp2(X,T,I,Y,O), 
	write('      '), ( isPackage(P,_) -> createOpDecl(Y,O,P); createOpDecl(Y,O) ), 
	( isProvOpSet2(X,T,I,_,Y,S) -> insertMemberClause(S); true ), 
	( isOpComment(_,Y,O,_) -> 
		write(';'), isOpComment(_,Y,O,D), nl, fileOrString(D) ), 
	fail. 
createProvOperations(X,T,I,P).
 
createOpImpBy('NIL',P,'NIL') :- isOpImpBy1('NIL',P,'NIL',T,I,K,O), 
	write('      '), createOpDecl(T,O,P),  
	write('      IMPLEMENTED_BY '), moduleName('NIL',T,I,K,P,_,_,N), printName(N), 
	dot, createOpDecl(T,O,P), fail.
createOpImpBy(X,T,I) :- isOpImpBy(X,T,I,O,F), 
	write('      '), createOpDecl(T,O),  
	write('      IMPLEMENTED_BY '), printName(F), 
	isPort(_,T1,O,D), memezTra(T1,T), createPortParam(T,O,D), nl, fail.
createOpImpBy(X,T,I) :- isOpImpBy2(X,T,I,P,G,O,F), 
	write('      '), createOpDecl(G,O,P),  
	write('      IMPLEMENTED_BY '), printName(F), 
	( ( isPort(_,G1,O,D), memezTra(G1,G) ) -> createPortParam(G,O,D);
	( isInverseOf(_,G,H), isPort(_,H1,O,D), memezTra(H1,H), 
	  inverseDir(D,E), createPortParam(H,O,E) ) ),
	nl, fail.
createOpImpBy(X,T,I).
 
createOpDecl(T,O) :- printName(O), fail.
createOpDecl(T,O) :- isPort(_,T1,O,D), memezTra(T1,T), createPortParam(T,O,D), fail.
createOpDecl(T,O) :- isInverseOf(_,T,U), isPort(_,U1,O,D), memezTra(U1,U), 
	inverseDir(D,E), createPortParam(U,O,E), fail.
createOpDecl(T,O) :- isDataSubprogram(P,T1,O,C), memezTra(T1,T), splitReference(C,Y,S,_), 
	( Y = 'NIL' -> createSpgParam(P,S); createSpgParam(Y,S) ), fail.
createOpDecl(T,O) :- isServerSubprogram(P,T1,O,C), memezTra(T1,T), splitReference(C,Y,S,_), 
	( Y = 'NIL' -> createSpgParam(P,S); createSpgParam(Y,S) ), fail.
createOpDecl(T,O) :- nl.

createOpDecl(T,O,P) :- printName(O), fail.
createOpDecl(T,O,P) :- isPort(_,T1,O,D), memezTra(T1,T), createPortParam(T,O,D), fail.
createOpDecl(T,O,P) :- isInverseOf(_,T,U), isPort(_,U1,O,D), memezTra(U1,U), 
	inverseDir(D,E), createPortParam(U,O,E), fail.
createOpDecl(T,O,P) :- isDataSubprogram(P,T1,O,C), memezTra(T1,T), splitReference(C,Y,S,_), 
	( Y = 'NIL' -> createSpgParam(P,S); createSpgParam(Y,S) ), fail.
createOpDecl(T,O,P) :- isServerSubprogram(P,T1,O,C), memezTra(T1,T), splitReference(C,Y,S,_), 
	( Y = 'NIL' -> createSpgParam(P,S); createSpgParam(Y,S) ), fail.
createOpDecl(T,O,P) :- nl.

/*----------------------*/
/* Operation Parameters */
/*----------------------*/

createSpgParam(X,U) :- ( isParameter(X,U1,_,_,_), memezTra(U1,U) ) -> write(' ('), nl, fail.
createSpgParam(X,U) :- firstParam(X,U,P,D,T), createParams(X,U,P,D,T), fail.
createSpgParam(X,U) :- ( isParameter(X,U1,_,_,_), memezTra(U1,U) ) -> write(' )'), fail. 

createPortParam(T,O,D) :- isPortType(_,T1,O,K,R), memezTra(T1,T), write(' ( '), 
	  ( ( K='event'; K='event data' ) -> printParam('event',D,'AADL::Event'); true ), 
	  ( K='event data' -> write('; '); true ),
	  ( ( K='data'; K='event data' ) -> printParam('port',D,R); true ), write(' )').
 
firstParam(X,U,P,D,T) :- isParameter(X,U1,P,D,T), memezTra(U1,U), !.
createParams(X,U,A,B,C) :- 
	write('          '), printParam(A,B,C), 
	isParameter(X,U,D,E,F), D\=A, write(';'), nl, 
	write('          '), printParam(D,E,F).

printParam(P,M,R) :- 
	printParamName(P), 
	printParamMode(M), 
	sp, printReference(R).

printParamName(P) :- 
	printName(P), write(' : ').
printParamMode('IN') :-  write('in').
printParamMode('OUT') :-  write('out').
printParamMode('IN OUT') :-  write('in out').

inverseDir('IN','OUT').
inverseDir('OUT','IN').
inverseDir('IN OUT','IN OUT').
 
/*------*/
/* OPCS */
/*------*/

createOpcs(X,T,I,K) :- hasOpcs(X,T,I,K,O), 
	write('      OPERATION '), createOpDecl(T,O), 
	write('        PSEUDO_CODE '), ( insertPseudo(X,T,I,O) -> true; none ), 
	write('        WCET '), ( isWcet(X,T,I,O,W) -> write(W); write('0') ), nl, 
	write('      END_OPERATION '), printName(O), nl,
	fail.
createOpcs(X,T,I,K).

insertPseudo(X,T,I,'thread') :- insertCall(_,T,I).

insertPseudo(X,T,I,O) :- isIntData(X,T,I,_,O,'IN'), nl, 
	write('--|'), write(O), write('_port := port|--'), nl.
insertPseudo(X,T,I,O) :- isIntData(X,T,I,_,O,'OUT'), nl, 
	write('--|port := '), write(O), write('_port|--'), nl.
insertPseudo(X,T,I,O) :- isIntData(X,T,I,_,O,'IN OUT'), nl, 
	write('--|'), write(O), write('_port := port'), nl, 
	write('port := '), write(O), write('_port|--'), nl.

insertPseudo(X,T,I,O) :- isProvOp2(X,T,I,T,O), isDataSubprogram(P,T1,O,C), memezTra(T1,T), 
	splitReference(C,P,R,J), insertCall(P,R,J).
insertPseudo(X,T,I,O) :- isProvOp2(X,T,I,T,O), isServerSubprogram(P,T1,O,C), memezTra(T1,T), 
	splitReference(C,P,R,J), insertCall(P,R,J).

insertCall(P,T,I) :- isCall(P,T,I,_,_) -> nl, write('--|'), fail.
insertCall(P,T,I) :- isCall(P,T,I,_,C), printCallReference(C), nl, fail.
insertCall(P,T,I) :- isCall(P,T,I,_,_) -> write('|--'), nl.
 
/*--------------------*/
/* Required Interface */
/*--------------------*/

createRequiredInterface(X,T,I,P) :- hasRequired(X,T,I,P), 
	write('  REQUIRED_INTERFACE '), nl, fail.
createRequiredInterface(X,T,I,P) :- isAttribute(X,T,I,A,Y), isComponent(_,Y,J,'DATA'), 
	(  J \= 'NIL' -> concat(Y,'_',J,Z); Z = Y ), 
	write('    OBJECT '), printName(Z), write(';'), nl, 
	write('      TYPES '), nl, 
	write('        PRAGMA STOOD_CONNEXION_TYPE (T=>1)'), nl, 
	write('        NONE '), nl, fail. 
createRequiredInterface(X,T,I,P).

hasRequired(X,T,I,P) :- isAttribute(X,T,I,_,Y), isComponent(_,Y,_,'DATA'), !.
 
/*----------*/
/* printing */
/*----------*/

none :- write('      NONE'), nl.

printName('') :- write('MISSING_IDENTIFIER'), !. 
printName(U) :- 
	( ( isReserved(V), memezTra(U,V) ) -> write('sif_keyword_'); true ), 
	write(U). 

printFreeText(C) :- write(' --|'), writeLong(C), write('|--'), nl. 

printTypeName(T) :-  splitName(T,S,P), 
	( P\='' -> ( splitName(P,O,_), printName(O), dot ); true ), 
	printType(S).

printType(T) :- 
	( ( isReserved(V), memezTra(T,V) ) -> write('sif_keyword_'); true ), 
	$name(T,L), $member(C,L), 
	( C =:= 39 -> ( $put(95), warning2(T) ); $put(C) ), fail.
printType(T).

write(X) :- $writename(X).
writeLong(X) :- $name(X,L), $member(E,L),  ( E =:= 124 -> $put(33); $put(E) ), fail. 
writeLong(X).
print(X) :- $telling(F), $told, $tell('user'), $writename(X), $tell(F,1).

fileOrString(C) :- 
	write('--|'), basicFileOrString(C), write('|--'), nl.

basicFileOrString(C) :- ( $conlength(C,L), L < 4 ) -> write(C); 
	( $substring(1,4,'tmp_',0,C,_) -> ( $see(C) -> printFile; writeLong(C) ); writeLong(C) ).

printFile :- $get0(C), 
	( C =\= -1 -> (  ( C =:= 124 -> $put(33); $put(C) ), printFile ); $seen ). 

nl :- $put(10).
tab :- $put(9).
sp :- $put(32).
dot :- $put(46).

/*-----------*/
/* utilities */
/*-----------*/

memezTra(A,B) :- $name(A,L1), $name(B,L2), id(L1,L2).

memezTra(A,B,C,D) :- $name(A,L1), $name(B,L2), $name(C,L3), $name(D,L4),
	$append(L2,L3,L5), $append(L5,L4,L6), id(L1,L6).

notMemezTra(A,B,C,D) :- $name(A,L1), $name(B,L2), $name(C,L3), $name(D,L4),
	$append(L2,L3,L5), $append(L5,L4,L6), not(id(L1,L6)).
	
id([],[]) :- true.
id([E|L1],[F|L2]) :- ( E =:= F ; E =:= F-32 ; E-32 =:= F ), id(L1,L2).

splitName(X,A,B) :- X \= '', /* X = B.A */
	$name(X,L1), $reverse(L1,R1), $name(X1,R1),
	$subdelim(1,'.',A2,0,X1,P) -> (
		$name(A2,R2), $reverse(R2,L2), $name(A,L2),
		$conlength(X,S), Q is S-P, $substring(1,Q,B3,P,X1,_),
		$name(B3,R3), $reverse(R3,L3), $name(B,L3) );
		( A=X, B='').

addSuffix(F1,S,F2) :- $name(F1,L1), $name(S,L2), $append(L1,L2,L3), $name(F2,L3). 
	
splitReference(N,M,T,I) :- splitName(N,A,B), /* N = M::T.I */
	( B \= '' -> ( I=A, C=B ); ( I='NIL', C=A ) ), 
	$name(C,L1), $reverse(L1,R1), $name(C1,R1),
	( $subdelim(1,':',T2,0,C1,P) -> (
		$name(T2,R2), $reverse(R2,L2), $name(T,L2), 
		$conlength(C,S), P1 is P+1, Q is S-P1,
		$substring(1,Q,M3,P1,C1,_), 
		$name(M3,R3), $reverse(R3,L3), $name(M,L3) );
		( T=C, M='NIL') ).

printReference(R) :- splitReference(R,P,T,I), 
	( P \= 'NIL' -> ( packageName(P,Q), printName(Q), dot ); true ), 
	( T \= 'NIL' -> ( printName(T) ); true ), 
	( isImplementation(I) -> ( write('_'), printName(I) ); true ).

printCallReference(C) :- splitReference(C,P,T,O), 
	( P \= 'NIL' -> ( packageName(P,Q), printName(Q), dot ); true ), 
	( O \= 'NIL' -> printName(O); printName(T) ).

rootPackage(P,R) :- $subdelim(1,':',Q,0,P,_) -> R = Q; R = P.
leafPackage(P,L) :- 
	$name(P,L1), $reverse(L1,R1), $name(P1,R1), 
	rootPackage(P1,P2),
	$name(P2,R2), $reverse(R2,L2), $name(L,L2).

isChildPackage(P,C) :- splitReference(C,P,_,_).

packageName(P,Q) :- P \= '', $name(P,L1), convert(L1,[],L2), $name(Q,L2).  /* :: -> _ */

convert([],R,L) :- $reverse(R,L), !. 
convert([C1|L1],L2,L) :- ( C1 =:= 58 ) -> 
	convert(L1,[95|L2],L); convert(L1,[C1|L2],L).
	
concat(A,B,C) :- $name(A,L1), $name(B,L2), 
	$append(L1,L2,L), $name(C,L).
concat(A,B,C,D) :- $name(A,L1), $name(B,L2), $name(C,L3), 
	$append(L1,L2,L4), $append(L4,L3,L), $name(D,L).
concat(A,B,C,D,E) :- $name(A,L1), $name(B,L2), $name(C,L3), $name(D,L4), 
	$append(L1,L2,L5), $append(L5,L3,L6), $append(L6,L4,L), $name(E,L).
concat(A,B,C,D,E,F) :- $name(A,L1), $name(B,L2), $name(C,L3), $name(D,L4), $name(E,L5),
	$append(L1,L2,L6), $append(L6,L3,L7), $append(L7,L4,L8), $append(L8,L5,L), $name(F,L).
	
toNum(S,N) :-  S \= '', $name(S,L1), remove(L1,[],L2), $name(N,L2).

remove([],R,L) :- $reverse(R,L), !.
remove([C1|L1],L2,L) :- numChar(C1) -> 
	remove(L1,[C1|L2],L); remove(L1,L2,L).

numChar(C) :- C > 47, C < 58. /* 0..9 */
numChar(46). /* . */

/*--------------*/
/* listPackages */
/*--------------*/

listPackages :- ( isPackage(P,'PUBLIC'); isPackage(P,'BOTH') ), extractAllPackages(P,0,Q), 
	( isPackage(Q) -> true; $assert(isPackage(Q)) ), fail.
listPackages.
	
extractAllPackages(P,A,Q) :- $subdelim(1,':',R,A,P,B) ->
	( Q = R; C is B + 1, extractAllPackages(P,C,Q) );
	( $conlength(P,L), M is L - A, $substring(1,M,Q,A,P,_) ).
	
/*----------*/
/* messages */
/*----------*/

warning1(U) :- $telling(F), $told, $tell('user'), 
	write('WARNING: '), write(U), 
	write(' is a reserved word for HOOD, and has been renamed into: '), 
	write(U), write('_'), nl,
	$told, $tell(F,1).

warning2(U) :- $telling(F), $told, $tell('user'), 
	write('WARNING: for parameter type: '), write(U), 
	write(' tick character was replaced by an underline character'), nl, 
	$told, $tell(F,1).
 
warning3(U) :- $telling(F), $told, $tell('user'), 
	write('WARNING: exception: '), write(U), 
	write(' has been inserted into module internals'), nl, 
	$told, $tell(F,1).

warning4(U) :- $telling(F), $told, $tell('user'), 
	write('WARNING: variable: '), write(U), 
	write(' has been inserted into module provided interface, as a constant'), nl, 
	$told, $tell(F,1).
	
debug(E,T) :- $telling(F), $told, $tell('user'), 
	write(E), sp, write(T), nl, 
	$told, $tell(F,1).

/*----------------*/
/* Reserved Words */
/*----------------*/

isReserved('ATTRIBUTES').
isReserved('ASER_BY_IT').
isReserved('ASER').
isReserved('ACTIVE').
isReserved('ABSTRACT').
isReserved('BY_VALUE').
isReserved('BY_REFERENCE').
isReserved('BY_POINTER').
isReserved('BY_IT').
isReserved('CONSTRAINED_OPERATIONS').
isReserved('CONSTRAINED_BY').
isReserved('CONSTANTS').
isReserved('CODE').
isReserved('CLASS').
isReserved('DESCRIPTION').
isReserved('DATAFLOWS').
isReserved('DATA').
isReserved('EXCEPTION_FLOWS').
isReserved('EXCEPTIONS').
isReserved('ENVIRONMENT').
isReserved('END_OPERATION').
isReserved('END_OBJECT').
isReserved('END').
isReserved('FORMAL_PARAMETERS').
isReserved('GENERIC').
isReserved('HSER_TOER').
isReserved('HSER').
isReserved('HANDLED_EXCEPTIONS').
isReserved('IS').
isReserved('INTERNALS').
isReserved('INSTANCE_RANGE').
isReserved('INSTANCE_OF').
isReserved('INHERITED').
isReserved('INHERITANCE').
isReserved('IN').
isReserved('IMPLEMENTED_BY').
isReserved('IMPLEMENTATION_OR_SYNCHRONIZATION_CONSTRAINTS').
isReserved('IMPLEMENTATION_OR_SYNCHRONISATION_CONSTRAINTS').
isReserved('IMPLEMENTATION_CONSTRAINTS').
isReserved('LSER_TOER').
isReserved('LSER').
isReserved('MTEX').
isReserved('MEMBER_OF').
isReserved('NONE').
isReserved('OUT').
isReserved('OSTD').
isReserved('OP_CONTROL').
isReserved('OPERATION_SETS').
isReserved('OPERATION_CONTROL_STRUCTURES').
isReserved('OPERATIONS').
isReserved('OPERATION').
isReserved('OBJECT_CONTROL_STRUCTURE').
isReserved('OBJECTS').
isReserved('OBJECT').
isReserved('PSEUDO_CODE').
isReserved('PROVIDED_INTERFACE').
isReserved('PROTOCOL').
isReserved('PROPAGATED_EXCEPTIONS').
isReserved('PRAGMA').
isReserved('PASSIVE').
isReserved('PARAMETERS').
isReserved('RWER').
isReserved('ROER').
isReserved('RLSER').
isReserved('RETURN').
isReserved('REQUIRED_INTERFACE').
isReserved('RASER').
isReserved('RAISED_BY').
isReserved('STATE').
isReserved('SIF_VERSION').
isReserved('TYPES').
isReserved('TOOL_VERSION').
isReserved('TOOL_NAME').
isReserved('TO').
isReserved('TARGET_LANGUAGE').
isReserved('USED_OPERATIONS').
isReserved('VIRTUAL_NODE').

/* HRT keywords */

isReserved('PROTECTED').
isReserved('CYCLIC').
isReserved('SPORADIC').
isReserved('REAL_TIME_ATTRIBUTES').
isReserved('WCET').
isReserved('BUDGET').
isReserved('CEILING_PRIORITY').
isReserved('PERIOD').
isReserved('OFFSET').
isReserved('MIN_ARRIVAL_TIME').
isReserved('MAX_ARRIVAL_FREQUENCY').
isReserved('DEADLINE').
isReserved('PRIORITY').
isReserved('PRECEDENCE_CONSTRAINTS').
isReserved('BEFORE').
isReserved('AFTER').
isReserved('EXECUTION_TIME_TRANSFORMATION').
isReserved('IMPORTANCE').
isReserved('HARD').
isReserved('SOFT').
isReserved('BACKGROUND').
isReserved('PAER').
isReserved('PSER').
isReserved('PSER_TOER').
isReserved('ASATC').
isReserved('LSATC').
isReserved('HSATC').
isReserved('LSATC_TOER').
isReserved('HSATC_TOER').
isReserved('INTEGRITY').
 
/*---------*/
/* counter */
/*---------*/

cptRst :- $globalset(cpt(0)).
cptPut(I) :- $globalset(cpt(I)).
cptGet(I) :- cpt(I).
cptInc :- cpt(I), J is I + 1, $globalset(cpt(J)).
cptDec :- cpt(I), J is I - 1, $globalset(cpt(J)).
cptAdd(I) :- cpt(J), K is J + I, $globalset(cpt(K)).
cptSub(I) :- cpt(J), K is J - I, $globalset(cpt(K)).
cptMul(I) :-  cpt(J), K is J * I, $globalset(cpt(K)).
cptDiv(I) :- cpt(J), K is J // I, $globalset(cpt(K)).
cptNeg :- cpt(I), J is -I, $globalset(cpt(J)).

