--  visibility on required modules : 
package body bank is
  function Find (me : IN Customer_Type) return Customer_ID;
  type Customer_Access is access Customer_Type;
  type Password_Access is access Password_Type;
  protected type Account_Type is 
  private
    function Balance return Integer;
    function Check (Password : IN Password_Type) return Boolean;
    function Customer return Customer_Type;
    procedure Deposit (Amount : IN Integer);
    procedure Initialize (Customer : IN Customer_Type; Password : IN Password_Type; Deposit : IN Integer; Available : OUT Boolean);
    function Password return Password_Type;
    My_Customer : Customer_Access;
    My_Password : Password_Access;
    My_Balance : Integer := 0;
  end Account_Type;
  Null_Customer : constant Customer_Type := "";
  Null_ID  : constant Customer_ID := 0;
  First_ID : constant Customer_ID := 1;
  Last_ID  : constant Customer_ID := N_Customer_IDs;
  Accounts      : array (First_ID .. Last_ID) of Account_Type;
  function Balance (me : IN Customer_Type; Password : IN Password_Type) return Integer is
  	ID : Customer_ID := Find(me);
  	Balance : Integer :=0;
  begin
  	if ID = Null_ID then
  		raise Wrong_Customer;
  	end if;
  	if Accounts(ID).Check(Password) then
  		return Accounts(ID).Balance;
  	else
  		raise Wrong_Password;
  	end if;
  end Balance;
  procedure Create (me : IN Customer_Type; Password : IN Password_Type; Deposit : IN Positive) is
   	Done : Boolean;
  begin
  	if Find (me) /= Null_ID then
  		raise Wrong_Customer;
  	end if;
  	for N in Accounts'range loop
  		Accounts(N).Initialize(me,Password,Deposit,Done);
  		if Done then
  			return;
  		end if;
  	end loop;
  raise No_More_IDs;
  end Create;
  procedure Deposit (me : IN Customer_Type; Amount : IN Positive) is
  	ID : Customer_ID := Find(me);
  begin
  	if ID = Null_ID then
  		raise Wrong_Customer;
  	end if;
  	Accounts(ID).Deposit(Amount);
  end Deposit;
  procedure Withdraw (me : IN Customer_Type; Password : IN Password_Type; Amount : IN Positive) is
  	ID : Customer_ID := Find(me);
  begin
  	if ID = Null_ID then
  		raise Wrong_Customer;
  	end if;
  	if Accounts(ID).Check(Password) then
  		Accounts(ID).Deposit(-Amount);
  	else
  		raise Wrong_Password;
  	end if;
  end Withdraw;
  function Is_Activated (me : IN Customer_ID) return Boolean is
  	The_Customer : Customer_Type := Accounts(me).Customer;
  begin
  	return The_Customer /= Null_Customer;
  end Is_Activated;
  function Get_Customer (me : IN Customer_ID) return Customer_Type is
  begin
     return Accounts(me).Customer;
  end Get_Customer;
  function Get_Password (me : IN Customer_ID) return Password_Type is
  begin
     return Accounts(me).Password;
  end Get_Password;
  function Get_Balance (me : IN Customer_ID) return Integer is
  begin
     return Accounts(me).Balance;
  end Get_Balance;
  function Find (me : IN Customer_Type) return Customer_ID is
  begin
  	for N in Accounts'range loop
  		if Accounts(N).Customer = me then
  			return N;
  		end if;
  	end loop;
  	return Null_ID;
  end Find;
  protected body Account_Type is 
    function Balance return Integer is
    begin
       return My_Balance;
    end Balance;
    function Check (Password : IN Password_Type) return Boolean is
    begin
       return My_Password /= null and then
              My_Password.all = Password;
    end Check;
    function Customer return Customer_Type is
    begin
       if My_Customer = null then
          return Null_Customer;
       else
          return My_Customer.all;
       end if;
    end Customer;
    procedure Deposit (Amount : IN Integer) is
    begin
       My_Balance := My_Balance + Amount;
    end Deposit;
    procedure Initialize (Customer : IN Customer_Type; Password : IN Password_Type; Deposit : IN Integer; Available : OUT Boolean) is
    begin
       if Customer = Null_Customer then
          raise Wrong_Customer;
       end if;
       if My_Customer = null then
          My_Customer := new Customer_Type'(Customer);
          My_Password := new Password_Type'(Password);
          My_Balance  := Deposit;
          Available   := True;
       else
          Available   := False;
       end if;
    end Initialize;
    function Password return Password_Type is
    begin
       if My_Customer = null then
          return "";
       else
          return My_Password.all;
       end if;
    end Password;
  end Account_Type;
end bank;
