-- Dining Philosophers - Ada 95 edition
-- 
-- Room.OBCS is responsible for assigning seats at the table, 
-- "left" and "right" chopsticks, and for reporting interesting 
-- events to the outside world.
-- 
-- Michael B. Feldman, The George Washington University, July 1995.
-- HOOD version by Pierre Dissaux, TNI, June 1998.


--  required interface : 
--   Required OPERATION :
--     OPERATION : open of object : windows
--     OPERATION : title of object : windows
--     OPERATION : borders of object : windows
--     OPERATION : put#1 of object : windows
--     OPERATION : new_line of object : windows
--     OPERATION : start_eating of object : phil
--     OPERATION : get_name of object : society
--     OPERATION : Clock of object : calendar
--   Required EXCEPTION : NONE
--   Required TYPE :
--     TYPE : Window of object : windows
--     TYPE : Philosopher of object : phil
--     TYPE : Philosopher_Ptr of object : phil
--     TYPE : States of object : phil
--     TYPE : Unique_DNA_Codes of object : society
--     TYPE : Stick of object : chop
--     TYPE : Stick_Ptr of object : chop
--     TYPE : Boolean of object : standard
--     TYPE : Integer of object : standard
--     TYPE : Natural of object : standard
--     TYPE : Positive of object : standard
--     TYPE : Time of object : calendar
--   Required CONSTANT : NONE
--   Required DATA : NONE

--  visibility on required modules : 
with phil;
use type phil.Philosopher;
use type phil.Philosopher_Ptr;
use type phil.States;
with society;
use type society.Unique_DNA_Codes;
with chop;
use type chop.Stick;
use type chop.Stick_Ptr;

package room is

  -- Room.Get_Stick is an access function to internal Sticks variable.
  -- It requires a chopstick ID (Which_Stick) to return a pointer to relevant
  --  protected object.
  -- (cf.FR2/Provide_chopsticks:)
  function get_stick (
    which_Stick : IN Positive)
     return Chop.Stick_Ptr;

  -- A dedicated state variable manages current state of dining room.
  -- This variable "Started" has a default value of "FALSE" and become "TRUE"
  --  after Start_Serving has been executed.
  -- Start_Serving and Report_State have both STATE and protocole constraints.
  task OBCS is
    entry start_serving;
    entry report_state (
      Which_Phil : IN Society.Unique_DNA_Codes; 
      Which_State : IN Phil.States; 
      How_Long : IN Natural := 0; 
      Which_Meal : IN Natural := 0);
  end OBCS;
  -- Room.Start_Serving is called by main procedure and renames OBCS.Start_Serving
  --  task entry.
  -- This procedure has no parameter. 
  -- (cf.FR1/Prepare&begin_diner:)
  procedure start_serving
     renames OBCS.start_serving;

  -- Room.Report_State is called by Phil.Start_Eating and renames OBCS.Report_State
  --  task entry.
  -- This procedure has four parameters:
  -- - Which_Phil: identifies actual Philosopher sending the message.
  -- - State: current state of sender.
  -- - How_Long: length of current state (or identifier of used chopstick).
  -- - Which_Meal: current meal.
  -- (cf.FR3/Report_events:)
  procedure report_state (
    Which_Phil : IN Society.Unique_DNA_Codes; 
    Which_State : IN Phil.States; 
    How_Long : IN Natural := 0; 
    Which_Meal : IN Natural := 0)
     renames OBCS.report_state;

end room;
