-- Dining Philosophers - Ada 95 edition
-- 
-- Manager for simple, nonoverlapping screen windows.
-- 
-- Michael B. Feldman, The George Washington University, July 1995.
-- HOOD version by Pierre Dissaux, TNI, June 1998.


--  required interface : 
--   Required OPERATION :
--     OPERATION : ClearScreen of object : screen
--     OPERATION : MoveCursor of object : screen
--     OPERATION : New_Line of object : text_io
--     OPERATION : Put of object : text_io
--   Required EXCEPTION : NONE
--   Required TYPE :
--     TYPE : Height of object : screen
--     TYPE : Width of object : screen
--     TYPE : Position of object : screen
--     TYPE : Character of object : standard
--     TYPE : String of object : standard
--   Required CONSTANT : NONE
--   Required DATA : NONE

--  visibility on required modules : 
with screen;
use type screen.Height;
use type screen.Width;
use type screen.Position;

package windows is

  type Window is private;

  -- Pre:	UpperLeft, Weight, and Width are defined
  -- Post:	returns a Window with the given upper-left corner, height, and width
  function open (
    UpperLeft : IN Screen.Position; 
    Height : IN Screen.Height; 
    Width : IN Screen.Width)
     return Window;

  -- Pre: 	me, Name, and Under are defined
  -- Post:	Name is displayed at the top of the window me, underlined with the
  --  character Under
  procedure title (
    me : IN OUT Window; 
    Name : IN String; 
    Under : IN Character);

  -- Pre:	All parameters are defined
  -- Post:	Draw border around current writable area in window with characters
  --  specified. 
  -- 	Call this BEFORE Title.
  procedure borders (
    me : IN OUT Window; 
    Corner : IN Character; 
    Down : IN Character; 
    Across : IN Character);

  -- Pre:	me, and P are defined, and P lies within the area of me
  -- Post:	Cursor is moved to the specified position.
  -- 	Coordinates are relative to the upper left corner of me, which is (1,1)
  procedure movecursor (
    me : IN OUT Window; 
    P : IN Screen.Position);

  -- Pre:	me, and Ch are defined.
  -- Post:	Ch is displayed in the window at the next available position.
  -- 	If end of column, go to the next row.
  -- 	If end of window, go to the top of the window.
  procedure put (
    me : IN OUT Window; 
    Ch : IN Character);

  -- Pre:	me, and S are defined.
  -- Post:	Ch is displayed in the window, "line-wrapped" if necessary
  procedure put (
    me : IN OUT Window; 
    S : IN String);

  -- Pre:	me is defined.
  -- Post:	Cursor moves to beginning of next line of me;
  -- 	line is not blanked until next character is written
  procedure new_line (
    me : IN OUT Window);

private

  -- First : coordinates of upper left corner;
  -- Last : coordinates of lower right corner;
  -- Current : current cursor position.
  type Window is tagged 
  record 
    First : screen.Position;
    Last : screen.Position;
    Current : screen.Position;
  end record;

end windows;
