with Automobile_Interface_Units;
package Aiu renames Automobile_Interface_Units;
with Auto_Simulator_Units;
package Asu renames Auto_Simulator_Units;
separate (Automobile_Interface)
function Engine return Aiu.Rev_Per_Minute is
begin
  return +0; -- Placeholder for now
end Engine;
with Mid_Sized_Auto;

separate (Automobile_Interface)
function Speed return Aiu.Feet_Per_Second is
begin
  return Mid_Sized_Auto.Current_Speed;
end Speed;
with Mid_Sized_Auto;

separate (Automobile_Interface)
procedure Throttle (Setting : Aiu.Millimeters) is

  Driver_Command : Aiu.Millimeters;

  function Gas_Pedal return Aiu.Millimeters is
  begin
    if Unprocessed_Command then
      if Command_Was_Received(More_Gas) then
        Command_Was_Received(More_Gas) := False;
        Unprocessed_Command := False;
        Gas := Gas + (+1);
        if Gas > Mid_Sized_Auto.Throttle_Limit then
          Gas := Mid_Sized_Auto.Throttle_Limit;
        end if;
        Brake_Is_On := False;
      elsif Command_Was_Received(Less_Gas) then
        Command_Was_Received(Less_Gas) := False;
        Unprocessed_Command := False;
        Gas := Gas - (+1);
        if Gas < +0 then
          Gas := +0;
        end if;
        Brake_Is_On := False;
      end if;
    end if;
    return Gas;
  end Gas_Pedal;

begin
  Driver_Command := Gas_Pedal;
  if Driver_Command > Setting then
    Mid_Sized_Auto.Accelerate_Body(Driver_Command);
    Throttle_Setting := Driver_Command;
  else
    Mid_Sized_Auto.Accelerate_Body(Setting);
    Throttle_Setting := Setting;
  end if;
end Throttle;
-- This body lets a user at a terminal interactively test the
-- performance of a cruise control algorithm.

with Ada.Integer_Text_Io; use Ada.Integer_Text_Io;
with Ada.Text_Io; use Ada.Text_Io;
with Di;
with Interfaces; use type Interfaces.Integer_32;

with Keyboard;

package body Automobile_Interface is

  type Command_Kind is 
      (Toggle_Brake, 
       Set_Desired_Speed, 
       Resume_Desired_Speed, 
       Turn_Control_Off, 
       More_Gas, 
       Less_Gas);

  Brake_Was_On : Boolean := True;

  Brake_Is_On : Boolean := False;

  Cruise_Control_Was_On : Boolean := True;

  Gas : Aiu.Millimeters := +0;

  Last_Gas : Aiu.Millimeters := Aiu.Type_Convert(Di.Last);

  Last_Mph : Integer_32 := Integer_32'First;

  Last_Throttle_Setting : Aiu.Millimeters := Type_Convert(Di.Last);

  Command_Was_Received : array (Command_Kind) of Boolean := (others => False);

  Throttle_Setting : Aiu.Millimeters := +0;

  Unprocessed_Command : Boolean := False;

  function Brake return Boolean is
  begin
    if Unprocessed_Command and then Command_Was_Received(Toggle_Brake) then
      Command_Was_Received(Toggle_Brake) := False;
      Unprocessed_Command := False;
      Brake_Is_On := not Brake_Is_On;
    end if;
    if Brake_Is_On /= Brake_Was_On then
      Put("Brake is ");
      if Brake_Is_On then
        Put_Line("On");
      else
        Put_Line("Off");
      end if;
    end if;
    Brake_Was_On := Brake_Is_On;
    if Brake_Is_On then
      Gas := +0;
      return True;
    else
      return False;
    end if;
  end Brake;

  procedure Cruise_Control -- display state of Cruise Control
      (On : Boolean;
       Speed : Aiu.Feet_per_second) is
  begin
    if On /= Cruise_Control_Was_On then
      Cruise_Control_Was_On := On;
      Put("Cruise Control is ");
      if On then
        Put("On, Cruising speed = ");
        Put(Integer(Aiu.Dimensionless(Speed) * 60 / 88), Width => 0);
        Put_Line(" mph");
      else
        Put_Line("Off");
      end if;
    end if;
  end Cruise_Control;

  function Cruise_Off return boolean is
  begin
    if Unprocessed_Command and then Command_Was_Received(Turn_Control_Off) then
      Command_Was_Received(Turn_Control_Off) := False;
      Unprocessed_Command := False;
      return True;
    end if;
    return False;
  end Cruise_Off;

  function Engine return Aiu.Rev_per_minute is separate;
  -- This function depends on the auto simulator.

  function Resume return Boolean is
  begin
    if Unprocessed_Command and then Command_Was_Received(Resume_Desired_Speed) then
      Command_Was_Received(Resume_Desired_Speed) := False;
      Unprocessed_Command := False;
      return True;
    end if;
    return False;
  end Resume;

  function Set_Speed return Boolean is
  begin
    if Unprocessed_Command and then Command_Was_Received(Set_Desired_Speed) then
      Command_Was_Received(Set_Desired_Speed) := False;
      Unprocessed_Command := False;
      return True;
    end if;
    return False;
  end Set_Speed;

  function Speed return Aiu.Feet_per_second is separate;
  -- This function depends on the auto simulator.

  procedure Speedometer(Speed : Aiu.Feet_per_second) is
    Mph : Integer_32;
  begin
    -- Show Gas, Brake, and Throttle diagnostic info first.
    if Gas /= Last_Gas then
      Last_Gas := Gas;
      Put("Gas = ");
      Put(Integer(Aiu.Dimensionless(Gas)), Width => 0);
      New_Line;
    end if;
    if Throttle_Setting /= Last_Throttle_Setting then
      Last_Throttle_Setting := Throttle_Setting;
      Put("Throttle = ");
      Put(Integer(Aiu.Dimensionless(Throttle_Setting)), Width => 0);
      New_Line;
    end if;
    -- 88 feet/sec = 60 MPH
    Mph := (Aiu.Dimensionless(Speed) * 60) / 88;
    if Mph /= Last_Mph then
      Put("Speed = ");
      Put(Integer(Mph), Width => 0);
      New_Line;
    end if;
    Last_Mph := Mph;
  end Speedometer;

  procedure Throttle(Setting : Aiu.Millimeters) is separate;
  -- This function depends on the auto simulator.

  procedure Update_Driver_Inputs is
  -- Gets the next command available, if any.

    Key : Character;

  begin
    -- If we already know there is an unprocessed command,
    -- then don't bother to check.
    if Unprocessed_Command then
      return;
    else
      if Keyboard.Hit then
        Key := Keyboard.Char;
        case Key is
          when 'S' | 's' =>
            Command_Was_Received(Set_Desired_Speed) := True;
            Unprocessed_Command := True;
          when 'R' | 'r' =>
            Command_Was_Received(Resume_Desired_Speed) := True;
            Unprocessed_Command := True;
          when 'O' | 'o' =>
            Command_Was_Received(Turn_Control_Off) := True;
            Unprocessed_Command := True;
          when 'B' | 'b' =>
            Command_Was_Received(Toggle_Brake) := True;
            Unprocessed_Command := True;
          when 'M' | 'm' =>
            Command_Was_Received(More_Gas) := True;
            Unprocessed_Command := True;
          when 'L' | 'l' =>
            Command_Was_Received(Less_Gas) := True;
            Unprocessed_Command := True;
          when 'Q' | 'q' =>
            raise Quit;
          when others =>
            -- Throw away invalid input keys.
            Unprocessed_Command := False;
        end case;
      end if;
      if Unprocessed_Command then
        Put_Line("Command = " & Key);
      end if;
    end if;
  end Update_Driver_Inputs;

  procedure Write_Header is
  begin
    Put_Line("Command syntax: <command_letter>[<repeat_count>]");
    Put_Line("S = Set Desired Speed");
    Put_Line("R = Resume Previous Speed");
    Put_Line("O = Turn Cruise Control Off");
    Put_Line("M = More Gas");
    Put_Line("L = Less Gas");
    Put_Line("B = Brake On/Off");
    Put_Line("D = Delay (no command for one cycle)");
    Put_Line("Q to quit");
    New_Line;
  end Write_Header;

begin
  Write_Header;
end Automobile_Interface;
with Aiu; use Aiu;
--pragma Elaborate_All (Aiu);

package Automobile_Interface is

--  pragma Elaborate_Body;


  -- Cruise control buttons.

  function Cruise_Off return Boolean;
  function Resume return Boolean;
  function Set_Speed return Boolean;


  -- Sensors.

  function Brake return Boolean;
  function Engine return Aiu.Rev_Per_Minute;
  function Speed return Aiu.Feet_Per_Second;


  -- Controls.

  procedure Throttle (Setting : Aiu.Millimeters);


  -- Displays.

  procedure Cruise_Control (On : Boolean; Speed : Aiu.Feet_Per_Second);
  -- Shows if the cruise control is on or off and what the cruising 
  -- speed is if it is on.

  procedure Speedometer (Speed : Aiu.Feet_Per_Second);
  


  -- Simulation.

  procedure Update_Driver_Inputs;
  -- Update_Driver_Inputs must be called every cycle.


  Quit : exception;
  -- Raised when the simulation is complete.
  -- (I.e., when 'Q' or 'q' is entered.)
    
end Automobile_Interface;
with Interfaces; use Interfaces;

package body Automobile_Interface_Units is

  function "*" 
      (Left : Feet_Per_Second;
       Right : Tu.Milliseconds) return Millimeters is
    Speed : Float;
    Time : Float;
    Distance : Float;
  begin
    Speed := Float(Dimensionless(Left));
    Time := Float(Tu.Dimensionless(Right));
    Distance := (Speed * 30.54)
      -- millimeters per second
        * (Time / 1000.0); -- seconds
    return Type_Convert(Integer_32(Distance));
  end "*";

end Automobile_Interface_Units;
with Di;
with Tu;

package Automobile_Interface_Units is

--  pragma Elaborate_Body;

  type Millimeters is new Di.Units;
  
  type Feet_Per_Second is new Di.Units;

  type Rev_Per_Minute is new Di.Units;

  function "*" 
      (Left : Feet_Per_Second;
       Right : Tu.Milliseconds) return Millimeters;

end Automobile_Interface_Units;
separate (Auto_Computer)
package body Diagnostics is

  -- This body displays the performance
  -- of a cruise control algorithm on
  -- the CRT screeen.

  use Tu;
  -- for Millisecond arithmetic

  package Float_Text_Io is new Float_Io(Float);
  use Float_Text_Io;

  Loop_Start_Time : Seconds;
  Loop_Stop_Time : Seconds;

  Available_Time : constant Seconds := Convert(Delta_T);

  procedure Open is
  begin
    null;
  end Open;

  procedure Start is
  begin
    Loop_Start_Time := Polled_Timer.Time_Used;
  end Start;

  procedure Stop is
    Used : Seconds;
  begin
    Loop_Stop_Time := Polled_Timer.Time_Used;
    Used := Loop_Stop_Time - Loop_Start_Time;
-- This is useless with a 1 second timer (gnat 2.04, OS/2).
--    Put("% CPU time used =");
--    Put(100.0 * Float(Dimensionless(Used)) / 
--        Float(Dimensionless(Available_Time)), Fore => 4, Aft => 2, Exp => 0);
--    New_Line;
  end Stop;

  procedure Close is
  begin
    null;
  end Close;

end Diagnostics;
separate (Auto_Computer)
procedure Do_Other_Stuff is
  -- A hook for other functions that 
  -- could be added later.
begin
  null;
  -- <optional miles per gallon,
  -- miles till empty, and so on,
  -- displayed here>
  -- <optional climate control here>
  -- <optional check of doors
  -- and lights here>
  -- <optional navigation here>
end Do_Other_Stuff;
with Ada.Text_Io; use Ada.Text_Io;
with Automobile_Interface;
pragma Elaborate_All(Automobile_Interface);
with Cruise;
pragma Elaborate_All(Cruise);
with Polled_Timer;
pragma Elaborate_All(Polled_Timer);

procedure Auto_Computer is 

  package Ai renames Automobile_Interface;

  package Body_Style_Cruise is
      new Cruise(Delta_T, Max_Throttle, Max_Rpm, Gain, Compensation);

  package Diagnostics is
    procedure Open;
    procedure Start;
    procedure Stop;
    procedure Close;
  end Diagnostics;

  package body Diagnostics is separate;  
  procedure Do_Other_Stuff is separate;

begin

  Diagnostics.Open;

  -- Start a timer that expires every Delta_T milliseconds.
  Polled_Timer.Open(
      --Change milliseconds to seconds
      Tu.Convert(Delta_T), Polled_Timer.Repeated);

  Polled_Timer.Start;

  loop

    while not Polled_Timer.Has_Expired loop
      null; 
    end loop;

    Diagnostics.Start;

    -- Update the simulation status of the car's controls.
    Ai.Update_Driver_Inputs;

    -- Control speed.
    Body_Style_Cruise.Adjust_Throttle;

    -- Display instruments.
    Ai.Speedometer(Ai.Speed);

    Ai.Cruise_Control
        (Body_Style_Cruise.Is_On,
         Body_Style_Cruise.Cruising_Speed);

    Do_Other_Stuff;

    Diagnostics.Stop;
    
  end loop;

exception
  when Ai.Quit =>
    Diagnostics.Close;
  when Program_Error =>
    Diagnostics.Close;
    Put_Line("Terminating due to Program_Error");
  when others =>
    Diagnostics.Close;
    Put_Line("Unexpected exception in Auto_Computer");
end Auto_Computer;
with Aiu;
with Interfaces; use Interfaces;
with Tu;
pragma Elaborate_All (Tu);

with Automobile_Interface;
--pragma Elaborate_All(Automobile_Interface);

separate (Auto_Simulator)
function Brake_Force return Asu.Lbs is
begin
  if Automobile_Interface.Brake then
    if Body_Speed /= +0.0 then
      declare
        Direction : Float;
      begin
        Direction := Dimensionless(Body_Speed) / abs(Dimensionless(Body_Speed));
        if +Direction /= Nan then
          return +(Direction * (Dimensionless(Max_Thrust)));
        end if;
      end;
    end if;
  end if;
  return +0.0;
end Brake_Force;
separate (Auto_Simulator)
function Drag return Asu.Lbs is
  Speed_Squared : Float;
begin
  Speed_Squared := Float(Dimensionless(Body_Speed))**2;
  return +(0.0343 * Speed_Squared);
end Drag;
separate (Auto_Simulator)
function Up_Hill return Asu.Lbs is 
begin
  return +0.0; -- Placeholder for now
end Up_Hill;
with Ada.Float_Text_Io; use Ada.Float_Text_Io;
with Ada.Integer_Text_Io; use Ada.Integer_Text_Io;
with Ada.Text_Io; use Ada.Text_Io;
with Df;
with Interfaces; use Interfaces;

package body Auto_Simulator is

  use Aiu;
  use Asu;

  type Fps is new Df.Units;
  -- Feet per second

  type Fps2 is new Df.Units;
  -- Feet per second squared

  Body_Speed : Fps;

  Coefficient_Of_Friction : constant Float := 0.7;

  Debug : constant Boolean := False;

  Max_Thrust : Asu.Lbs := Coefficient_Of_Friction * Car_Weight;

  function Divide (N, D : Float) return Float;
  -- Needed here by Plus_Infinity and Nan.

  procedure Push_Body_With (Thrust : Asu.Lbs);
  -- Needed here by Accelerate_Body.

  function Up_Hill return Asu.Lbs is separate;
  -- Needed here by Push_Body_With

  Nan : constant Asu.Lbs := Type_Convert(Divide(0.0, 0.0));

  Plus_Infinity : constant Asu.Lbs := Type_Convert(Divide(1.0, 0.0));

  function "/" (Left : Asu.Hp; Right : Fps) return Asu.Lbs is
    L, R : Float;
  begin
    -- 1 Hp = 550 ft-lbs/sec.
    L := Asu.Dimensionless(Left);
    R := Dimensionless(Right);
    return Asu.Type_Convert(550.0 * L / R);
  end "/";

  function "/" 
      (Left : Asu.Lbs; -- pounds of force
       Right : Asu.Lbs) -- pounds of mass
      return  Fps2 is
    L, R : Float;
  begin
    L := Asu.Dimensionless(Left);
    R := Asu.Dimensionless(Right);
    -- One pound of force will accelerate one pound of mass
    -- at 32.2 feet per second squared.
    return Type_Convert(32.2 * L / R);
  end "/";

  function "*" (Left : Fps2; Right : Tu.Milliseconds) return Fps is
    L, R : Float;
  begin
    L := Dimensionless(Left);
    R := Float(Tu.Dimensionless(Right));
    return Type_Convert(L * R / 1000.0);
  end "*";

  procedure Accelerate_Body (Setting : Aiu.Millimeters) is
    Throttle_Fraction : Float;
    Thrust : Asu.Lbs;
    Power : Asu.Hp;
  begin
    if Debug then
      Put("Accelerator setting = ");
      Put(Integer(Dimensionless(Setting)));
      New_Line;
    end if;
    -- Compute horsepower created by engine.
    Throttle_Fraction := Float(Aiu.Dimensionless(Setting)) / Float(Aiu.Dimensionless(Max_Throttle));
    Power := Throttle_Fraction * Max_Horsepower;
    if Debug then
      Put("Power = ");
      Put(Float(Dimensionless(Power)));
      New_Line;
    end if;
    if Power = +0.0 then
      Thrust := +0.0;
    else
      -- Convert that to thrust applied to the body.
      -- Evaluation may raise Numeric_Error or Constraint_Error
      -- *** but not in OS/2 gnat! ***
      Thrust := Power / Body_Speed;
    end if;
    if Thrust = Plus_Infinity then
      Thrust := Throttle_Fraction * Max_Thrust;
    elsif Thrust > Max_Thrust then
      Thrust := Max_Thrust;
    elsif Thrust = Nan or else Thrust < +0.0 then
      Thrust := +0.0;
    end if;
    if Debug then
      Put("Thrust = ");
      Put(Float(Dimensionless(Thrust)));
      New_Line;
    end if;
    -- Move the car with it.
    Push_Body_With(Thrust);
  exception
    when Numeric_Error | Constraint_Error =>
      Thrust := +0.0;
      Push_Body_With(Thrust);
    when others =>
      Put_Line("Unexpected exception in Accelerate_Body");
      raise;
  end Accelerate_Body;

  function Brake_Force return Asu.Lbs is separate;

  function Current_Speed return Aiu.Feet_Per_Second is
    Speed : Integer_32;
  begin
    Speed := Integer_32(Dimensionless(Body_Speed));
    return Aiu.Type_Convert(Speed);
  end Current_Speed;

  -- To work around non-overflowing IEEE arithmetic.
  -- (Static expressions that evaluate to +/- infinity or NaN
  -- are rejected.)
  function Divide (N, D : Float) return Float is
  begin
    return N / D;
  end Divide;

  function Drag return Asu.Lbs is separate;


  procedure Push_Body_With (Thrust : Asu.Lbs) is
    Acceleration : Fps2;
    Delta_Speed : Fps;
  begin
    Acceleration := (Thrust - Up_Hill - Drag - Brake_Force) / Car_Weight;
    -- Simple integrator.
    Delta_Speed := Acceleration * Delta_T;
    if Debug then
      if Acceleration /= +0.0 then
        Put("Acceleration = ");
        Put(Float(Dimensionless(Acceleration)));
        New_Line;
      end if;
      if Thrust /= +0.0 then
        Put("Thrust = ");
        Put(Float(Dimensionless(Thrust)));
        New_Line;
      end if;
      if Brake_Force /= +0.0 then
        Put("Brake_Force = ");
        Put(Float(Dimensionless(Brake_Force)));
        New_Line;
      end if;
      if Delta_Speed /= +0.0 then
        Put("Delta_Speed = ");
        Put(Float(Dimensionless(Delta_Speed)));
        New_Line;
      end if;
    end if;
    Body_Speed := Body_Speed + Delta_Speed;
  end Push_Body_With;

begin
  Body_Speed := +0.0;
end Auto_Simulator;

with Df;

package Auto_Simulator_Units is

  type Hp is new Df.Units;
  
  type Lbs is new Df.Units;

end Auto_Simulator_Units;
with Automobile_Interface_Units;
use Automobile_Interface_Units;
--pragma Elaborate_All (Automobile_Interface_Units);

package Compensation_Memory is
  Last_Speed : Automobile_Interface_Units.Feet_Per_Second := +0;
end Compensation_Memory;
with Automobile_Interface;
--pragma Elaborate_All (Automobile_Interface);
with Df;

package body Cruise is

  package Ai renames Automobile_Interface;

  use Aiu;
  use Tu;

  type Gs is new Df.Units;

  -- Non-volatile variables
  Cruise_Control_On : Boolean;
  Desired_Speed : Aiu.Feet_Per_Second;
  Last_Speed : Aiu.Feet_Per_Second;
  Throttle_Setting : Aiu.Millimeters;

  function Wheels_Spinning return Boolean;
  -- Needed by Adjust_Throttle.

  function "/" 
      (Left : Aiu.Feet_Per_Second; Right : Tu.Milliseconds) return Gs is
    Speed : Float;
    Time : Float;
    Feet_Per_Second_Squared : Float;
  begin
    Speed := Float(Dimensionless(Left));
    Time := Float(Dimensionless(Right)) / 1000.0;
    Feet_Per_Second_Squared := Speed / Time;
    return +(Feet_Per_Second_Squared / 32.2);
  end "/";

  function Integrate (Speed : Aiu.Feet_Per_Second) return Aiu.Millimeters is
    Integrated_Error : Aiu.Millimeters;
    Adjustment : Aiu.Millimeters;
  begin
    Integrated_Error := Speed * Delta_T;
    Adjustment := Integrator_Gain * Integrated_Error;
    return Throttle_Setting + Adjustment;
  end Integrate;

  procedure Adjust_Throttle is
    Perceived_Speed : Aiu.Feet_Per_Second;
  begin

    Perceived_Speed := Feedback_Compensated(Ai.Speed);

    -- Check driver's controls.
    if Ai.Set_Speed then
      Desired_Speed := Perceived_Speed;
      Cruise_Control_On := True;
    end if;

    if Ai.Resume then
      Cruise_Control_On := True;
    end if;

    if Ai.Brake or Ai.Cruise_Off then
      Cruise_Control_On := False;
      -- Note:  Could set Desired_Speed to 0 here if desired.
    end if;

    -- Safety checks.
    if Ai.Engine > Max_Rpm then
      Cruise_Control_On := False;
    end if;
    if Wheels_Spinning then
      Cruise_Control_On := False;
    end if;

    if Cruise_Control_On then
      Throttle_Setting := Integrate(Desired_Speed - Perceived_Speed);
      -- Throttle has physical limits.
      if Throttle_Setting < +0 then
        Throttle_Setting := +0;
      elsif Throttle_Setting > Max_Throttle then
        Throttle_Setting := Max_Throttle;
      end if;
    else
      -- Throttle_Setting = 0 when CC is off.
      Throttle_Setting := +0; 
    end if;

    Ai.Throttle(Throttle_Setting);

  exception
    when Ai.Quit => -- Quit is expected to occur.
      raise;
  end Adjust_Throttle;

  function Cruising_Speed return Aiu.Feet_Per_Second is
  begin
    return Desired_Speed;
  end Cruising_Speed;

  function Is_On return Boolean is
  begin
    return Cruise_Control_On;
  end Is_On;

  function Wheels_Spinning return Boolean is
    Acceleration : Gs;
    New_Speed : Aiu.Feet_Per_Second;
  begin
    New_Speed := Ai.Speed;
    Acceleration := (New_Speed - Last_Speed) / Delta_T;
    Last_Speed := New_Speed;
    -- Wheels must be spinning if apparent acceleration
    -- is greater than 1 G.
    return Acceleration > +1.0;
  end Wheels_Spinning;

begin
  Cruise_Control_On := False;
  Desired_Speed := +0;
  Last_Speed := +0;
  Throttle_Setting := +0;
end Cruise;
with Aiu;
with Interfaces; use Interfaces;
with Tu;
--pragma Elaborate_All(Aiu);
--pragma Elaborate_All(Tu);

package Cruise is

--  pragma Elaborate_Body;

  procedure Adjust_Throttle;
  -- Controls the throttle if the Cruise Control unit is enabled.

  function Is_On return Boolean;
  -- True if cruise control is enabled.

  function Cruising_Speed return Aiu.Feet_Per_Second;
  -- Returns the desired cruising speed.

end Cruise;
with Dim_Float;
package Df renames Dim_Float;
with Dim_Int;
package Di renames Dim_Int;
with Float_Units;
with Interfaces; use Interfaces;

package Dim_Float is new Float_Units
    (Float_Type => Float,
     Integer_Type => Integer_32);
with Integer_Units;
with Interfaces; use Interfaces;

package Dim_Int is new Integer_Units(Integer_32);
package body FLOAT_UNITS is

  -- This version is specialized for non-overflowing IEEE arithmetic.

  function Float_Type_Divide (L, R : Float_Type'Base) return Float_Type'Base;
  function Units_Divide (L, R : Units) return Units;

  Plus_Infinity_Float : constant Float_Type'Base := Float_Type_Divide(1.0, 0.0); 
  Minus_Infinity_Float : constant Float_Type'Base := Float_Type_Divide(-1.0, 0.0); 
  Nan_Float : constant Float_Type'Base := Float_Type_Divide(0.0, 0.0); 

  Plus_Infinity_Units : constant Units := Units_Divide(1.0, 0.0); 
  Minus_Infinity_Units : constant Units := Units_Divide(Units(Float_Type'(-1.0)), 0.0); 
  Nan_Units : constant Units := Units_Divide(0.0, 0.0); 

  function Float_Type_Divide (L, R : Float_Type'Base) return Float_Type'Base is
  begin
    return L / R;
  end Float_Type_Divide;

  function Units_Divide (L, R : Units) return Units is
  begin
    return L / R;
  end Units_Divide;

  function Type_Convert(X : Float_type) return Units is
  begin
    if X = Plus_Infinity_Float then
      return Plus_Infinity_Units;
    elsif X = Minus_Infinity_Float then
      return Minus_Infinity_Units;
    elsif X = Nan_Float then
      return Nan_Units;
    else
      return Units(X);
    end if;
  end Type_Convert;

  function "+"(X : Float_type) return Units is
  begin
    return Units(X);
  end "+";

  function "-"(X : Float_type) return Units is
    NEG : Float_Type'Base;
  begin
    NEG := -X;
    return Units(NEG);
  end "-";

  function "+"(RIGHT : Units)
    return Units is
  begin
    return RIGHT;
  end "+";

  function "-"(RIGHT : Units)
    return Units is
      NEG : Float_Type'Base;
  begin
    NEG := -1.0 * Float_Type'Base(RIGHT);
    return Units(NEG);
  end "-";

  function "abs"(RIGHT : Units)
    return Units is
  begin
    return Units(abs(Float_type(RIGHT)));
  end "abs";

  function "+"(LEFT, RIGHT : Units)
    return Units is
  begin
    return Units(Float_Type'Base(LEFT) + Float_Type'Base(RIGHT));
  end "+";

  function "-"(LEFT, RIGHT : Units)
    return Units is
  begin
    return Units(Float_Type'Base(LEFT) - Float_Type'Base(RIGHT));
  end "-";

  function "*"(LEFT : Float_Type'Base; RIGHT : Units)
    return Units is
  begin
    return Units(LEFT * Float_Type'Base(RIGHT));
  end "*";

  function "*"(LEFT : Units; RIGHT : Float_Type'Base)
    return Units is
  begin
    return Units(Float_Type'Base(LEFT) * RIGHT);
  end "*";

  function "/"(LEFT : Units; RIGHT : Float_Type'Base)
    return Units is
  begin
    return Units(Float_Type'Base(LEFT) / RIGHT);
  end "/";

  function "/"(LEFT, RIGHT : Units)
    return Float_Type'Base is
  begin
    return Float_Type'Base(LEFT) / Float_Type'Base(RIGHT);
  end "/";

  function "<"(LEFT, RIGHT : Units)
    return boolean is
  begin
    return Float_Type'Base(LEFT) < Float_Type'Base(RIGHT);
  end "<";

  function "<="(LEFT, RIGHT : Units)
    return boolean is
  begin
    return Float_Type'Base(LEFT) <= Float_Type'Base(RIGHT);
  end "<=";

  function ">"(LEFT, RIGHT : Units)
    return boolean is
  begin
    return Float_Type'Base(LEFT) > Float_Type'Base(RIGHT);
  end ">";

  function ">="(LEFT, RIGHT : Units)
    return boolean is
  begin
    return Float_Type'Base(LEFT) >= Float_Type'Base(RIGHT);
  end ">=";

  function "/"(LEFT, RIGHT : Units) return Integer_Type'Base is
    -- divide and truncate toward zero
    EXACT : Float_Type'Base;
    ROUNDED, TRUNCATED : Integer_Type'Base;
  begin
    EXACT := LEFT / RIGHT;
    ROUNDED := Integer_Type'Base(EXACT);
    if Float_Type'Base(abs(ROUNDED)) > abs(EXACT) then
      if ROUNDED > 0 then
        TRUNCATED := ROUNDED-1;
      else
        TRUNCATED := ROUNDED+1;
      end if;
    else
      TRUNCATED := ROUNDED;
    end if;
    return TRUNCATED;
  end "/";

  function "rem"(LEFT, RIGHT : Units) return Units is
    COMPLETE_CYCLES : Integer_Type'Base;
    REMAINDER       : Units;
  begin
    COMPLETE_CYCLES := LEFT / RIGHT;
    REMAINDER := LEFT - Float_Type'Base(COMPLETE_CYCLES) * RIGHT;
    return REMAINDER;
  end "rem";

  function "mod"(LEFT, RIGHT : Units) return Units is
    REMAINDER : Units;
  begin
    REMAINDER := LEFT rem RIGHT;
    if LEFT * RIGHT > 0.0 then
      return REMAINDER;
    elsif REMAINDER = 0.0 then
      return REMAINDER;
    else
      return REMAINDER + RIGHT;
    end if;
  end "mod";

  function Dimensionless(LEFT : Units)
    return Float_Type'Base is
  begin
    return Float_Type'Base(LEFT);
  end Dimensionless;

  function First return Float_Type'Base is
  begin
    return MIN;
  end First;

  function Last return Float_Type'Base is
  begin
    return MAX;
  end Last;

begin
  if Float_Type'Machine_Overflows then
    raise Program_Error;
  end if;
end FLOAT_UNITS;
package FLOAT_UNITS is

--  pragma Elaborate_Body;

  -- This package provides useful parent types for derived
  -- dimensional units. That is, it makes it possible to
  -- do this:

  -- package DIMENSIONAL_FLOAT is new FLOAT_UNITS
  --  (Float, Integer_Type => Interfaces.Integer_32);
  -- type Feet is new DIMENSIONAL_FLOAT.Units;

  -- package TEMPERATURE is new FLOAT_UNITS
  --  (Float_Type => Float, MIN => -273.16, MAX => 32000.0,
  --  Integer_Type => integer);
  -- type Celsius is new TEMPERATURE.Units;

  -- Objects of type Feet can be added to other objects of
  -- type feet to produce a result in Feet, but Ada will not
  -- allow you to erroneously multiply objects of type Feet
  -- together and assign the result to another object of
  -- type Feet. Ada will raise CONSTRAINT_ERROR if you try
  -- to assign a value colder than absolute zero to an
  -- object of type Celsius.

  type Units is private;

  -- Because Units are PRIVATE types, Ada can't be sure they
  -- are real numbers. Therefore, she will not let you do
  -- this:

    -- TARGET_RANGE : Feet := 587.0;

  -- Since you have to be able to assign numeric values to
  -- objects of type Feet, I have given you three type
  -- conversion routines.

  -- The preferred way is to do this:

    -- TARGET_RANGE := Feet := Type_Convert(587.0);

  -- This makes it obvious to other programmers that you
  -- have converted a numeric literal to a special kind of
  -- data type.

  -- But if you do this a lot, your program will be filled
  -- with so many Type_Convert() statements that people will
  -- get so used to them that they will ignore them. They
  -- lose their value, so you might as well use an operator
  -- which is 11 characters shorter.

    -- TARGET_RANGE : Feet := +587.0; -- or -587.0;

  function Type_Convert(X : Float_Type'Base) return Units;
  function "+"(X : Float_Type'Base) return Units;
  function "-"(X : Float_Type'Base) return Units;

  function "+"(RIGHT : Units)
    return Units;
  function "-"(RIGHT : Units)
    return Units;
  function "abs"(RIGHT : Units)
    return Units;
  function "+"(LEFT, RIGHT : Units)
    return Units;
  function "-"(LEFT, RIGHT : Units)
    return Units;
  function "*"(LEFT : Float_Type'Base; RIGHT : Units)
    return Units;
  function "*"(LEFT : Units; RIGHT : Float_Type'Base)
    return Units;
  function "/"(LEFT : Units; RIGHT : Float_Type'Base)
    return Units;
  function "/"(LEFT, RIGHT : Units)
    return Float_Type'Base;
  -- "=" and "/=" are already defined for private types
  function "<"(LEFT, RIGHT : Units)
    return boolean;
  function "<="(LEFT, RIGHT : Units)
    return boolean;
  function ">"(LEFT, RIGHT : Units)
    return boolean;
  function ">="(LEFT, RIGHT : Units)
    return boolean;

  -- The modulo operation for Units is provided to make
  -- it easy to normalize angular measurements.
  -- The division operator for Units which returns integers
  -- truncates toward zero (rather than rounding) to make it
  -- consistent with integer division, and it lets you do
  -- this:

  -- type Radians is new DIMENSIONAL_FLOAT.Units;

  -- PI           : constant Radians := 3.14159;
  -- ANGLE        : Radians;
  -- REVOLUTIONS  : integer; -- (if Integer_Type is integer)

  -- ANGLE := ANGLE mod (2.0 * PI);
  -- REVOLUTIONS := ANGLE / (2.0 * PI);

  function "/"(LEFT, RIGHT : Units)
    return Integer_Type'Base; -- truncates toward zero
  function "rem"(LEFT, RIGHT : Units)
    return Units;
  function "mod"(LEFT, RIGHT : Units)
    return Units;

  -- Suppose you want to multiply two objects of type Feet
  -- together to get an answer in Square_feet. You need to
  -- write your own multiplication routine which converts
  -- the objects to dimensionless numbers, multiplies them,
  -- and converts them to type Square_feet. You are given
  -- the Dimensionless function to do this.

    -- function "*"(LEFT,RIGHT : Feet) return Square_feet is
    --   L, R : float; -- or whatever Float_Type is
    -- begin
    --   L := Dimensionless(LEFT);
    --   R := Dimensionless(RIGHT);
    --   return Type_Convert(L * R);
    -- end "*";

  -- The Dimensionless function is also handy for printing
  -- values. Assuming you have instantiated Float_Io for
  -- Float_Type and called it Float_Type_Io, you can do
  -- this:

    -- Float_Type_Io.Put(Dimensionless(TARGET_RANGE));

  function Dimensionless(LEFT : Units)
    return Float_Type'Base;

  -- Since private types (and the underlying real type)
  -- don't have 'FIRST and 'LAST attributes, I have given
  -- you functions that will act like those attributes.

  function First return Float_Type'Base;
  function Last return Float_Type'Base;
    -- For example:
    --    ABSOLUTE_ZERO : constant Celsius
    --      := Type_Convert(TEMPERATURE.First);

private

  type Units is new Float_Type'Base range MIN..MAX;

end FLOAT_UNITS;
package body Integer_Units is

  function Type_Convert (X : Integer_Type'Base) return Units is
  begin
    return Units(X);
  end Type_Convert;

  function "+" (Right : Integer_Type) return Units is
  begin
    return Units(Right);
  end "+";

  function "-" (Right : Integer_Type) return Units is
    Neg : Integer_Type'Base;
  begin
    Neg := -1 * Right;
    return Units(Neg);
  end "-";

  function "+" (Right : Units)
    return Units is
  begin
    return Right;
  end "+";

  function "-" (Right : Units)
    return Units is
      Neg : Integer_Type'Base;
  begin
    Neg := -1 * Integer_Type'Base(Right);
    return Units(Neg);
  end "-";

  function "abs" (Right : Units)
    return Units is
  begin
    return Units(abs(Integer_Type'Base(Right)));
  end "abs";

  function "+" (Left, Right : Units)
    return Units is
  begin
    return Units(Integer_Type'Base(Left) + Integer_Type'Base(Right));
  end "+";

  function "-" (Left, Right : Units)
    return Units is
  begin
    return Units(Integer_Type'Base(Left) - Integer_Type'Base(Right));
  end "-";

  function "*" (Left : Integer_Type'Base; Right : Units)
    return Units is
  begin
    return Units(Left * Integer_Type'Base(Right));
  end "*";

  function "*" (Left : Units; Right : Integer_Type'Base)
    return Units is
  begin
    return Units(Integer_Type'Base(Left) * Right);
  end "*";

  function "/" (Left : Units; Right : Integer_Type'Base)
    return Units is
  begin
    return Units(Integer_Type'Base(Left) / Right);
  end "/";

  function "/" (Left, Right : Units)
    return Integer_Type'Base is
  begin
    return Integer_Type'Base(Left) / Integer_Type'Base(Right);
  end "/";

  function "/" (Left, Right : Units)
    return Float is
      Exact_Quotient : Float;
  begin
    Exact_Quotient := Float(Left) / Float(Right);
    return Exact_Quotient;
  end "/";

  function "rem" (Left, Right : Units)
    return Units is
  begin
    return Units
      (Integer_Type'Base(Left) rem Integer_Type'Base(Right));
  end "rem";

  function "mod" (Left, Right : Units)
    return Units is
  begin
    return Units
      (Integer_Type'Base(Left) mod Integer_Type'Base(Right));
  end "mod";

  function Dimensionless (Left : Units)
    return Integer_Type'Base is
  begin
    return Integer_Type'Base(Left);
  end Dimensionless;

  function "<" (Left, Right : Units)
    return Boolean is
  begin
    return Integer_Type'Base(Left) < Integer_Type'Base(Right);
  end "<";

  function "<=" (Left, Right : Units)
    return Boolean is
  begin
    return Integer_Type'Base(Left) <= Integer_Type'Base(Right);
  end "<=";

  function ">" (Left, Right : Units)
    return Boolean is
  begin
    return Integer_Type'Base(Left) > Integer_Type'Base(Right);
  end ">";

  function ">=" (Left, Right : Units)
    return Boolean is
  begin
    return Integer_Type'Base(Left) >= Integer_Type'Base(Right);
  end ">=";

  function First return Integer_Type'Base is
  begin
    return Min;
  end First;

  function Last return Integer_Type'Base is
  begin
    return Max;
  end Last;

end Integer_Units;
package Integer_Units is

--  pragma Pure;

  -- This package provides useful parent types for derived
  -- dimensional units. That is, it makes it possible to
  -- do this:

  -- package Dimensional_Integer is new
  --   Integer_Units(integer);
  -- type Feet is new Dimensional_Integer.Units;

  -- package Temperature is new Integer_Units
  --  (Integer_Type => Interfaces.Integer_16;
  --   Min => -273;
  --   Max => Interfaces.Integer_16'Last);
  --   -- Absolute Zero is -273 degrees C.
  -- type Celsius is new Temperature.Units;

  -- Objects of type Feet can be added to other objects of
  -- type Feet to produce a result in Feet, but Ada will not
  -- allow you to erroneously multiply objects of type Feet
  -- together and assign the result to another object of
  -- type Feet. Ada will raise Constraint_Error if you try
  -- to assign a value colder than absolute zero to an
  -- object of type Celsius.

  type Units is private;

  -- Because Units are private types, Ada can't be sure they
  -- are integers. Therefore, she will not let you do this:

    -- Target_Range : Feet := 587;

  -- Since you have to be able to assign numeric values to
  -- objects of type Feet, I have given you three type
  -- conversion routines.

  -- The preferred way is to do this:

    -- Target_Range : Feet := Type_Convert(587);

  -- This makes it obvious to other programmers that you
  -- have converted a numeric literal to a special kind of
  -- data type.

  -- But if you do this a lot, your program will be filled
  -- with so many Type_Convert() statements that people will
  -- get so used to them that they will ignore them. They
  -- loose their value, so you might as well use an operator
  -- which is 11 characters shorter.

    -- Target_Range : Feet := +587; -- or -587;

  function Type_Convert (X : Integer_Type'Base) return Units;
  function "+" (Right : Integer_Type'Base) return Units;
  function "-" (Right : Integer_Type'Base) return Units;

  -- The derived data types will inherit all the operations
  -- in the package below. These are all the operations that
  -- make sense for dimensional quantities.

  function "+" (Right : Units)
    return Units;
  function "-" (Right : Units)
    return Units;
  function "abs" (Right : Units)
    return Units;
  function "+" (Left, Right : Units)
    return Units;
  function "-" (Left, Right : Units)
    return Units;
  function "*" (Left : Integer_Type'Base; Right : Units)
    return Units;
  function "*" (Left : Units; Right : Integer_Type'Base)
    return Units;
  function "/" (Left : Units; Right : Integer_Type'Base)
    return Units;
  function "/" (Left, Right : Units)
    return Integer_Type'Base;
  function "/" (Left, Right : Units)
    return Float;
  function "rem" (Left, Right : Units)
    return Units;
  function "mod" (Left, Right : Units)
    return Units;
  -- "=" and "/=" are already defined for private types
  function "<" (Left, Right : Units)
    return Boolean;
  function "<=" (Left, Right : Units)
    return Boolean;
  function ">" (Left, Right : Units)
    return Boolean;
  function ">=" (Left, Right : Units)
    return Boolean;

  -- Suppose you want to multiply two objects of type Feet
  -- together to get an answer in Square_feet. You need to
  -- write your own multiplication routine which converts
  -- the objects to dimensionless integers, multiplies them,
  -- and converts them to type Square_feet. You are given
  -- the Dimensionless function to do this.

      -- function "*"(Left,Right : Feet)
      --   return Square_feet is
      --   L, R : integer; -- or Integer_32 or whatever
      --                   -- type you used to instantiate
      --                   -- this package.
      -- begin
      --   L := Dimensionless(Left);
      --   R := Dimensionless(Right);
      --   return Type_Convert(L * R);
      -- end "*";

  -- The Dimensionless function is also handy for printing
  -- values. If you instantiated this package for type
  -- Integer_X, (where X is 8, 16, or 32) you can do this:

      -- Put(Integer_X'Image(Dimensionless(Target_Range)));

  function Dimensionless(Left : Units)
    return Integer_Type'Base;

  -- Since private types don't have 'Fisrt and 'Last
  -- attributes, I have given you functions that
  -- will act like those attributes.

  function First return Integer_Type'Base;
  function Last return Integer_Type'Base;
    -- For example,
    --  Absolute_Zero : constant Celsius
    --    := Type_Convert(Temperature.First);

private

  type Units is new Integer_Type'Base range Min..Max;

end Integer_Units;
with Text_Io; use Text_Io;
package body Keyboard is
-- This version takes its input from a file 
-- named "kbdinput.dat" (which must be
-- present in the execution (sub)directory),
-- instead of the keyboard.
-- Command syntax: <character>[<repeat_count>]
-- 'D' or 'd' are not passed to the caller but are
-- skipped (they represent a one call 'delay' to 
-- Keyboard.Hit).  Encountering an end of file is 
-- represented by the value 'q' (for "quit").

  package Int_Io is new Integer_Io(Integer);
  use Int_Io;

  Available : Boolean := False;
  File : File_Type;
  Repeat_Count : Integer := 0;
  Value : Character := 'q';

  procedure Check_Input is
  begin
    if Repeat_Count > 0 then
      Repeat_Count := Repeat_Count - 1;
      if Value /= 'D' and then Value /= 'd' then
        Available := True;
      end if;
      return;
    end if;
    if not Available and then Repeat_Count = 0 then
      if not End_Of_File(File) then
        Text_Io.Get(File, Value);
        if not End_Of_Line(File) then
          begin
            Get(File, Repeat_Count);
          exception
            when others =>
              Repeat_Count := 1;
          end;
        else
          Repeat_Count := 1;
        end if; -- not EOF after Get
        if Repeat_Count <= 0 then 
          Repeat_Count := 1;
        end if;
        if Value = 'D' or Value = 'd' then
          return;
        else
          -- Make the first repeat available.
          Repeat_Count := Repeat_Count - 1;
          Available := True;
          return;
        end if;
      else
        Value := 'q';
      end if;
      Available := True;
    end if; -- not Available and then Repeat_Count = 0
  end Check_Input;

  function Hit return Boolean is
  begin
    Check_Input;
    return Available;
  end Hit;

  function Char return Character is
  begin
    if Available then
      Available := False;
      return Value;
    else
      Text_Io.Put_Line("Keyboard.Char called when not Keyboard.Hit");
    end if;
  end Char;

begin
  Open(File, In_File, ".\kbdinput.dat");
exception
  when Status_Error |
       Mode_Error |
       Name_Error |
       Use_Error |
       Device_Error =>
    Put_Line("Can't open the keyboard input file.");
    raise;
  when others =>
    Put_Line("others exception from Open");
    raise;
end Keyboard;
package Keyboard is

  pragma Elaborate_Body;

  function Hit return Boolean;
  -- Returns True if a key has been struck.
  -- Hit must be called until it returns True,
  -- indicating that a character is available.
  -- then Char must be called to consume and 
  -- retrieve the value of the key.

  function Char return Character;
  -- If a character is available (as determined by
  -- a preceding call to Hit), returns its value.
  -- Otherwise, it raises Program_Error.

end Keyboard;
with Aiu;
with Asu;
with Auto_Simulator;
with Tu;

-- The following use clauses make the various "+" operators 
-- which are used below visible:

use Aiu;
-- for Millimeters (Max_Throttle);

use Asu;
-- for Hp (Max_Horsepower)
-- for Lbs (Car_Weight)

use Tu;
-- for Milliseconds (Delta_T)

package Mid_Sized_Auto is new Auto_Simulator
    (Delta_T => +1000, 
     Max_Throttle => +100,
     Max_Horsepower => +98.0,
     Car_Weight => +2692.0);
     -- including 200 lb driverwith Aiu;
with Auto_Computer;
with Compensation;
with Tu;

-- The following use clauses make the various "+" operators 
-- which are used below visible:

use Aiu;
  -- for Millimeters (Max_Throttle)
  -- for Rev_Per_Minute (Max_Rpm)

use Tu;
  -- for Milliseconds (Delta_T)

procedure Mid_Size_Auto_Computer is
  new Auto_Computer
      (Delta_T => +1000,
       Max_Throttle => +100,
       Max_Rpm => +4000,
       Gain => 1,
       Compensation => Compensation);
-- This body is for longer time periods on the IBM PC.

with CALENDAR;
package body POLLED_TIMER is

--  CALENDAR_TICK : float := 65_536.0 / 1_193_180.0;
  CALENDAR_TICK : float := 1.0; -- Modified for gnat 2.x under OS/2.
  CALENDAR_MAX  : float := 86_400.0;

  TIMER_PERIOD      : Duration;
  TIMER_MODE        : Modes;
  START_TIME, DONE  : CALENDAR.Time;
  TIME_ALREADY_USED : Duration;
  STOPPED           : boolean;

  procedure Open(PERIOD : Seconds;
                 MODE   : Modes) is
  begin
    if Dimensionless(PERIOD) > CALENDAR_MAX then
      raise INVALID_PERIOD;
    end if;
    if Dimensionless(PERIOD) < CALENDAR_TICK then
      raise INVALID_PERIOD;
    end if;
    TIMER_PERIOD := Duration(Dimensionless(PERIOD));
    TIME_ALREADY_USED := 0.0;
    STOPPED := TRUE;
    TIMER_MODE   := MODE;
  end Open;

  procedure Close is
  begin
    null;
  end Close;

  procedure Start is
    use CALENDAR; -- for "+" and "-"
  begin
    START_TIME := CALENDAR.Clock;
    DONE := START_TIME + TIMER_PERIOD
      - TIME_ALREADY_USED;
    STOPPED := FALSE;
  end Start;

  procedure Restart is
  begin
    TIME_ALREADY_USED := 0.0;
    Start;
  end Restart;

  function Has_Expired return boolean is
    use CALENDAR; -- for ">=" and "+"
  begin
    if Dimensionless(Time_Left) > 0.0 then
      return FALSE;
    else
      case TIMER_MODE is
        when SINGLE =>
          Stop;
          return TRUE;
        when REPEATED =>
          TIME_ALREADY_USED := CALENDAR.Clock - DONE;
          Start;
          return TRUE;
      end case;
    end if;
  end Has_Expired;

  procedure Stop is
    use CALENDAR; -- for "-"
  begin
    TIME_ALREADY_USED := CALENDAR.Clock
      - START_TIME;
    STOPPED := TRUE;
  end Stop;

  function Time_Used return Seconds is
    TIME : float;
    use CALENDAR; -- for "-"
  begin
    if STOPPED then
      TIME := float(TIME_ALREADY_USED);
    else
      TIME := float(TIMER_PERIOD
         - (DONE - CALENDAR.Clock));
    end if;
    return Type_Convert(TIME);
  end Time_Used;

  function Time_Left return Seconds is
    DIFFERENCE : float;
    use CALENDAR; -- for "-"
  begin
    if STOPPED then
      DIFFERENCE := float(TIMER_PERIOD - TIME_ALREADY_USED);
    else
      DIFFERENCE := float(DONE - CALENDAR.Clock);
    end if;
    return Type_Convert(DIFFERENCE);
  end Time_Left;

  function Max_Period return Seconds is
  begin
    return Type_Convert(CALENDAR_MAX);
  end Max_Period;

  function Single_Tick return Seconds is
  begin
    return Type_Convert(CALENDAR_TICK);
  end Single_Tick;

end POLLED_TIMER;
with TIME_UNITS;
use  TIME_UNITS;
--   Seconds

package POLLED_TIMER is

--  pragma Elaborate_Body;

  type Modes is (SINGLE, REPEATED);
    -- Program the timer to expire once, then stop (SINGLE),
    -- or program it to start another timing cycle as soon
    -- as the first one finishes (REPEATED).

  procedure Open(PERIOD : Seconds;
                 MODE   : Modes);
    -- Initializes the timer.
    -- It may raise INVALID_PERIOD.  If it does, use the two
    -- functions at the end of this package to find out what
    -- went wrong.

  INVALID_PERIOD : exception;

  procedure Close;
    -- Turns the timer off.  On some target systems (like
    -- the IBM PC), it restores the physical timer chip to
    -- its default condition.  (On the IBM PC, the
    -- POLLED_TIMER uses the timer chip normally used by the
    -- speaker for timing.  If you don't call Close at the
    -- end of the application program, then subsequent
    -- programs may not be able to use the speaker.)

  procedure Start;
    -- Starts the timer running in the MODE determined by
    -- Set.

  procedure Restart;
    -- Reloads the counter with the desired time delay and
    -- starts the counter again.

  function Has_Expired return boolean;
    -- Poll this function to see if the time delay has
    -- expired yet.

  procedure Stop;
    -- Holds the timer at its current count until started or
    -- restarted again.

  function Time_Used return Seconds;
    -- Tells you how many seconds have gone by since you
    -- started the timer.  It does NOT stop the timer.

  function Time_Left return Seconds;
    -- Tells you how many seconds until the time delay will
    -- expire.  It does NOT stop the timer.

  -- The following two functions tell the limitations of the
  -- underlying physical timer. They can be called before
  -- setting the period to make sure the timer can support
  -- the desired time resolution, or they can be called
  -- after Set raises INVALID_PERIOD to find out what went
  -- wrong.

  function Max_Period return Seconds;
  function Single_Tick return Seconds;

end POLLED_TIMER;
with Ada.Integer_Text_Io; use Ada.Integer_Text_Io;
with Interfaces; use Interfaces;

package body Time_Units is

  Twelve_Hours : constant Milliseconds
      := Type_Convert(43_200_000);
        -- (12 * 60 * 60 * 1000)

  One_Day : constant Milliseconds
      := 2 * Twelve_Hours;

  function Image(Time : Milliseconds)
    return Time_Strings is
      T, Hours, Min, Sec, Msec
        : Integer_32;
      Time_String : Time_Strings
        := "HH:MM:SS.TTT";
        --  123456789012  (to help count positions)
  begin
    T := Dimensionless(Time);
    Hours := T / 3600000;
    T := T mod 3600000;
    Min := T / 60000;
    T := T mod 60000;
    Sec := T / 1000;
    Msec := T mod 1000;

    Put(Time_String(1 .. 2), Integer(Hours));
    Put(Time_String(4 .. 5), Integer(Min));
    Put(Time_String(7 .. 8), Integer(Sec));
    Put(Time_String(10 .. 12), Integer(Msec));

    return Time_String;

  end Image;

  function Value(Time : Time_strings)
    return Milliseconds is
      T, Hours, Min, Sec, Msec
        : Integer_32;
  begin
    Hours := Integer_32'Value(Time(1..2));
    Min   := Integer_32'Value(Time(4..5));
    Sec   := Integer_32'Value(Time(7..8));
    Msec  := Integer_32'Value(Time(10..12));

    T := Msec + 1000 * (Sec + 60 * (Min + 60 * Hours));

    return Type_Convert(T);

  end Value;

  function Convert(Time : Milliseconds)
    return Seconds is
      T : Float;
  begin
    if abs(Dimensionless(Time)) > 999_999 then
      raise Range_Error;
    end if;
    T := Float(Dimensionless(Time)) / 1000.0;
    return Type_Convert(T);
  end Convert;

  function Convert(Time : Seconds)
    return Milliseconds is
      T : Float;
  begin
    if abs(Dimensionless(Time)) > 999.999 then
      raise Range_Error;
    end if;
    T := Dimensionless(Time) * 1000.0;
    return Type_Convert(
      Integer_32(T));
  end Convert;

  function "*" (Left : Seconds; Right : Hertz)
    return Float is
  begin
    return Dimensionless(Left) * Dimensionless(Right);
  end "*";

  function "*" (Left : Hertz; Right : Seconds)
    return Float is
  begin
    return Dimensionless(Left) * Dimensionless(Right);
  end "*";

  function "/" (Left : Float; Right : Seconds)
    return Hertz is
  begin
    return Type_Convert(
      Left / Dimensionless(Right));
  end "/";

  function "/" (Left : Float; Right : Hertz)
    return Seconds is
  begin
    return Type_Convert(
      Left / Dimensionless(Right));
  end "/";

  procedure Normalize(Time  : in out Milliseconds) is
  begin
    Time := Time mod ONE_DAY;
  end Normalize;

  function ">="(Left, Right : Milliseconds)
    return Boolean is
      L, R : Milliseconds;
  begin
    L := Left;
    Normalize(L);
    R := Right;
    Normalize(R);
    if Dimensionless(L - R) > Dimensionless(TWELVE_Hours) or
       Dimensionless(L - R) < 0 then
         return False;
    else
      return True;
    end if;
  end ">=";

  function "<"  (Left,Right : Milliseconds)
    return Boolean is
  begin
    return not (Left >= Right);
  end "<";

  function "<=" (Left,Right : Milliseconds)
    return Boolean is
  begin
    return Left < Right or Left = Right;
  end "<=";

  function ">"  (Left,Right : Milliseconds)
    return Boolean is
  begin
    return not (Left <= Right);
  end ">";

end Time_Units;
with Dim_Int;
with Dim_Float;

package Time_Units is

--  pragma Elaborate_Body;

  -- There are two kinds of time measurements
  -- in this package -- Milliseconds and Seconds.

  -- Milliseconds are intended primarily to
  -- represent the time of day to the nearest
  -- millisecond.

  -- Seconds are intended primarily to
  -- represent time intervals in scientific
  -- equations.

  -- Therefore, the type Milliseconds has
  -- subprograms that convert time values
  -- to HH:MM:SS.TTT strings and vice versa,
  -- but Seconds does not.  Milliseconds
  -- "roll over" at midnight, and have special
  -- relational operators that know that
  -- 11:30 PM is "earlier" than 1:15 AM.
  -- Seconds do not have corresponding
  -- concepts.

  -- Seconds, on the other hand, are likely
  -- to be used in equations in scientific
  -- applications.  Therefore they have special
  -- arithmetic operators that know the
  -- relationship between the period of a
  -- cycle and its frequency.  Milliseconds
  -- have no corresponding concept.

  -- It is conceivable that one may want to
  -- convert between Milliseconds and Seconds.
  -- This will work for small values of
  -- Milliseconds, but will fail in general.
  -- The reason is that Seconds is derived
  -- from a Floating point type that has only
  -- 6 digits of accuracy.  The Convert functions
  -- will check to make sure the number of
  -- milliseconds is small before making the
  -- conversion.  If not, the Convert functions
  -- will raise Range_Error.

  type Milliseconds is new Dim_Int.Units;

  type Seconds is new Dim_Float.Units;

  type Hertz is new Dim_Float.Units;

  Range_Error : exception;
    -- Raised by Convert if the number of
    -- milliseconds is too large to be
    -- accurately represented by Seconds
    -- (which has only 6 digits of accuracy).

  -- The following functions convert time to
  -- a character string and vice versa.

  subtype Time_Strings is String(1..12);
    -- HH:MM:SS.TTT

  function Image(Time : Milliseconds)
    return Time_Strings;

  function Value(Time : Time_Strings)
    return Milliseconds;

  -- The following functions do the conversions
  -- between Milliseconds and Seconds.

  function Convert(Time : Milliseconds)
    return Seconds;
    -- May raise Range_Error;

  function Convert(Time : Seconds)
    return Milliseconds;
    -- May raise Range_Error;

  -- The following operators accomodate the
  -- relationship between time and frequency.

  function "*" (Left : Seconds; Right : Hertz)
    return Float;

  function "*" (Left : Hertz; Right : Seconds)
    return Float;

  function "/" (Left : Float; Right : Seconds)
    return Hertz;

  function "/" (Left : Float; Right : Hertz)
    return Seconds;

  -- The next routine effectively performs a "mod"
  -- function to convert negative times and times
  -- exceeding 24 hours to the equivalent value
  -- in the range 0 to 24 hours.

  procedure Normalize(Time  : in out Milliseconds);

  -- The following subprograms take into account the fact
  -- that time rolls over at midnight.  They require that
  -- the time difference be less than 12 hours.  The
  -- expression Left < Right tells if the Left time is
  -- earlier than the Right time. 

  function "<"  (Left, Right : Milliseconds)
    return Boolean;
  function "<=" (Left, Right : Milliseconds)
    return Boolean;
  function ">"  (Left, Right : Milliseconds)
    return Boolean;
  function ">=" (Left, Right : Milliseconds)
    return Boolean;

end Time_Units;
with Time_Units;
package Tu renames Time_Units;
