

--  visibility on required modules : 

with nt_console;
with ada.text_io;

package body board is

  package text_io renames ada.text_io;
  procedure next (me : IN board; x_co : IN OUT board_index; y_co : IN OUT board_index; dir : IN natural; res : OUT boolean);

  function find_turned (me : IN board; x : IN board_index; y : IN board_index) return natural;

  procedure turn_counters (me : IN OUT board; x : IN board_index; y : IN board_index; total : OUT natural);

  function no_turned (me : IN board; o_x : IN board_index; o_y : IN board_index; dir : IN natural; n : IN natural := 0) return natural;

  procedure capture (me : IN OUT board; x_co : IN board_index; y_co : IN board_index; dir : IN natural);

  procedure set_up (me : IN OUT board) is
    black_counter : counter.counter;
    white_counter : counter.counter;
  begin
    counter.set(black_counter, counter.BLACK);
    counter.set(white_counter, counter.WHITE);
    for x in me.sqrs'range(1) loop
      for y in me.sqrs'range(2) loop
        cell.initialise(me.sqrs(x,y));
      end loop;
    end loop;
    cell.add(me.sqrs(SIZE/2,SIZE/2), black_counter);
    cell.add(me.sqrs(SIZE/2,SIZE/2+1), white_counter);
    cell.add(me.sqrs(SIZE/2+1,SIZE/2), white_counter);
    cell.add(me.sqrs(SIZE/2+1,SIZE/2+1), black_counter);
    me.score(counter.BLACK) := 2;
    me.score(counter.WHITE) := 2;
  end set_up;

  procedure add (me : IN OUT board; x : IN integer; y : IN integer; move_is : IN move_status) is
    plays_with : counter.counter;
    turned : natural;
  begin
    counter.set(plays_with, me.player);
    me.last_move(me.player) := move_is;
    if move_is=OK then
      turn_counters(me, x, y, turned);
      cell.add(me.sqrs(x, y), plays_with);
      me.score(me.player) := me.score(me.player)+turned+1;
      me.score(me.opponent) := me.score(me.opponent)-turned;
    end if;
  end add;

  procedure now_playing (me : IN OUT board; c : IN counter.counter_colour) is
  begin
    me.player := c;
    case c is
      when counter.WHITE => me.opponent := counter.BLACK;
      when counter.BLACK => me.opponent := counter.WHITE;
    end case;
  end now_playing;

  procedure display (me : IN OUT board) is
    dashes : string(1..me.sqrs'length*4+1) := (others=>'-');
    package io is new text_io.integer_io(integer);
  begin
    nt_console.clear_screen;
    text_io.put(dashes);
    text_io.new_line;
    for x in me.sqrs'range(1) loop
      text_io.put('|');
      for y in me.sqrs'range(2) loop
        text_io.put(" ");
        cell.display(me.sqrs(x,y));
        text_io.put(" |");
      end loop;
      text_io.new_line;
      text_io.put(dashes);
      text_io.new_line;
    end loop;
    text_io.new_line;
    text_io.put("Player X has ");
    io.put(integer(me.score(counter.BLACK)), width=>2);
    text_io.put(" counters");
    text_io.new_line;
    text_io.put("Player O has ");
    io.put(integer(me.score(counter.WHITE)), width=>2);
    text_io.put(" counters");
    text_io.new_line;
  end display;

  function check_move (me : IN board; x : IN integer; y : IN integer) return move_status is
  begin
    if x=0 and then y=0 then
      return PASS;
    elsif x in board_index and then y in board_index then
      if cell.holds(me.sqrs(x,y))=cell.EMPTY then
        if find_turned(me, x, y)>0 then
          return OK;
        end if;
      end if;
    end if;
    return INVALID;
  end check_move;

  function status (me : IN board) return state_of_game is
  begin
    if me.score(me.opponent)=0 then
      return WIN;
    end if;
    if ( me.sqrs'length(1)*me.sqrs'length(2)=
         me.score(me.opponent)+me.score(me.player) ) 
      or 
       ( me.last_move(counter.BLACK)=PASS and        me.last_move(counter.WHITE)=PASS ) 
      then
      if me.score(me.opponent)=me.score(me.player) then
        return DRAW;
      end if;
      if me.score(me.opponent)<me.score(me.player) then
        return WIN;
      else
        return LOSE;
      end if;
    end if;
    return PLAY;
  end status;

  function contents (me : IN board; x : IN integer; y : IN integer) return cell.cell_holds is
  begin
    return cell.holds(me.sqrs(x, y));
  end contents;

  procedure next (me : IN board; x_co : IN OUT board_index; y_co : IN OUT board_index; dir : IN natural; res : OUT boolean) is
    x, y : natural;
  begin
    x := x_co;
    y := y_co;
    case dir is
      when 1 =>           y := y+1;
      when 2 => x := x+1; y := y+1;
      when 3 => x := x+1;
      when 4 => x := x+1; y := y-1;
      when 5 =>           y := y-1;
      when 6 => x := x-1; y := y-1;
      when 7 => x := x-1;
      when 8 => x := x-1; y := y+1;
      when others => raise constraint_error;
    end case;
    if x in board_range and then y in board_range then
      x_co := x;
      y_co := y;
      res := TRUE;
    else
      res := FALSE;
    end if;
  end next;

  function find_turned (me : IN board; x : IN board_index; y : IN board_index) return natural is
    sum : natural := 0;
  begin
    if cell.holds(me.sqrs(x,y))=cell.EMPTY then
      for dir in 1..8 loop
        sum := sum + no_turned(me, x, y, dir);
      end loop;
    end if;
    return sum;
  end find_turned;

  procedure turn_counters (me : IN OUT board; x : IN board_index; y : IN board_index; total : OUT natural) is
    num_cap : natural := 0;
    captured : natural;
  begin
    if cell.holds(me.sqrs(x, y))=cell.EMPTY then
      for dir in 1..8 loop
        captured := no_turned(me, x, y, dir);
        if captured>0 then
          capture(me, x, y, dir);
          num_cap := num_cap + captured;
        end if;
      end loop;
    end if;
    total := num_cap;
  end turn_counters;

  function no_turned (me : IN board; o_x : IN board_index; o_y : IN board_index; dir : IN natural; n : IN natural := 0) return natural is
    ok : boolean;
    nxt : cell.cell_holds;
    col : counter.counter_colour;
    x : board_index := o_x;
    y : board_index := o_y;
  begin
    next(me, x, y, dir, ok);
    if ok then
      nxt := cell.holds(me.sqrs(x, y));
      if nxt = cell.EMPTY then
        return 0;
      else
        col := cell.to_colour(nxt);
        if col=me.opponent then
          return no_turned(me, x, y, dir, n+1);
        elsif col=me.player then
          return n;
        end if;
      end if;
    else
      return 0;
    end if;
  end no_turned;

  procedure capture (me : IN OUT board; x_co : IN board_index; y_co : IN board_index; dir : IN natural) is
    ok : boolean;
    x, y : board_index;
    nxt : cell.cell_holds;
  begin
    x := x_co;
    y := y_co;
    next(me, x, y, dir, ok);
    if ok then
      nxt := cell.holds(me.sqrs(x,y));
      if cell.to_colour(nxt)=me.opponent then
        cell.flip(me.sqrs(x, y));
        capture(me, x, y, dir);
      else
        return;
      end if;
    else
      raise constraint_error;
    end if;
  end capture;


end board;
