--  visibility on required modules : 
with system;
use type system.Address;
with device_register;
use type device_register.csr;
with Ada.Real_Time; use Ada.Real_Time;
with RTA; use RTA;
with ch4csr;
with operator_console;
with data_logger;
with pump_controller;
with ch4_status;
package body ch4_sensor is
  procedure initialise;
  jitter_range : constant ch4_reading := 40;
  ch4_present : ch4_reading;
  ch4csr : device_register.csr;
  for ch4csr'address use 16#AA18#;
  ch4dbr : ch4_reading;
  for ch4dbr'address use 16#AA1A#;
  for ch4dbr'size use one_word;
  now : Duration;
  task THREAD is
    pragma PRIORITY(ch4_sensor_RTATT.INITIAL_THREAD_PRIORITY);
  end THREAD;
  procedure initialise is
  -- enable device
  ch4csr.device := D_Enabled;
  end initialise;
  procedure OPCS_thread is
  begin
    -- start conversion
    ch4csr.operation := Set;
    -- wait for conversion
    now := clock;
    loop
      exit when clock >= now + milliseconds(100);
    end loop;
    if nor ch4csr.done then
      operator_console.alarm(ch4_device_error);
    else
      -- read device register for sensor value
      ch4_present := ch4dbr;
      if ch4_present  > ch4_high then
        pump_controller.not_safe;
        ch4_status.write(motor_unsafe);
        operator_console.alarm(high_methane);
      end if;
      data_logger.ch4_log(ch4_present);
    end if;
  exception
    when others =>
      operator_console.alarm(unknown_error);
      pump_controller.not_safe;
      ch4_status.write(motor_unsafe);
  end OPCS_thread;
  task body THREAD is
    T : Ada.Real_Time.TIME := Ada.Real_Time.CLOCK;
  begin
    if T < RTA.SYSTEM_START_UP_TIME then
      T := RTA.SYSTEM_START_UP_TIME;
    end if;
    T := T + ch4_sensor_RTATT.OFFSET(RTA.START_MODE);
    loop
      delay until T;
      OPCS_thread;
      T := T + ch4_sensor_RTATT.PERIOD(RTA.CURRENT_MODE);
    end loop;
  end THREAD;
end ch4_sensor;
