/* Begin Body of cyclic                                                       */
/* ---------- include provided interface: ----------------------------------- */
#include "cyclic.h"
/* ---------- include required interface: ----------------------------------- */
#include "client.h"
#include "shared.h"

/* ---------- internal data: ------------------------------------------------ */
int reg = 0;

static int status = 0;

/* ---------- provided operation bodies: ------------------------------------ */
/*  constrained operation OPCS:                                               */
void OPCS_start(void)
{ 
  client__notify(); 
}

/*  constrained operation OPCS:                                               */
void OPCS_timer(void)
{ 
  shared__write(reg); 
}

/*  constrained operation OPCS:                                               */
void OPCS_stop(void)
{ 
  cyclic__OBCS_stop(); 
}

/* ---------- obcs body: ---------------------------------------------------- */

/*  OBCS main task:                                                           */
Process * cyclic__OBCS;

/*  input and output channels:                                                */

typedef struct {
  int  call_;
} cyclic__start_in;
typedef struct {
  int  return_;
} cyclic__start_out;
Channel * cyclic__start_req;
Channel * cyclic__start_rep;

typedef struct {
  int  call_;
} cyclic__stop_in;
typedef struct {
  int  return_;
} cyclic__stop_out;
Channel * cyclic__stop_req;
Channel * cyclic__stop_rep;

/*  OBCS stop channel:                                                        */
Channel * cyclic__OBCS_end;

/*  channels messages:                                                        */
static cyclic__start_in start_msg_in;
static cyclic__start_out start_msg_out;
static cyclic__stop_in stop_msg_in;
static cyclic__stop_out stop_msg_out;

/*  constrained operation calling procedure:                                  */

void cyclic__start(void)
{
  ChanOut(cyclic__start_req, &start_msg_in, sizeof(cyclic__start_in));
  ChanIn(cyclic__start_rep, &start_msg_out, sizeof(cyclic__start_out));
}

void cyclic__stop(void)
{
  ChanOut(cyclic__stop_req, &stop_msg_in, sizeof(cyclic__stop_in));
  ChanIn(cyclic__stop_rep, &stop_msg_out, sizeof(cyclic__stop_out));
}

/*  OBCS initialization procedure:                                            */
Process * cyclic__OBCS_init(void)
{
  cyclic__start_req = ChanAlloc();
  if (cyclic__start_req == NULL) return NULL;
  cyclic__start_rep = ChanAlloc();
  if (cyclic__start_rep == NULL) return NULL;
  cyclic__stop_req = ChanAlloc();
  if (cyclic__stop_req == NULL) return NULL;
  cyclic__stop_rep = ChanAlloc();
  if (cyclic__stop_rep == NULL) return NULL;
  cyclic__OBCS_end = ChanAlloc();
  if (cyclic__OBCS_end == NULL) return NULL;
  cyclic__OBCS = ProcAlloc(cyclic__OBCS_proc, 0, 0);
  if (cyclic__OBCS == NULL) return NULL;
  ProcRun(cyclic__OBCS);
  return cyclic__OBCS;
}

/*  OBCS execution procedure:                                                 */
void cyclic__OBCS_proc(Process * cyclic__OBCS)
{
  int i;
  int running = 1;
  while (running) {
    if (status == 0) {
      i = ProcAlt(cyclic__start_req, cyclic__OBCS_end, NULL);
      switch (i) {
        case 0 :
            /* LSER start */
            ChanIn(cyclic__start_req,
              &start_msg_in, sizeof(cyclic__start_in));
            ChanOut(cyclic__start_rep,
              &start_msg_out, sizeof(cyclic__start_out));
            OPCS_start();
            start_msg_out.return_ = 1;
            status = 1;
          break;
        case 1 :
          /* STOP OBCS */
          running = ChanInInt(cyclic__OBCS_end);
          break;
      }
    }
    if (status == 1) {
      /* BY_IT timer */
      int DeadLine = ProcTimePlus(ProcTime(),10);
      OPCS_timer();
      i = ProcTimerAlt(DeadLine, cyclic__stop_req, cyclic__OBCS_end, NULL);
      switch (i) {
        case -1 :
          break;
        case 0 :
            /* HSER stop */
            ChanIn(cyclic__stop_req,
              &stop_msg_in, sizeof(cyclic__stop_in));
            OPCS_stop();
            stop_msg_out.return_ = 1;
            ChanOut(cyclic__stop_rep,
              &stop_msg_out, sizeof(cyclic__stop_out));
            status = 0;
          break;
        case 1 :
          /* STOP OBCS */
          running = ChanInInt(cyclic__OBCS_end);
          break;
      }
    }
  }
}

/*  OBCS stop procedure:                                                      */
void cyclic__OBCS_stop(void)
{
  ChanOutInt(cyclic__OBCS_end, 0);
}


/*  OBCS termination procedure:                                               */
void cyclic__OBCS_term(void)
{
  free(cyclic__start_req);
  free(cyclic__start_rep);
  free(cyclic__stop_req);
  free(cyclic__stop_rep);
  free(cyclic__OBCS_end);
  ProcAllocClean(cyclic__OBCS);
}


/* End Body of cyclic                                                         */
