/* Begin Body of server                                                       */
/* ---------- include provided interface: ----------------------------------- */
#include "server.h"

/* ---------- internal define constants: ------------------------------------ */
#define TRUE 1

#define FALSE 0

/* ---------- internal data: ------------------------------------------------ */
static int status = 0;

/* ---------- provided operation bodies: ------------------------------------ */
/*  constrained operation OPCS:                                               */
void OPCS_op1(shared__T_value x /*in out*/, shared__T_value y /*in*/)
{ 
  shared__write(x); 
}

/*  constrained operation OPCS:                                               */
void OPCS_op2(shared__T_value x /*in*/)
{ 
  shared__write(x); 
}

/*  constrained operation OPCS:                                               */
void OPCS_op3(void)
{ 
  shared__T_value* tmp; 
  tmp = (shared__T_value*)shared__read; 
}

/* ---------- obcs body: ---------------------------------------------------- */

/*  OBCS main task:                                                           */
Process * server__OBCS;

/*  ASER operations Server Tasks:                                             */
Process * server__op3_st;

/*  input and output channels:                                                */

typedef struct {
  shared__T_value  y;
  shared__T_value  x;
  int  call_;
} server__op1_in;
typedef struct {
  shared__T_value  x;
  int  return_;
} server__op1_out;
Channel * server__op1_req;
Channel * server__op1_rep;

typedef struct {
  shared__T_value  x;
  int  call_;
} server__op2_in;
typedef struct {
  int  return_;
} server__op2_out;
Channel * server__op2_req;
Channel * server__op2_rep;

typedef struct {
  int  call_;
} server__op3_in;
Channel * server__op3_req;

/*  OBCS stop channel:                                                        */
Channel * server__OBCS_end;

/*  Server Tasks stop channels:                                               */
Channel * server__op3_end;

/*  channels messages:                                                        */
static server__op1_in op1_msg_in;
static server__op1_out op1_msg_out;
static server__op2_in op2_msg_in;
static server__op2_out op2_msg_out;
static server__op3_in op3_msg_in;

/*  constrained operation calling procedure:                                  */

void server__op1(shared__T_value x /*in out*/, shared__T_value y /*in*/)
{
  op1_msg_in.y = y;
  op1_msg_in.x = x;
  ChanOut(server__op1_req, &op1_msg_in, sizeof(server__op1_in));
  ChanIn(server__op1_rep, &op1_msg_out, sizeof(server__op1_out));
  x = op1_msg_out.x;
}

void server__op2(shared__T_value x /*in*/)
{
  op2_msg_in.x = x;
  ChanOut(server__op2_req, &op2_msg_in, sizeof(server__op2_in));
  ChanIn(server__op2_rep, &op2_msg_out, sizeof(server__op2_out));
}

void server__op3(void)
{
  ChanOut(server__op3_req, &op3_msg_in, sizeof(server__op3_in));
}

/*  OBCS initialization procedure:                                            */
Process * server__OBCS_init(void)
{
  server__op1_req = ChanAlloc();
  if (server__op1_req == NULL) return NULL;
  server__op1_rep = ChanAlloc();
  if (server__op1_rep == NULL) return NULL;
  server__op2_req = ChanAlloc();
  if (server__op2_req == NULL) return NULL;
  server__op2_rep = ChanAlloc();
  if (server__op2_rep == NULL) return NULL;
  server__OBCS_end = ChanAlloc();
  if (server__OBCS_end == NULL) return NULL;
  server__OBCS = ProcAlloc(server__OBCS_proc, 0, 0);
  if (server__OBCS == NULL) return NULL;
  ProcRun(server__OBCS);
  return server__OBCS;
}

/*  Server Tasks initialization procedure:                                    */
Process * server__op3_init(void)
{
  server__op3_req = ChanAlloc();
  if (server__op3_req == NULL) return NULL;
  server__op3_end = ChanAlloc();
  if (server__op3_end == NULL) return NULL;
  server__op3_st = ProcAlloc(server__op3_proc, 0, 0);
  if (server__op3_st == NULL) return NULL;
  ProcRunHigh(server__op3_st);
  return server__op3_st;
}


/*  OBCS execution procedure:                                                 */
void server__OBCS_proc(Process * server__OBCS)
{
  int i;
  int running = 1;
  while (running) {
    if (status == 0) {
      i = ProcAlt(server__op1_req, server__op2_req, server__OBCS_end, NULL);
      switch (i) {
        case 0 :
          if (TRUE) {
            /* HSER op1 */
            ChanIn(server__op1_req,
              &op1_msg_in, sizeof(server__op1_in));
            OPCS_op1(
              op1_msg_in.x, 
              op1_msg_in.y);
            op1_msg_out.return_ = 1;
            op1_msg_out.x = op1_msg_in.x;
            ChanOut(server__op1_rep,
              &op1_msg_out, sizeof(server__op1_out));
            status = 1;
          }
          break;
        case 1 :
          if (TRUE) {
            /* LSER op2 */
            ChanIn(server__op2_req,
              &op2_msg_in, sizeof(server__op2_in));
            ChanOut(server__op2_rep,
              &op2_msg_out, sizeof(server__op2_out));
            OPCS_op2(
              op2_msg_in.x);
            op2_msg_out.return_ = 1;
            status = 2;
          }
          break;
        case 2 :
          /* STOP OBCS */
          running = ChanInInt(server__OBCS_end);
          break;
      }
    }
    if (status == 1) {
      i = ProcAlt(server__OBCS_end, NULL);
      switch (i) {
        case 0 :
          /* STOP OBCS */
          running = ChanInInt(server__OBCS_end);
          break;
      }
    }
    if (status == 2) {
      i = ProcAlt(server__OBCS_end, NULL);
      switch (i) {
        case 0 :
          /* STOP OBCS */
          running = ChanInInt(server__OBCS_end);
          break;
      }
    }
  }
}

/*  OBCS stop procedure:                                                      */
void server__OBCS_stop(void)
{
  ChanOutInt(server__OBCS_end, 0);
}


/*  Server Task execution procedure:                                          */
void server__op3_proc(Process * server__op3_st)
{
  int i;
  int running = 1;
  while (running) {
    i = ProcAlt(server__op3_req, server__op3_end, NULL);
    switch (i) {
      case 0 :
        if (status == 1 && TRUE) {
          /* ASER op3 */
          ChanIn(server__op3_req,
            &op3_msg_in, sizeof(server__op3_in));
          OPCS_op3();
          status = 0;
        }
        if (status == 2 && TRUE) {
          /* ASER op3 */
          ChanIn(server__op3_req,
            &op3_msg_in, sizeof(server__op3_in));
          OPCS_op3();
          status = 0;
        }
        break;
      case 1 :
        /* STOP op3 */
        running = ChanInInt(server__op3_end);
        break;
    }
  }
}

/*  Server Task stop procedure:                                               */
void server__op3_stop(void)
{
  ChanOutInt(server__op3_end, 0);
}


/*  OBCS termination procedure:                                               */
void server__OBCS_term(void)
{
  free(server__op1_req);
  free(server__op1_rep);
  free(server__op2_req);
  free(server__op2_rep);
  free(server__OBCS_end);
  ProcAllocClean(server__OBCS);
}

/*  Server Tasks termination procedure:                                       */
void server__op3_term(void)
{
  free(server__op3_req);
  free(server__op3_end);
  ProcAllocClean(server__op3_st);
}


/* End Body of server                                                         */
