/************************************************************************
*									*
* The SB-Prolog System							*
* Copyright SUNY at Stony Brook, 1986; University of Arizona, 1987	*
*									*
************************************************************************/

/*-----------------------------------------------------------------
SB-Prolog is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the SB-Prolog General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
SB-Prolog, but only under the conditions described in the
SB-Prolog General Public License.   A copy of this license is
supposed to have been given to you along with SB-Prolog so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies. 
------------------------------------------------------------------ */
/* buffer.c */

#include "builtin.h"

/*
#define   EXAMINE_BUFFCODE
#define   DEBUG_SUBSTRING
#define   DEBUG_CONLENGTH
#define   DEBUG_ALLOC
*/


#define BACKWARDS 0      /* Direction parameter=0 means backwards, */
                         /* !=0 means forwards */

#define GET_E_LENGTH(ptr) (GET_LENGTH(ptr) < 65535 ? \
                (LONG)GET_LENGTH(ptr) : *(LONG_PTR)(GET_NAME(ptr)-4))

extern PSC_REC_PTR   insert();

static CHAR perm = PERM;
static CHAR temp = TEMP;

/* create a psc_entry on the heap. Len = 4 */
LONG makenullbuff()
{
    LONG            addr;
    PSC_REC_PTR     ptr;

    *hreg = (LONG)(hreg + 1);         /* fake first of pair */
    addr  = (LONG)hreg++;
    ptr   = (PSC_REC_PTR)hreg;
    hreg += 2;                        /* no ep and no hash link */
    GET_ETYPE(ptr)  = T_BUFF;
    GET_ARITY(ptr)  = 0;
    GET_LENGTH(ptr) = 0;
    GET_NAME(ptr)   = (CHAR_PTR)hreg;
    return( addr | CS_TAG );
}


b_ALLOC_BUFF()
{  /* r1 = size;
      r2 = Buffer,
      r3 = 0=>perm, 1=>heap, 2=>subbuff
      r4 = Super-Buffer if there is one
      r5 = Retcode
      long lengths (>= 65535) are kept in 4 bytes immediately preceeding
      the name
   */
   register LONG       op1, op3, op4;
   register LONG_PTR   top;
   PSC_REC_PTR         ptr, pptr;
   LONG                rc, size, len, disp;
   LONG_PTR            loc;
   LONG                addr;

   /* r1 should be bound to a number; r2 is free */
   op1  = GREGC(1);  DEREF(op1);
   size = INTVAL(op1);
   op3  = GREGC(3);  DEREF(op3);
   rc = 0;

#ifdef  DEBUG_ALLOC
   printf( "b_ALLOC_BUFF: " );
#endif

   switch ( INTVAL(op3) ) {
     case 0: /* perm */
             alloc_perm( size, &addr ); /* addr: ptr to struct node for buff */

#ifdef  DEBUG_ALLOC
   printf( "case 0 : perm\n" );
   printf( "            : size = %d   addr = %08x\n", size, addr );
#endif
             if ( ! unify(GREGC(2), (addr | CS_TAG))) {
                FAIL0;
             }
             break;
     case 1: /* heap */
             addr = makenullbuff();
             ptr = GET_STR_PSC(addr);

#ifdef  DEBUG_ALLOC
   printf( "case 1 : heap\n" );
   printf( "            : hreg = %08x size = %d  shift = %x\n", 
                          hreg, size, (size + 3 )>>2 );
   printf( "            : addr = %08x   ptr = %08x\n", addr & ~CS_TAG, ptr );
#endif
             if ( size < 65535 ) GET_LENGTH(ptr) = size;
             else {
                GET_LENGTH(ptr) = 65535;
                *hreg++ = size;
             }
             GET_NAME(ptr) = (CHAR_PTR)hreg;

             hreg += (size+3) >> 2;
             *(LONG_PTR)(GET_NAME(ptr)) = 4;    /* disp of next free */

#ifdef  DEBUG_ALLOC
   printf( "            : hreg = %08x \n", hreg );
   printf( "            : addr = %08x  =>  %08x\n",
                          (addr & ~CS_TAG), *(LONG_PTR)(addr & ~CS_TAG) );
   printf( "            : ptr  = %08x\n", ptr );
   printf( "            : etype = %d  arity = %d  len = %d\n",
                          GET_ETYPE(ptr), GET_ARITY(ptr), GET_LENGTH(ptr) );
   printf( "            : nameptr = %08x\n", GET_NAME(ptr) );
   printf( "            : ep      = %08x\n", GET_EP(ptr) );
#endif

             if ( ! unify(GREGC(2), (addr | CS_TAG))) {
                FAIL0;
             }
             break;
     case 2: /* subbuffer */
#ifdef  DEBUG_ALLOC
   printf( "case 2 : subbuffer\n" );
#endif
             op4 = GREGC(4);  DEREF(op4); /* super buffer */
             ptr = GET_STR_PSC(op4);
             len = GET_E_LENGTH(ptr);
             disp = *(LONG_PTR)(GET_NAME(ptr));
             loc = (LONG_PTR)(GET_NAME(ptr)+disp);

             if ( disp+12+size > len )
                rc = 1;
             else {
                *loc = (LONG)(loc+1);
                addr = (LONG)loc++;
                pptr = (PSC_REC_PTR)loc;
                GET_ETYPE(pptr) = T_BUFF;
                GET_ARITY(pptr) = 0;
                loc += 2; /*no ep, no hash link*/
                if ( size < 65535 )
                   GET_LENGTH(pptr) = size;
                else {
                   GET_LENGTH(pptr) = 65535;
                   *loc++ = size;
                }
                GET_NAME(pptr) = (CHAR_PTR) loc;
                *(LONG_PTR)(GET_NAME(pptr)) = 4; /* disp of next free */
                loc += (size+3) >> 2;
                disp = (LONG)(((CHAR_PTR)loc) - GET_NAME(ptr));
                *(LONG_PTR)(GET_NAME(ptr)) = disp;
                if ( ! unify(GREGC(2), (addr | CS_TAG))) {
                   FAIL0;
                }
             }
   }
   if ( ! unify(GREGC(5), MAKEINT(rc)) ) {
      FAIL0;
   }
}

b_BUFF_CODE()
{  /* b_BUFF_CODE inserts a word into, or extracts a word from, a buffer.
      It is VERY low-level and implementation dependent. It is used to
      generate byte-code into a buffer, and retrieve a word from a buffer.
      On entry, r1 is bound to a buffer, r2 is the offset in the buffer,
      r3 contains a number indicating what internal word to generate,
      r4 contains a term from which the word to insert in the buffer is
      extracted, or a variable that is bound to the word extracted from
      the buffer
   */

   PSC_REC_PTR        ptr, pscptr4;
   register LONG      op1, op4;
   register LONG_PTR  top;
   FILE              *fdes;
   int                disc, disp, i, arity;
   char               s[100];
   LONG               tempvar, temp1, temp2;

   op4 = GREGC(1);  DEREF(op4);
   if ( ! ISCONSTR(op4) ) {
       errmsg(11);
       FAIL0;
       return;
    }
    ptr = GET_STR_PSC(op4);
    op4 = GREGC(2);  DEREF(op4);  disp = INTVAL(op4);
    op4 = GREGC(3);  DEREF(op4);  disc = INTVAL(op4);
    op4 = GREGC(4);  DEREF(op4);

#ifdef  EXAMINE_BUFFCODE
        printf( "b_BUFF_CODE : case ==== %2d ==== ", disc );
#endif

    switch (disc) {
      case  0: /* ppsc: constant/structure, UNTAGGED psc_ptr to buff */
               /* make permanent if could be dangling pointer */

               if ((LONG)GET_NAME(ptr)<(LONG)GET_STR_PSC(op4)) {
                  pscptr4 = GET_STR_PSC(op4);
                  op4 = (LONG)insert(GET_NAME(pscptr4), GET_LENGTH(pscptr4),
                                     GET_ARITY(pscptr4), &perm) | CS_TAG;
               }
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", 
                 GET_NAME(ptr)+disp, 
                 (LONG)(GET_STR_PSC(op4)) );
#endif
               *(LONG_PTR)(GET_NAME(ptr)+disp) = (LONG)(GET_STR_PSC(op4));
               break;
      case  1: /* pppsc: cons/struc, UNTAGGED ptr to psc_ptr to buff */
               /* make permanent if could be dangling pointer        */
               if ((LONG)GET_NAME(ptr)<(LONG)UNTAGGED(op4)) {
                  pscptr4 = GET_STR_PSC(op4);
                  op4 = (LONG)insert(GET_NAME(pscptr4), GET_LENGTH(pscptr4),
                                     GET_ARITY(pscptr4), &perm);
               }
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", 
                 GET_NAME(ptr)+disp, 
                 UNTAGGED(op4) );
#endif

               *(LONG_PTR)(GET_NAME(ptr)+disp) = UNTAGGED(op4);
               break;
      case  2: /* pn: number, UNTAGGED LONG value to buff */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", 
                 GET_NAME(ptr)+disp, 
                 INTVAL(op4) );
#endif

               *(LONG_PTR)(GET_NAME(ptr)+disp) = INTVAL(op4);
               break;
      case  3: /* ps: number, UNTAGGED short value to buff */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x        %04x\n", 
                 GET_NAME(ptr)+disp, 
                 INTVAL(op4) );
#endif

               *(WORD_PTR)(GET_NAME(ptr)+disp) = INTVAL(op4);
               break;
      case  4: /* ga: return address of location in buffer, as tagged num */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x\n", 
                 GET_NAME(ptr)+disp );
#endif

               if ( ! unify(op4, MAKEINT(GET_NAME(ptr)+disp)) ) {
                  FAIL0;
               }
               break;
      case  5: /* gn: return word at location in buffer, as tagged num */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", 
                 GET_NAME(ptr)+disp, 
                 *(LONG_PTR)(GET_NAME(ptr)+disp) );
#endif

               if ( ! unify(op4, MAKEINT(*(LONG_PTR)(GET_NAME(ptr)+disp))) ) {
                  FAIL0;
               }
               break;
      case  6: /* gs: return short at location in buffer, as tagged num */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x        %04x\n", 
                 GET_NAME(ptr)+disp, 
                 *(WORD_PTR)(GET_NAME(ptr)+disp) );
#endif

               if ( ! unify(op4, MAKEINT(*(WORD_PTR)(GET_NAME(ptr)+disp))) ) {
#ifdef  EXAMINE_BUFFCODE
        printf(" FAIL in case 6\n");
#endif
                  FAIL0;
               }
               break;
      case  7: /* gepb: return buff that 1st arg pred ep points to       */
               /* assumes ep points to 2nd word in buff, first is ptr to */
               /*   its own psc_ptr                                      */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", 
                 GET_EP(ptr)-1, 
                 *(LONG_PTR)(GET_EP(ptr)-1) );
#endif

               if ( ! unify(op4, *(LONG_PTR)(GET_EP(ptr)-1))) {
                  FAIL0;
               }
               break;
      case  8: /* gpb: return buff that word at disp-buff points to */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x    %08x    %08x\n", 
                 GET_NAME(ptr)+disp,
                 *(LONG_PTR)(GET_NAME(ptr)+disp),
                 *(LONG_PTR)(GET_NAME(ptr)+disp)-4,
                 *(LONG_PTR)((*(LONG_PTR)(GET_NAME(ptr)+disp))-4) );
#endif

               if ( ! unify(op4,
                     *(LONG_PTR)((*(LONG_PTR)(GET_NAME(ptr)+disp))-4))) {
                  FAIL0;
               }
               break;
      case  9: /* pep: set ep of term to point to 4th byte in buff */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %8x\n", 
                 GET_EP(ptr), 
                 (LONG_PTR)(GET_NAME(GET_STR_PSC(op4))+4));
#endif

               GET_EP(ptr) = (LONG_PTR)(GET_NAME(GET_STR_PSC(op4))+4);
               GET_ETYPE(ptr) = T_DYNA;
               /* and set 1st word of buff to point to buff psc ptr */
               /* *(LONG_PTR)GET_NAME(GET_STR_PSC(op4)) = op4; */
               break;
      case 10: /* pbr: set word p points to 4th byte in buff */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", 
                 GET_NAME(ptr)+disp, 
                 GET_NAME(GET_STR_PSC(op4))+4 );
#endif

               *(CHAR_PTR*)(GET_NAME(ptr)+disp) = GET_NAME(GET_STR_PSC(op4))+4;
               /* and set 1st word of buff to point to buff psc ptr */
               /* *(LONG_PTR)GET_NAME(GET_STR_PSC(op4)) = op4; */
               break;
      case 11: /* rep: reset ep of first arg to undefined */
               if (IS_PRED(ptr) || IS_DYNA(ptr)) {
                  GET_EP(ptr) = 0;
                  GET_ETYPE(ptr) = T_ORDI;
               }
               break;
      case 12: /* fv: free variable to buff, for use with getival instr */
               temp1 = (LONG)(GET_NAME(ptr)+disp);
               if (! (temp1 & 3) )
                  FOLLOW(temp1) = temp1;
               else {
                  curr_fence = (CHAR_PTR)(((LONG)curr_fence + 3) & 0xfffffffc);
                  FOLLOW(temp1) = (LONG)curr_fence;
                  FOLLOW(curr_fence) = (LONG)curr_fence;
                  curr_fence += 4;
               }
               break;
      case 13: /* execb: branch into buffer,  using 4th arg as call */
               pcreg = (WORD_PTR)(GET_NAME(ptr)+disp);
               arity = GET_ARITY(GET_STR_PSC(op4));
               UNTAG(op4);
               for ( i=1; i <= arity; i++ )
                  GREGC(i) = FOLLOW((LONG_PTR)op4+i);
               break;
      case 14: /* ptv: number or const, tagged word value to buff */
               /* make permanent if could be dangling pointer     */
               if (ISCONSTR(op4) && (LONG)GET_NAME(ptr)<(LONG)UNTAGGED(op4)) {
                  pscptr4 = GET_STR_PSC(op4);
                  op4 = (LONG)((LONG)insert(GET_NAME(pscptr4),
                                            GET_LENGTH(pscptr4),
                                            GET_ARITY(pscptr4), &perm)
                                            | CS_TAG);
               }
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", 
                 GET_NAME(ptr)+disp, 
                 UNTAGGED(op4) );
#endif

               *(LONG_PTR)(GET_NAME(ptr)+disp) = op4;
               break;
      case 15: /* ptp: put str-tagged pointer to second location */
               *(LONG_PTR)(GET_NAME(ptr)+disp) =
                     (LONG)((LONG)(GET_NAME(ptr)+INTVAL(op4)) | CS_TAG);
               break;
      case 16: /* ptl: put list-tagged pointer to second location */
               *(LONG_PTR)(GET_NAME(ptr)+disp) =
                     (LONG)((LONG)(GET_NAME(ptr)+INTVAL(op4)) | LIST_TAG);
               break;
      case 17: /* pvar: put variable into buffer */
               tempvar = (LONG)(GET_NAME(ptr)+disp);
               FOLLOW(tempvar) = tempvar;
               if (op4 >= (LONG)GET_NAME(ptr) &&
                   op4 <= (LONG)(GET_NAME(ptr)+GET_E_LENGTH(ptr)))
                  /* already points into the buffer */
                  FOLLOW(op4) = tempvar; /* point word to it */
               else {
                  FOLLOW(op4) = tempvar;
                  PUSHTRAIL(op4);      /* and trail! */
               }
               break;
      case 18: /* ubv: unify value with buffer */
               if ( ! unify(GET_NAME(ptr)+disp, op4)) {
                  FAIL0;
               }
               break;
      case 19: /* cep: copy ep of last arg to ep of first */
               pscptr4 = GET_STR_PSC(op4);
               if ( ! IS_BUFF(ptr) ) {
                  GET_EP(ptr) = GET_EP(pscptr4);
                  GET_ETYPE(ptr) = GET_ETYPE(pscptr4);
                }
                break;
      case 20: /* pepb: copy ep of last arg to loc in buff */
               pscptr4 = GET_STR_PSC(op4);
               FOLLOW(GET_NAME(ptr)+disp) = (LONG)(GET_EP(pscptr4));
               break;
      case 21: /*gnb: return next buffer along the hash chain */
               op1 = *(LONG_PTR)(GET_NAME(ptr) + disp);
               op1 = *(LONG_PTR)(op1 + 12) - 12;
               if ( ! unify(op4, *(LONG_PTR)op1)) {
                  FAIL0;
               }
               break;
      case 22: /*dis: disassemble buffer for debugging assert         */
               /*op1: buffer; op2: 0 for "w" 1 for "a"; op4: filename */
               pscptr4 = GET_STR_PSC(op4);
               temp1 = (LONG)pcreg;
               temp2 = num_line;
               num_line = 1;
               namestring(pscptr4, s);
               if (disp == 0)
                  fdes = fopen(s, "w");
               else
                  fdes = fopen(s, "a");
               fprintf(fdes,
  "\n /* New Buffer Below: pscptr, arity, length, nameptr, backptr *\/\n\n");
               fprintf(fdes, "%x, %d, %d, %x, %x \n", ptr, GET_ARITY(ptr),
                             GET_LENGTH(ptr), GET_NAME(ptr), GET_EP(ptr));
               pcreg = (WORD_PTR)(GET_NAME(ptr) + 4);
               while (pcreg < (WORD_PTR)(GET_NAME(ptr) + GET_LENGTH(ptr)) )
                  print_inst(fdes, pcreg);
               fflush( fdes );
               fclose( fdes );
               pcreg = (WORD_PTR)temp1;
               num_line = temp2;
               break;
      case 23: /* ps: short number, UNTAGGED word value to buff */
               *(WORD_PTR)(GET_NAME(ptr)+disp) = INTVAL(op4);
               break;
      case 24: /* gs: return short number at location in buffer,  */
               /* as tagged num                                   */
               if ( ! unify(op4, MAKEINT(*(WORD_PTR)(GET_NAME(ptr)+disp)))) {
                  FAIL0;
               }
               break;
      case 25: /* bb: build a buffer for a pointer in a buffer, 0 len  */
               temp1 = makenullbuff();
               pscptr4 = GET_STR_PSC(temp1);
               GET_NAME(pscptr4) = *(CHAR_PTR *)(GET_NAME(ptr)+disp);

#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", GET_NAME(pscptr4), GET_NAME(ptr)+disp  );
#endif
               if ( ! unify(op4, temp1) ) {
                  FAIL0;
               }
               break;
      case 26: /* pba: put buffer address into buffer */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", GET_NAME(ptr)+disp, GET_NAME(pscptr4) );
#endif
               pscptr4 = GET_STR_PSC(op4);
               FOLLOW(GET_NAME(ptr)+disp) = (LONG)(GET_NAME(pscptr4));
               break;
      case 27: /* pf: put float (in WAM format, word) into buffer */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", GET_NAME(ptr)+disp, op4 );
#endif
               *(LONG_PTR)(GET_NAME(ptr)+disp) = op4;
               break;
	case 28: /* gppsc: get ptr to psc_table entry */
	    if (!unify(op4, (*(LONG_PTR)(GET_NAME(ptr)+disp))|CS_TAG)) {
               FAIL0;
            }
	    break;
	case 29: /* gf: get float (in WAM format, word) from buffer */
	    if (!unify(op4, *(LONG_PTR)(GET_NAME(ptr)+disp))) {
               FAIL0;
            }
	    break;

      case 30: /* pb: number, UNTAGGED byte value to buff */
#ifdef  EXAMINE_BUFFCODE
        printf( "%08x    %08x\n", GET_NAME(ptr)+disp, INTVAL(op4) );
#endif
               *(BYTE_PTR)(GET_NAME(ptr)+disp) = INTVAL(op4);
               break;
      case 31: /* gb: return byte at location in buffer, as tagged num */
#ifdef  EXAMINE_BUFFCODE
        printf( "\n" );
#endif
               if ( ! unify(op4, MAKEINT(*(BYTE_PTR)(GET_NAME(ptr)+disp))) ) {
                  FAIL0;
               }
               break;
    }
}

b_TRIMBUFF()
{  /* reg1 : new length (if <0 scan for 0x00)
      reg2 : buffer
      reg3 : 0 => perm, 1 => temp, 2 => temp in superbuff (r4)
      reg4 : superbuff to trim (if 2)
   */

   register LONG      op, op1, op3;
   register LONG_PTR  top;
   PSC_REC_PTR        ptr, bptr;
   LONG               len;

   op3 = GREGC(3);  DEREF(op3);   op3 = INTVAL(op3);
   op  = GREGC(2);  DEREF(op );   ptr = GET_STR_PSC(op);
   op1 = GREGC(1);  DEREF(op1);   len = INTVAL(op1);

#ifdef  DEBUG_ALLOC
   printf( "b_TRIMBUFF: op1 = %08x  op = %08x  op3 = %08x\n", op1, op, op3 );
   printf( "          : ptr = %08x  len =%08x\n", ptr, len );
   printf( "          : type = %d  arity = %d  len = %d\n", 
                        GET_ETYPE(ptr), GET_ARITY(ptr), GET_LENGTH(ptr) );
   printf( "          : name = %08x  %s\n", GET_NAME(ptr), GET_NAME(ptr) );
   printf( "          : ep   = %08x\n", GET_EP(ptr) );
#endif

   if ( len < 0 ) {
      if (GET_NAME(ptr)) {
         len = strlen(GET_NAME(ptr));
#ifdef  DEBUG_ALLOC
         printf( " len < 0  : name = %08x  name = %s  newlen = %d\n",
                     GET_NAME(ptr), GET_NAME(ptr), len );
#endif
      } else {
#ifdef  DEBUG_ALLOC
         printf( " len < 0  : failing  (FAIL0)\n" );
#endif
         FAIL0;
      }
   }
   else if (op3 == 0) { /* perm */
      if (curr_fence == (CHAR_PTR)(GET_NAME(ptr) + GET_LENGTH(ptr)))
         curr_fence = (CHAR_PTR)(GET_NAME(ptr) + len);
   }
   else if (op3 == 1) { /* heap */
      if (hreg == (LONG_PTR)(GET_NAME(ptr) + GET_E_LENGTH(ptr)))
         hreg = (LONG_PTR)(GET_NAME(ptr) + len);
    }
    else {  /* in superbuffer */
      op = GREGC(4); DEREF(op); bptr = GET_STR_PSC(op);
      if ((LONG)GET_NAME(bptr)+*(LONG_PTR)GET_NAME(bptr) ==
              (LONG)GET_NAME(ptr) + GET_E_LENGTH(ptr))
          *(LONG_PTR)(GET_NAME(bptr)) = /* new disp */
              (((LONG)GET_NAME(ptr)+len-(LONG)GET_NAME(bptr))+3)&0xfffffffc;
   }
   if ( GET_LENGTH(ptr) < 65535 )
      GET_LENGTH(ptr) = len;
   else
      *(LONG_PTR)(GET_NAME(ptr)-4) = len;
}


b_SUBSTRING()
{  /* reg1 = direction (1 = forwards for read,  0 = backwards for write)
      reg2 = numbytes
      reg3 = Internal constant
      reg4 = Initial location in the input buffer
      reg5 = Input Buffer  (must be a valid constant)
      reg6 = Final location in the input buffer after reading from or
             writing into the buffer

      Forwards: If the internal constant parameter is free upon entry, it
                takes the first numbytes of the input buffer and creates
                an internal constant.  Reg 6 gets bound to the location
                in the input buffer directly FOLLOWing the constant. If the
                constant parameter is already bound, it is checked against
                the one in the buffer to see if they unify.  The numbytes
                parameter must unify with the length of the constant.

     Backwards: Binds numbytes(if not bound) to the length of the internal
                constant. Copies the internal constant into the Input Buffer,
                and returns in reg 6 an index into the input buffer which
               directly FOLLOWs the constant.
   */

   CHAR_PTR     Bnameptr, Cnameptr;/* Buffer Nameptr, Constant Nameptr */
   PSC_REC_PTR       buff_psc;     /* pointer to psc rec of buffer */
   PSC_REC_PTR       addr;         /* Holds result from insert */
   int               offset;       /* Offset into buffer  */
   int               numbytes;     /* Numbytes in buffer for constant */
   int               i;            /* Counter */
   LONG              op1, op2, op3, op4, op5;
   register LONG_PTR top;

   op1 = GREGC(1);  DEREF(op1);          /* direction */
   op2 = GREGC(2);                       /* length */
   op3 = GREGC(3);  DEREF(op3);          /* constant substring */
   op4 = GREGC(4);  DEREF(op4);          /* offset */
   op5 = GREGC(5);  DEREF(op5);          /* constant, long string */


          /* check the direction param for error */
   if ( ! ISINTEGER(op1) ) {
      errmsg(0);
      FAIL0;
      return;
   }
         /* check input buffer - ?valid constant */
   if ( ! ISATOM(op5) ) {
      errmsg(8);  FAIL0;
      return;
   }
   buff_psc = GET_STR_PSC(op5);

         /* check that offset is valid */
   if ( !ISINTEGER(op4) ) {
      errmsg(2);  FAIL0;
      return;
   }
   offset = INTVAL(op4);
   if ((GET_LENGTH(buff_psc) < offset) || (offset < 0)) {
      errmsg(3);  FAIL0;
      return;
   }
         /* get length of substring */
   if ( (ISATOM(op3) ) &&
      ( ! unify(op2, MAKEINT(GET_LENGTH(GET_STR_PSC(op3)))))) {
         FAIL0;
         return;
   }
   DEREF(op2);
   if ( ! ISINTEGER(op2) ) {
      errmsg(4);  FAIL0;
      return;
   }
   numbytes = INTVAL(op2);
         /* check no buffer overflow */
   if ((numbytes < 0) || ((offset+numbytes) > GET_LENGTH(buff_psc))) {
      errmsg(3);  FAIL0;
      return;
   }

   if ( INTVAL(op1) != BACKWARDS ) {
      /* find the constant,  or insert a new one */
      temp = TEMP;
      addr = insert(GET_NAME(buff_psc)+offset,numbytes,0,&temp); /*&perm*/
      if ( ! unify(op3, (LONG)addr | CS_TAG)) {
         FAIL0;
         return;
      }
   }
   else {   /* going backwards */
      Bnameptr = GET_NAME(buff_psc);
      if ( ! ISATOM(op3)) {
         errmsg(5);  FAIL0;
         return;
      }
      Cnameptr = GET_NAME(GET_STR_PSC(op3));
      for ( i=0;  i < numbytes;  i++ )    /* copy into buffer */
          Bnameptr[i+offset] = Cnameptr[i];


#ifdef  DEBUG_SUBSTRING
printf( "b_SUBSTRING: (1)dir = %d   (2)bytes = %d   (4)ofs = %d\n",
                      INTVAL(op1), INTVAL(op2), INTVAL(op4) );
printf( "           : const = %08x\n", GET_STR_PSC(op3) );
printf( "           :    type = %d  arity = %d  len = %d\n",
                           GET_ETYPE(GET_STR_PSC(op3)),
                           GET_ARITY(GET_STR_PSC(op3)),
                           GET_LENGTH(GET_STR_PSC(op3)) );
printf( "           :    name = %08x : %s\n",  
                           GET_NAME(GET_STR_PSC(op3)),
                           GET_NAME(GET_STR_PSC(op3))  );
printf( "           : buff  = %08x\n", GET_STR_PSC(op5) );
printf( "           :    type = %d  arity = %d  len = %d\n",
                           GET_ETYPE(GET_STR_PSC(op5)),
                           GET_ARITY(GET_STR_PSC(op5)),
                           GET_LENGTH(GET_STR_PSC(op5)) );
printf( "           :    name = %08x : %s\n",  
                           GET_NAME(GET_STR_PSC(op5)),
                           GET_NAME(GET_STR_PSC(op5))  );
printf( "           : Cname = %s\n", Cnameptr );
printf( "           : Bname = %s\n", Bnameptr );
#endif

   }
          /* check or set out location */
   if ( ! unify(GREGC(6), MAKEINT(offset+numbytes))) {
      FAIL0;
   }
}


b_SUBNUMBER()
{  /* reg1 = direction (forwards for read,  backward for write)
      reg2 = number of bytes storing the length
      reg3 = numeric constant
      reg4 = Initial location in buffer
      reg5 = Input Buffer
      reg6 = Final location in buffer

      Forward case: Takes numbytes out of Input Buffer, converts
                    it to a number.  (starting from initial loc in buffer)
                    Binds reg 6 to location in the buffer FOLLOWing the
                    string representation of the number.

      Backward case: Takes a numeric constant, stores its name in the
                     Input buffer and binds reg6 to the location in the
                     buffer FOLLOWing the string representation of number
   */

   CHAR_PTR      Bnameptr;         /* Buffer nameptr */
   PSC_REC_PTR   buff_psc;          /* pointer to psc rec of buffer */
   int           num;               /* Number from subnumber op */
   int           numbytes;          /* Number of bytes */
   int           xtra;              /* Number of leading zeros */
   int           i;                 /* Counter */
   int           offset;            /* Offset into buffer */
   char          s[10];             /* String representation of number */
   LONG          op1, op2, op3, op4, op5;
   register LONG_PTR  top;

   op1 = GREGC(1);  DEREF(op1);     /* direction */
   op2 = GREGC(2);  DEREF(op2);     /* length */
   op3 = GREGC(3);  DEREF(op3);     /* numeric constant,  substring */
   op4 = GREGC(4);  DEREF(op4);     /* offset */
   op5 = GREGC(5);  DEREF(op5);     /* buffer constant, long string */

         /* check the direction param for error */
   if ( ! ISINTEGER(op1) ) {
      errmsg(0);  FAIL0;
      return;
   }
         /* check input buffer - ?valid constant */
   if ( ! ISATOM(op5) ) {
      errmsg(8);  FAIL0;
      return;
   }
   buff_psc = GET_STR_PSC(op5);
         /* check that offset is valid */
   if ( ! ISINTEGER(op4) ) {
      errmsg(2);  FAIL0;
      return;
   }
   offset = INTVAL(op4);
   if ((GET_LENGTH(buff_psc) < offset) || (offset < 0)) {
      errmsg(3);  FAIL0;
      return;
   }
   if ( ! ISINTEGER(op2) ) {    /* Number of bytes which is the length */
      errmsg(6);                /* of the string representation of the */
      FAIL0;                    /* number MUST be specified in both    */
      return;                   /* the forward and backward case.      */
   }
   numbytes = INTVAL(op2);    /* get the length of const into numbytes  */
         /* check no buffer overflow */
   if ((numbytes < 0) || ((offset+numbytes) > GET_LENGTH(buff_psc))) {
      errmsg(3);  FAIL0;
      return;
   }

  if ( INTVAL(op1) != BACKWARDS) {
         /* get numeric equivalent out */
      num = getnum(numbytes,GET_NAME(buff_psc)+offset);
      if ( ! unify(op3, MAKEINT(num)) ) {     /* unify reg3 with number */
         FAIL0;
         return;
      }
   }
   else { /* going backwards */
      if ( ! ISINTEGER(op3) ) {
         errmsg(6);          /* no number to be written */
         FAIL0;
         return;
      }
      num = INTVAL(op3);               /* get number to be written */
      Bnameptr = GET_NAME(buff_psc);   /* get buffer name pointer */
      itoa(num,s);                     /* make s string representing num */
      xtra = numbytes - strlen(s);     /* number of leading zeros */
      if ( xtra < 0 ) {                /* number too large */
         errmsg(10);
         FAIL0;
         return;
      }
      for( i=0; i < xtra; i++ )      /* put leading zeros in if any */
         Bnameptr[i+offset] = '0';
      for (i = xtra; i < strlen(s) + xtra;  i++ )   /* put character rep of */
         Bnameptr[i+offset] = s[(i-xtra)];          /* number into buffer*/
   }
          /* check or set out location */
   if ( ! unify(GREGC(6), MAKEINT(offset+numbytes)) ) {
      FAIL0;
   }
}


b_SUBDELIM()
{  /* reg1 = direction       (forwards for read,  backwards for write )
      reg2 = delimiter
      reg3 = internal constant
      reg4 = Initial location in buffer
      reg5 = Input buffer
      reg6 = Final location in buffer

      Forwards: Takes the characters preceeding the delimiter
                in the input buffer, and creates an internal
                constant with that name.  Binds reg[6] to the
                location in the buffer FOLLOWing the delimiter.

      Backwards: Puts the internal constant into the buffer,
                 appends the delimiter to it, binds reg[6] to
                 the final location in the input buffer
   */

   PSC_REC_PTR  addr;                  /* Holds result from insert */
   CHAR_PTR     Bnameptr;              /* Buffer Nameptr */
   CHAR_PTR     Cnameptr;              /* Constant Nameptr */
   CHAR_PTR     Dnameptr;              /* Delimiter Nameptr */
   PSC_REC_PTR  buff_psc, con_psc;     /* pointers to psc recs */
   int          offset;                /* Offset into buffer */
   int          Blen;                  /* Buffer length */
   int          Clen;                  /* Constant length */
   int          i;                     /* Counter */
   LONG         op1, op2, op3, op4, op5;
   register LONG_PTR     top;

   op1 = GREGC(1);  DEREF(op1);        /* direction */
   op2 = GREGC(2);  DEREF(op2);        /* delimiter */
   op3 = GREGC(3);  DEREF(op3);        /* constant,  substring */
   op4 = GREGC(4);  DEREF(op4);        /* offset */
   op5 = GREGC(5);  DEREF(op5);        /* buffer constant, long string */

          /* check the direction param for error */
   if ( ! ISINTEGER(op1) ) {
      errmsg(0);  FAIL0;
      return;
   }
          /* check input buffer - ?valid constant */
   if ( ! ISATOM(op5) ) {
      errmsg(8);  FAIL0;
      return;
   }
   buff_psc = GET_STR_PSC(op5);
          /* check that offset is valid */
   if ( ! ISINTEGER(op4) ) {
      errmsg(2);  FAIL0;
      return;
   }
   offset = INTVAL(op4);
   if ( (GET_LENGTH(buff_psc) < offset) || (offset < 0) ) {
      errmsg(3);  FAIL0;
      return;
   }

   if ( ! ISATOM(op2) ) {           /* delimiter must be given */
      errmsg(7);  FAIL0;
      return;
   }

   Bnameptr = GET_NAME(buff_psc);            /* get nameptr for Buffer */
   Dnameptr = GET_NAME(GET_STR_PSC(op2));    /* get nameptr for delimiter */
   if ( INTVAL(op1) != BACKWARDS ) {
      Blen = GET_LENGTH(buff_psc);           /* length of Buffer */
      Clen = 0;                          /* get the length of the constant */
      while ((Bnameptr[offset+Clen]!=Dnameptr[0]) && (offset+Clen<Blen))
          Clen++;
      if (offset+Clen >= Blen) {
         FAIL0;
         return;
      }
           /* create constant of length Clen */
      temp = TEMP;
      addr = insert(Bnameptr+offset,Clen,0,&temp);
          /* unify this with reg3 */
      if ( ! (unify(op3, (LONG)addr | CS_TAG)) ) {
         FAIL0;
         return;
      }
   }
   else {/* going backwards */
      if ( ISATOM(op3) ) {  /* make sure there is a constant to write out */
         con_psc  = GET_STR_PSC(op3);
         Cnameptr = GET_NAME(con_psc);      /* get cnst nameptr */
         Clen     = GET_LENGTH(con_psc);    /* get length of constant */
         for( i=0; i < Clen; i++ )          /* copy constant into buffer */
            Bnameptr[offset+i] = Cnameptr[i];
            Bnameptr[offset + Clen] = Dnameptr[0];
            /* copy delimiter into buffer */
      }
      else {         /* no constant to be written out */
         errmsg(5);  FAIL0;
         return;
      }
   }
      /* validate or bind outloc */
   if ( ! unify(GREGC(6), MAKEINT(offset+Clen+1)) ) {
      FAIL0;
   }
 }


b_CONLENGTH()
{  /* reg 1 is an internal constant or a number
      reg 2 is the length of the constant or number
   */

   LONG              op1;
   register LONG_PTR top;
   int               len;

   op1 = GREGC(1);  DEREF(op1);

   if ( ISATOM(op1) ) {
#ifdef  DEBUG_CONLENGTH
printf( "b_CONLENGTH : ptr = %08x\n", GET_STR_PSC(op1) );
printf( "            :   type = %d  arity = %d  len = %d\n", 
                         GET_ETYPE(GET_STR_PSC(op1)),
                         GET_ARITY(GET_STR_PSC(op1)),
                         GET_LENGTH(GET_STR_PSC(op1)) );
printf( "            :   name = %08x   %s\n",
                         GET_NAME(GET_STR_PSC(op1)),
                         GET_NAME(GET_STR_PSC(op1)) );
#endif

      len = GET_E_LENGTH(GET_STR_PSC(op1));
   }
   else if ( ISINTEGER(op1) )
      len = numlength(INTVAL(op1));
   else if ( ISFLOAT(op1) ) {
      printf( "conlength: FLOAT case not implemented\n" );
      len = 0;
   }
   else {
      errmsg(9);  FAIL0;
      return;
   }

   if ( ! unify(MAKEINT(len), GREGC(2)) ) {
      FAIL0;
   }
}

/*****************************************************************************/
/* Routine name: errmsg                                                      */
/* Input Parameter: errnum  type: short integer                              */
/* Purpose:  To output a relevant message when an error occurs.              */
/*****************************************************************************/
errmsg( errnum )
int     errnum;
{
   switch( errnum ) {
   case  0: printf( "Error: Direction parameter must must be a 0 or 1.\n" );
   case  1: printf( "Error: Delimiter not found in buffer.\n" );
   case  2: printf( "Error: Index into buffer is not an integer.\n" );
   case  3: printf( "Error: Index into buffer is out of range.\n" );
   case  4: printf(
   "Error: Constant and length params are both free or bound improperly.\n"  );
   case  5: printf( "Error: Nothing to write out in sub* backwards.\n" );
   case  6: printf( "Error: Length must be bound in subnumber operation.\n" );
   case  7: printf( "Error: Delimiter must be bound in subdelim operation.\n" );
   case  8: printf( "Error: Input buffer is free or bound improperly.\n" );
   case  9: printf( "Error: Improper argument to Conlength.\n" );
   case 10: printf( "Error: Number too large for field in Subnumber.\n" );
   case 11: printf( "Error: Illegal arg to buff_code.\n" );
   }
}
