/************************************************************************
*									*
* The SB-Prolog System							*
* Copyright SUNY at Stony Brook, 1986; University of Arizona, 1987	*
*									*
************************************************************************/

/*-----------------------------------------------------------------
SB-Prolog is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the SB-Prolog General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
SB-Prolog, but only under the conditions described in the
SB-Prolog General Public License.   A copy of this license is
supposed to have been given to you along with SB-Prolog so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies. 
------------------------------------------------------------------ */
/* meta.c */

#include "builtin.h"

extern double floatval();
extern LONG   makefloat();

b_VAR()
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);  DEREF(op);
   if ( ISNONVAR(op) ) {
      FAIL0;
   }
}

b_NONVAR()
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);   DEREF(op);
   if ( ! ISNONVAR(op) ) {
      FAIL0;
   }
}

b_ATOM()
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);  DEREF(op);
   if ( ! ISATOM(op) ) {
      FAIL0;
   }
}

b_ATOMIC()
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);  DEREF(op);
   if ( ! (ISNUM(op) || ISATOM(op)) ) {
      FAIL0;
   }
}

b_INTEGER()
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);  DEREF(op);
   if ( ! ISINTEGER(op) ) {
      FAIL0;
   }
}

b_REAL()
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);  DEREF(op);
   if ( ! ISFLOAT(op) ) {
      FAIL0;
   }
}

b_STRUCTURE()
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);  DEREF(op);
   if ( ! (ISLIST(op) || (ISCONSTR(op) && (GET_STR_ARITY(op) != 0))) ) {
      FAIL0;
   }
}

b_TERMREP() /* r1 is term (var);  r2 is integer that is its rep */
{
   register LONG      op;
   register LONG_PTR  top;

   op = GREGC(1);  DEREF(op);
   if ( ! unify( MAKEINT(op), GREGC(2)) ) {
      FAIL0;
   }
}

b_FLOOR0()  /* F, I, N */
{
/*  N is bound to a number which gives the direction of conversion:
    if N = 0 then F is bound to a real and I is a variable, while if
    N = 1 then F is a variable and I is bound to an integer.  No
    checking for the above is done.
*/
    register LONG      op1, op2, op3;
    register LONG_PTR  top;

    op1 = GREGC(1);  DEREF(op1);
    op2 = GREGC(2);  DEREF(op2);
    op3 = GREGC(3);  DEREF(op3);
    if ( INTVAL(op3) == 0 ) {    /* F=float, I=free */
       FOLLOW(op2) = MAKEINT((LONG)(floatval(op1)));
       PUSHTRAIL(op2);
    } else {                 /* F=free, I=integer */
       FOLLOW(op1) = makefloat((double)(INTVAL(op2)));
       PUSHTRAIL(op1);
    }
}
