/************************************************************************
*									*
* The SB-Prolog System							*
* Copyright SUNY at Stony Brook, 1986; University of Arizona, 1987	*
*									*
************************************************************************/

/*-----------------------------------------------------------------
SB-Prolog is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the SB-Prolog General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
SB-Prolog, but only under the conditions described in the
SB-Prolog General Public License.   A copy of this license is
supposed to have been given to you along with SB-Prolog so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies. 
------------------------------------------------------------------ */
/* other.c */

#include "builtin.h"
#include  <errno.h>
#include  <sys/types.h>
#include  <netdb.h>
#include  <sys/socket.h>
#include  <netinet/in.h>
/*#include  <arpa/inet.h>*/

#define  STR_LIM 256  /* limit on the length of the name of a constant */
#define  isalpha(x)  (((x >= 'a') && (x <= 'z')) || ((x >= 'A') && (x <= 'Z')))
#define  isdigit(x)  ((x >= '0') && (x <= '9'))

extern float   floatval();
extern LONG    insert();
extern int     d_trace, d_hitrace;
extern int     errno;
extern LONG    maxmem, maxpspace, maxtrail;

FILE  *curr_in, *curr_out;   /* current input, output streams */

typedef union {
   char *name;
   int num;
} call_args;

static call_args   call_arg[10];
static char        s[256];

b_SYSTEM0()  /* r1: a list of int (string) for CShell commands */
{
   register LONG      op1;
   register LONG_PTR  top;
   char               s[256];

   op1 = GREGC(1);  DEREF(op1);
   namestring( GET_STR_PSC(op1), s );
   if ( ! unify( MAKEINT(system(s)), GREGC(2)) ) {
      FAIL0;
   }
}


getgenargs( rno )
int         rno;
{  /* rno is number of register containing list of args
      This routine converts them into array cal_arg, and
      returns the number of args
   */
   register LONG      op2, op3;
   register LONG_PTR  top;
   PSC_REC_PTR        ptr;
   int                i;

   op2 = GREGC(rno);  DEREF(op2);
   i = 1;
   while ( ! ISNIL(op2) ) {
      UNTAG(op2);
      op3 = FOLLOW(op2);  DEREF(op3);
      if ( ISATOM(op3) ) {
         ptr = GET_STR_PSC(op3);
         if ( GET_ETYPE(ptr) == T_ORDI ) {
            namestring( ptr, s );
            call_arg[i].name = s;
         }
         else if (GET_ETYPE(ptr) == T_BUFF) {
            call_arg[i].name = GET_NAME(ptr);
         }
      }
      else if ( ISINTEGER(op3) )
         call_arg[i].num = INTVAL(op3);
      else
         quit( "Unknown syscall argument\n" );
      op2 += 4;
      DEREF(op2);
      i++;
    }
    return(i);
}


b_SYSCALL()  /* r1: call # ; R2: a list of parameters; R3: returned value */
{
    register LONG      op1;
    register LONG_PTR  top;
    int                n, r;

    op1 = GREGC(1);  DEREF(op1);
    n   = INTVAL(op1);              /* syscall number */
    if (n == 33) {                  /* access */
        getgenargs(2);
        r = access(call_arg[1].name, call_arg[2].num);
    }
    else {
        printf("syscall not implemented");
        r = 0;
    }
/* not implemented ... 
    switch ( getgenargs(2) ) {
      case  1: r = syscall(n);
               break;
      case  2: r = syscall(n, call_arg[1]);
               break;
      case  3: r = syscall(n, call_arg[1], call_arg[2]);
               break;
      case  4: r = syscall(n, call_arg[1], call_arg[2], call_arg[3]);
               break;
      case  5: r = syscall(n, call_arg[1], call_arg[2], call_arg[3],
                              call_arg[4]);
               break;
      case  6: r = syscall(n, call_arg[1], call_arg[2], call_arg[3],
                              call_arg[4], call_arg[5]);
               break;
      case  7: r = syscall(n, call_arg[1], call_arg[2], call_arg[3],
                              call_arg[4], call_arg[5], call_arg[6]);
               break;
      default: quit( "Too many arguments for syscall\n" );
               break;
   } */
   if ( ! unify( GREGC(3), MAKEINT(r)) ) {
      FAIL0;
   }
}


b_BROCALL()
{  /* R1: call #;
      R2: buffer containing args in 4 byte fields;
      R3: buffer to put return value in.
   */
   register LONG_PTR  top;
   register LONG      op1, op;
   PSC_REC_PTR        rptr;
   LONG_PTR           aptr;

   op1  = GREGC(1);  DEREF(op1);  /* brocall number */
   op   = GREGC(2);  DEREF(op);
   aptr = (LONG_PTR)GET_NAME(GET_STR_PSC(op));    /* buff with args */
   op   = GREGC(3);  DEREF(op);
   rptr = GET_STR_PSC(op);                        /* buff for result */
   switch ( INTVAL(op1) ) {
   case  2: *(LONG_PTR)GET_NAME(rptr) = (LONG) getenv(*aptr); 
            break;

/* Communication subsystem system calls. Have not included byteorder (ntohl,
   ntohs, htonl, htons). Each call is from manual entry 3N,  except
   getpeername,  which is from 2.
*/

/* case 21: *(LONG_PTR)GET_NAME(rptr) = (LONG)gethostent(); break; */
   case 22: *(LONG_PTR)GET_NAME(rptr) = (LONG)gethostbyname(*aptr);    break;
   case 23: *(LONG_PTR)GET_NAME(rptr) = (LONG)gethostbyaddr(*aptr);    break;
   case 24: *(LONG_PTR)GET_NAME(rptr) = (LONG)sethostent(*aptr);       break;
   case 25: *(LONG_PTR)GET_NAME(rptr) = (LONG)endhostent();            break;

   case 26: *(LONG_PTR)GET_NAME(rptr) = (LONG)getnetent();             break;
   case 27: *(LONG_PTR)GET_NAME(rptr) = (LONG)getnetbyname(*aptr);     break;
   case 28: *(LONG_PTR)GET_NAME(rptr) = (LONG)getnetbyaddr(*aptr);     break;
   case 29: *(LONG_PTR)GET_NAME(rptr) = (LONG)setnetent(*aptr);        break;
   case 30: *(LONG_PTR)GET_NAME(rptr) = (LONG)endnetent();             break;

   case 31: *(LONG_PTR)GET_NAME(rptr) = (LONG)getprotoent();           break;
   case 32: *(LONG_PTR)GET_NAME(rptr) = (LONG)getprotobyname(*aptr);   break;
   case 33: *(LONG_PTR)GET_NAME(rptr) = (LONG)getprotobynumber(*aptr); break;
   case 34: *(LONG_PTR)GET_NAME(rptr) = (LONG)setprotoent(*aptr);      break;
   case 35: *(LONG_PTR)GET_NAME(rptr) = (LONG)endprotoent();           break;

   case 36: *(LONG_PTR)GET_NAME(rptr) = (LONG)getservent();            break;
   case 37: *(LONG_PTR)GET_NAME(rptr) = (LONG)getservbyname(*aptr);    break;
   case 38: *(LONG_PTR)GET_NAME(rptr) = (LONG)getservbyport(*aptr);    break;
   case 39: *(LONG_PTR)GET_NAME(rptr) = (LONG)setservent(*aptr);       break;
   case 40: *(LONG_PTR)GET_NAME(rptr) = (LONG)endservent();            break;

/* case 41: *(LONG_PTR)GET_NAME(rptr) = (LONG)inet_addr(*aptr); break; */
   case 42: *(LONG_PTR)GET_NAME(rptr) = (LONG)inet_network(*aptr);     break;
   case 43: *(LONG_PTR)GET_NAME(rptr) = (LONG)inet_ntoa(*aptr);     break;
/* case 44: *(LONG_PTR)GET_NAME(rptr) = (LONG)inet_makeaddr(*aptr); break; */
   case 45: *(LONG_PTR)GET_NAME(rptr) = (LONG)inet_lnaof(*aptr);       break;
   case 46: *(LONG_PTR)GET_NAME(rptr) = (LONG)inet_netof(*aptr);       break;

/* case 47: *(LONG_PTR)GET_NAME(rptr) = (LONG) get_peername(*aptr); break; */
/* case 50: *(LONG_PTR)GET_NAME(rptr) = (LONG)perror(*aptr);        break; */

   default: printf( "Illegal brocall number\n" );
            FAIL0;
            return;
   }
}

b_ERRNO()
{
   if ( ! unify( GREGC(1), MAKEINT(errno)) ) {
      FAIL0;
   }
}

b_CALL()  /* R1: The predicate to be called */
{
   callv_sub();  /* since cpreg has been saved by call "call",
                    should not be saved again, the same as exec
                 */
}

b_LOAD()
{  /* R1: the byte code file to be loaded
      R2: the return code, 0 => success
   */
   register LONG      op1;
   register LONG_PTR  top;
   PSC_REC_PTR	      psc;

   op1 = GREGC(1);  
   DEREF(op1);
   psc = GET_STR_PSC(op1);
           
   if ( ! unify( MAKEINT(dyn_loader(psc)), GREGC(2)) ) {
      FAIL0;
   }
}

b_STATISTICS()
{
   print_statistics();
}

b_STATISTICS0()
{
   if ( ! unify( GREGC(1), MAKEINT(maxpspace*sizeof(LONG))) ) {
      FAIL0;     /* max program area */
   }
   if ( ! unify( GREGC(2), MAKEINT((((LONG)curr_fence-(LONG)pspace)))) ) {
      FAIL0;     /* program area in use */
   }
   if ( ! unify( GREGC(3), MAKEINT((hreg-heap_bottom))) ) {
      FAIL0;     /* global stack in use */
   }
   if ( ! unify( GREGC(4), MAKEINT((ereg<breg?ereg:breg)-hreg)) ) {
      FAIL0;     /* stack area (local, global) free */
   }
   if ( ! unify( GREGC(5), MAKEINT(local_bottom-(ereg<breg?ereg:breg))) ) {
      FAIL0;     /* local stack in use */
   }
   if ( ! unify( GREGC(6), MAKEINT(maxmem*sizeof(LONG))) ) {
      FAIL0;     /* total stack area size */
   }
   if ( ! unify( GREGC(7), MAKEINT(maxtrail*sizeof(LONG))) ) {
      FAIL0;     /* total trail size */
   }
   if ( ! unify( GREGC(8), MAKEINT(trreg-tstack)) ) {
      FAIL0;
   }
}

b_TRACE()
{
   hitrace = 1;
}

b_PILTRACE()
{
   trace = 1;
}

b_UNTRACE()
{
   hitrace = trace = 0;
}

/*
b_DETRACE()
{
   hitrace = d_hitrace;
   trace = d_trace;
}
*/

b_SYMTYPE()  /* R1 term, R2 type field of psc-entry of root sym of term */
{
   register LONG      op1;
   register LONG_PTR  top;

   op1 = GREGC(1);
typd:
   switch ( op1 & 3 ) {
     case FREE: NDEREF(op1, typd);
     case LIST:
     case NUM : quit( "Symtype: illegal first arg\n" );
     case CS  : if (! unify(MAKEINT(GET_ETYPE(GET_STR_PSC(op1))), GREGC(2))) {
                   FAIL0;
                }
   }
}

b_HASHVAL()  /* R1 Arg, R2 size of hashtab, R3 hashval for this arg */
{
   register LONG      op1, op2, op3;
   register LONG_PTR  top;

   op1 = GREGC(1);
   op2 = GREGC(2);  DEREF(op2);  op2 = INTVAL(op2);
   op3 = GREGC(3);  DEREF(op3);
sotd0:
   switch( op1 & 3 ) {
     case FREE: NDEREF(op1, sotd0);
                printf( "Indexing for asserted predicate with var arg\n" );
                FAIL0;
     case NUM : if ( ISINTEGER(op1) )
                   op1 = INTVAL(op1);
                else
                   op1 = (LONG)(floatval(op1));
                break;
     case LIST: op1 = *((LONG_PTR)UNTAGGED(list_str));
                break;
     case CS  : op1 = (LONG)GET_STR_PSC(op1);
                break;
   }
   if ( ! unify(op3, MAKEINT( IHASH(op1, op2) )) ) {
      FAIL0;
   }
}

b_FLAGS()
{  /* R1 contains number of bit to get or set (must be integer);
      R2 contains setting of 0 or 1,
      or is variable and setting will be returned
   */
   register LONG      op1, op2, res;
   register LONG_PTR  top;

   op1 = GREGC(1);  DEREF(op1);  op1 = INTVAL(op1);
   op2 = GREGC(2);  DEREF(op2);
   if ( ISNONVAR(op2) ) {
      if ( op1 > 9 )
         flags[op1-10] = op2;
      else {
         op2 = INTVAL(op2);
         switch ( op1 ) {
           case 0: trace      = op2;   break;
           case 1: hitrace    = op2;   break;
           case 2: overflow_f = op2;   break;
           case 3: trace_sta  = op2;   break;
         }
         call_intercept = hitrace | trace_sta;
      }
   } else {
      if ( op1 > 9 )
         res = flags[op1-10];
      else {
         switch ( op1 ) {
           case 0: res = trace;       break;
           case 1: res = hitrace;     break;
           case 2: res = overflow_f;  break;
           case 3: res = trace_sta;   break;
         }
         res = MAKEINT(res);
      }
      FOLLOW(op2) = res;
   }
}

print_statistics()
{
    LONG_PTR  lstktop;

    if ( breg < ereg)
       lstktop = breg;
    else
       lstktop = ereg - *(cpreg-5);

    printf( "Maximum available stack size: %d\n", maxmem );
    printf( "  Local stack: %d in use, %d max used.\n",
               local_bottom-lstktop, local_bottom-mlocaltop);
    printf( "  Heap stack: %d in use, %d max used.\n",
               hreg-heap_bottom, mheaptop-heap_bottom);

    printf( "Permanent space: %d, %d in use.\n", maxpspace,
               ((int) curr_fence - (int) pspace)/4);

    printf( "Trail stack: %d, %d in use, %d max used.\n",
               maxtrail, trail_bottom-trreg, trail_bottom-mtrailtop);
}

b_READNAME()
{  /* Ch, Name, NextCh: reads a sequence of letters, digits
      dollar signs and underscores, makes the sequence into a
      constant and inserts it into the PSC table if necessary,
      and returns a pointer to the PSC entry as Name.  NextCh is
      the first character read which cannot join this sequence.

      It is assumed that reg1 contains a character; and that regs
      2 and 3 are free.  No checking is done here.  This builtin
      is only supposed to be called from $read_tokens/3 anyway.
   */

   register LONG      op;
   register LONG_PTR  top;
   char               pname[STR_LIM], *name, ch;
   char               perm = PERM;
   int                len = 1;
   int                done = 0;
   LONG               ptr;

   op = GREGC(1);  DEREF(op);  op = INTVAL(op);

   name = pname;
   *(name++) = (char)op;   /* first character in the sequence */

   while ( ! done && len <= STR_LIM ) {
      ch = getc(curr_in);
      if ( isalpha(ch) || isdigit(ch ) || (ch == '$') || (ch ==  '_')) {
         *(name++) = ch;
         len++;
      } else {
         *name = '\0';
         done = 1;
      }
   }

   if (ch == EOF) {
      clearerr( curr_in );
      printf( "! unexpected end of file after %s\n", pname );
   }
   if (len > STR_LIM) {
      *name = '\0';
      len--;
      printf( "*** Name of constant too long: %s\n", pname );
   }

   ptr = insert( pname, len, 0, &perm ) | CS_TAG;
   if ( ! unify(GREGC(2), ptr) ) {
       FAIL0;
   }
   op = GREGC(3);  DEREF(op);
   FOLLOW(op) = MAKEINT(ch);
   PUSHTRAIL(op);
}
