/************************************************************************
*									*
* The SB-Prolog System							*
* Copyright SUNY at Stony Brook, 1986; University of Arizona, 1987	*
*									*
************************************************************************/

/*-----------------------------------------------------------------
SB-Prolog is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the SB-Prolog General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
SB-Prolog, but only under the conditions described in the
SB-Prolog General Public License.   A copy of this license is
supposed to have been given to you along with SB-Prolog so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies. 
------------------------------------------------------------------ */
#include "sim.h"
#include "aux.h"

extern double  floatval();
extern LONG    makefloat();
/***********************************************************************/
static  eval_error( functor, arity )
CHAR_PTR            functor;
int                 arity;
{
   printf( "Error: unknown functor %s/%d in arithmetic expression\n",
                   functor, arity);
}
static  integer_required( functor )
CHAR_PTR                  functor;
{
   printf( "%s: integer required\n", functor );
}
/***********************************************************************/
eval( op, val)
LONG  op;
LONG_PTR  val;
{
   register  LONG_PTR  top;
   PSC_REC_PTR         p;
   CHAR_PTR            functor;
   int                 arity, length, floatp;
   LONG                op1, op2;
   double              result;

   switch (op & 0x3) {
    case FREE:
       printf( "Error: unbound variable in arithmetic expression\n" );
       FAIL0;
       return -1;
    case CS  :
       floatp = 0;
       p = GET_STR_PSC(op);
       functor = GET_NAME(p);
       arity   = GET_ARITY(p);
       length  = GET_LENGTH(p);
       if (length == 1)
         switch (*functor) {
         case '+':
            switch (arity) {
              case 1  : op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
                        return eval(op1, val);
              case 2  : op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
                        floatp = eval(op1, &op1);
                        op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
                        floatp = floatp | eval(op2, &op2);
                        if (floatp >= 0)
                          *val = MAKENUM(NUMVAL(op1) + NUMVAL(op2));
                        return floatp;
              default : eval_error( functor, arity );
                        FAIL0;
                        return -1;
            }  /* switch arity */
         case '-':
            switch (arity) {
              case 1  : op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
                        floatp = eval(op1, &op1);
                        if (floatp >= 0) {
                           *val = MAKENUM(-(NUMVAL(op1)));
                           return floatp;
                        }
              case 2  : op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
                        floatp = eval(op1, &op1);
                        op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
                        floatp = floatp | eval(op2, &op2);
                        if (floatp >= 0)
                           *val = MAKENUM(NUMVAL(op1) - NUMVAL(op2));
                        return floatp;
              default : eval_error( functor, arity );
                        FAIL0;
                        return -1;
            } /* switch (arity) */
         case '*':
            if (arity == 2) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (floatp >= 0)
                  *val = MAKENUM(NUMVAL(op1) * NUMVAL(op2));
               return floatp;
            } else {
               eval_error( functor, arity );
               FAIL0;
               return -1;
            }
         case '/':
            if (arity == 2) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (NUMVAL(op2) == 0) {
                  printf("div: division by zero! ... failing execution ...\n");
                  return -1;
               }
               if (floatp >= 0) {
                  result = ((double)(NUMVAL(op1)))/((double)(NUMVAL(op2)));
                  if (INTEGRAL(result)) {
                     floatp = 0;
                     *val = MAKEINT(result);
                  } else {
                     floatp = 1;
                     *val = makefloat(result);
                  }
                  return floatp;
               }
            } else {
               eval_error( functor, arity );
               FAIL0;
               return -1;
            }
        case '\\':
            if (arity == 1) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               if (floatp == 0) {
                  *val = MAKEINT(~(INTVAL(op1)));
                  return floatp;
               } else {
                  integer_required( "neg" );
                  return -1;
               }
            } else {
               eval_error( functor, arity );
               FAIL0;
               return -1;
            }
         }
         else if (arity == 2) {
            if ( !strcmp( functor, "//" ) && arity == 2 ) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (floatp >= 0)
                  *val = MAKEINT(((NUMVAL(op1) / NUMVAL(op2))));
               return 0;
            }
            else if ( !strcmp( functor, "mod" ) && arity == 2 ) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (floatp == 0) {
                  *val = MAKEINT(INTVAL(op1) % INTVAL(op2));
                  return 0;
               } else {
                  integer_required( "mod" );
                  return -1;
               }
            }
            else if ( !strcmp( functor, "/\\" ) && arity == 2 ) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (floatp == 0) {
                  *val = MAKEINT(INTVAL(op1) & INTVAL(op2));
                  return 0;
               } else {
                  integer_required( "and" );
                  FAIL0;
                  return -1;
               }
            }
            else if ( !strcmp( functor, "\\/" ) && arity == 2 ) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (floatp == 0) {
                  *val = MAKEINT(INTVAL(op1) | INTVAL(op2));
                  return 0;
               } else {
                  integer_required( "or" );
                  FAIL0;
                  return -1;
               }
            }
            else if ( !strcmp(functor, ">>" ) && arity == 2 ) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (floatp == 0) {
                   *val = MAKEINT(INTVAL(op1) >> INTVAL(op2));
                   return 0;
               } else {
                  integer_required( "rshift" );
                  FAIL0;
                  return -1;
               }
            }
            else if ( !strcmp( functor, "<<" ) && arity == 2 ) {
               op1 = *(((LONG_PTR)(UNTAG(op))) + 1);   DEREF(op1);
               floatp = eval(op1, &op1);
               op2 = *(((LONG_PTR)(UNTAG(op))) + 2);   DEREF(op2);
               floatp = floatp | eval(op2, &op2);
               if (floatp == 0) {
                  *val = MAKEINT(INTVAL(op1) << INTVAL(op2));
                  return 0;
               } else {
                  integer_required( "lshift" );
                  FAIL0;
                  return -1;
               }
            }
            else {
               eval_error( functor, arity );
               FAIL0;
               return -1;
            }
         }
         else {
            eval_error( functor, arity );
            FAIL0;
            return -1;
         }
    case NUM:
       *val = op;
       return ISFLOAT(op);
    case LIST:
       FAIL0; return -1;
   }
}
