/************************************************************************
*									*
* The SB-Prolog System							*
* Copyright SUNY at Stony Brook, 1986; University of Arizona, 1987	*
*									*
************************************************************************/

/*-----------------------------------------------------------------
SB-Prolog is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the SB-Prolog General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
SB-Prolog, but only under the conditions described in the
SB-Prolog General Public License.   A copy of this license is
supposed to have been given to you along with SB-Prolog so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies. 
------------------------------------------------------------------ */
/* init.c */

#include "sim.h"
#include "inst.h"
#include "aux.h"
#include <string.h>
/*
#define  DEBUG
*/
LONG  maxmem    = 150000;  
LONG  maxpspace = 100000;
LONG  maxtrail;

int  d_trace;
int  d_hitrace;

extern char *malloc();

/****************************************************************************/

init_sim( argc, argv )
int       argc;
char     *argv[];
{
   int        i, itmp; 
   LONG       tempint;
   char       perm = PERM;
   WORD_PTR   temp, insert();
   char       c, *charp;

   /* for debugging malloc on sun (include /usr/lib/debug/malloc.o in link)
      malloc_debug(2);
   */

   for ( i = 0; i < BUCKET_CHAIN; i++ ) {
      MAKE_FREE( LONG_PTR, hash_table[i][TEMP] );
      MAKE_FREE( LONG_PTR, hash_table[i][PERM] );
   }


   trace = hitrace = disassem = num_line = 0;
   trace_sta = d_trace = d_hitrace = call_intercept = 0;
   interrupt_code = 0;

   maxtrail = maxmem / 5;

   if (argc == 1)
      quit("Usage: sim [-Ttdns] [-m s_size] [-p p_size] [-b tr_size] [-ui num] pil_file_name ...\n");

   for( i=1; i < argc;  i++ ) {
      if (*argv[i] == '-') {
         charp = argv[i];
         while (c = *++charp) {
            switch(c) {
              case 't': trace = d_trace = 1;
                        break;
              case 'T': hitrace = d_hitrace = 1;
              case 's': trace_sta = 1;
                        call_intercept = 1;
                        break;
              case 'n': num_line = 1;
                        break;
              case 'd': disassem = 1;
                              break;
              case 'm': i++;
                        sscanf(argv[i], "%d", &maxmem);
                        maxtrail = maxmem / 5;
                        break;
              case 'p': i++;
                        sscanf(argv[i], "%d", &maxpspace);
                        break;
              case 'b': i++;
                        sscanf(argv[i], "%d", &maxtrail);
                        break;
              case 'u': i++; /* skip user field on this pass */
                        charp++;
                        break;
              default : printf("Unknown option %c\n", c);
            }  /* switch */
         }  /* while */
      }  /* if */
   }  /* for */

   pspace = (LONG_PTR) (malloc(maxpspace*sizeof(LONG)));
   if (!pspace) quit("Not enough core!\n");
   max_fence = (LONG_PTR) (pspace+maxpspace);

   memory = (LONG_PTR) (malloc(maxmem*sizeof(LONG)));
   if (!memory) quit("Not enough core!\n");

   tstack = (LONG_PTR) (malloc(maxtrail*sizeof(LONG)));
   if (!tstack) quit("Not enough core!\n");
   heap_bottom  = memory;
   local_bottom = memory + maxmem;
   trail_bottom = tstack + maxtrail;

   breg  = ereg      = mlocaltop = local_bottom - 2;
   hreg  = mheaptop  = heap_bottom;
   trreg = mtrailtop = trail_bottom - 1;
   hbreg = heap_bottom - 1;
   trap_vector[0] = pspace;
   curr_fence     = (CHAR_PTR)pspace;

   /* install fail and halt instructions     */
   /* format is  opcode + pad  (word + word) */
   *((WORD_PTR)curr_fence) = fail;   curr_fence += sizeof( WORD );
   *((WORD_PTR)curr_fence) =    0;   curr_fence += sizeof( WORD );

   cpreg = (LONG_PTR)curr_fence;                /* halt on final success */
   *(local_bottom - 1) = (LONG)curr_fence;      /* halt on final failure */

   *((WORD_PTR)curr_fence) = halt;   curr_fence += sizeof( WORD );
   *((WORD_PTR)curr_fence) =    0;   curr_fence += sizeof( WORD );

   temp = insert("[]", 2, 0, &perm);
   nil_sym = (LONG)temp | CS_TAG;

   temp = insert(".", 1, 2, &perm);
   list_str = (LONG)temp | CS_TAG;
   list_psc = (PSC_REC_PTR)FOLLOW(temp);

   temp = insert(",", 1, 2, &perm);
   comma_psc = (PSC_REC_PTR)FOLLOW(temp);

   temp = insert("_$interrupt", 11, 2, &perm);
   interrupt_psc = (PSC_REC_PTR)FOLLOW(temp);
   trap_vector[1] = 0;
   inst_begin = 0;

   /* now strip off user parameters */
   for ( i=0;  i < 10;  i++ ) { flags[i] = nil_sym; }
   for ( i=1;  i < argc; i++ ) {
      if (*argv[i] == '-') {
         charp = argv[i];
         while (c = *++charp) {
            switch(c) {
              case 'p':
              case 'b':
              case 'm': i++;
              case 't':
              case 'T':
              case 's':
              case 'n':
              case 'd': break;
              case 'u': itmp = *++charp - '0';
                        i++; /* value; retrieved by flags(index+10,V) */
                        if ( *argv[i] >= 48 && *argv[i] <= 57 ) { /*num*/
                           sscanf( argv[i], "%d", &tempint);
                           flags[itmp] = MAKEINT(tempint);
                        } else {   /* make it a constant */
                           temp = insert(argv[i], strlen(argv[i]), 0, &perm);
                           flags[itmp] = (LONG)temp | CS_TAG;
                        }
                        break;
            } /* switch */
         } /* while */
      } /* if */
   } /* for */

#ifdef  DEBUG
   printf( "init_sys\n" );
   printf( "pspace       = %d\n", pspace );
   printf( "max_fence    = %d\n", max_fence );
   printf( "memory       = %d\n", memory  );
   printf( "tstack       = %d\n", tstack );
   printf( "heap_bottom  = %d\n", heap_bottom );
   printf( "local_bottom = %d\n", local_bottom );
   printf( "trail_bottom = %d\n", trail_bottom  );
   printf( "breg         = %d\n", breg );
   printf( "ereg         = %d\n", ereg );
   printf( "mlocaltop    = %d\n", mlocaltop );
   printf( "hreg         = %d\n", hreg );
   printf( "mheaptop     = %d\n", mheaptop );
   printf( "trreg        = %d\n", trreg );
   printf( "mtrailtop    = %d\n", mtrailtop );
   printf( "hbreg        = %d\n", hbreg  );
   printf( "curr_fence   = %d\n", curr_fence );
   printf( "cpreg        = %d\n", cpreg  );
   printf( "list_str     = %d\n", list_str );
   printf( "list_psc     = %d\n", list_psc );
   printf( "comma_psc    = %d\n", comma_psc  );
#endif

}  /* init_sym */

/****************************************************************************/

init_loading( argc, argv )
int           argc;
char         *argv[];
{
   char  c;
   int   i;
   int   n = 0; /* number of the initial loaded files */

   for ( i=1; i < argc; i++ ) {
      if ( *argv[i] == '-' ) {
         c = *(argv[i] + 1);
         switch( c ) {
           case 'u':
           case 'm':
           case 'p':
           case 'b': i++;
           default : break;
         }
      } else {
         if ( loader(argv[i]) ) 
            quit( "Error in loading initial files\n" );
         n++;
      }
   }  /* for */
   if ( n == 0 ) 
      quit( "No input file!\n" );

}  /* init_system */

/****************************************************************************/
