/************************************************************************
*									*
* The SB-Prolog System							*
* Copyright SUNY at Stony Brook, 1986; University of Arizona, 1987	*
*									*
************************************************************************/

/*-----------------------------------------------------------------
SB-Prolog is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the SB-Prolog General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
SB-Prolog, but only under the conditions described in the
SB-Prolog General Public License.   A copy of this license is
supposed to have been given to you along with SB-Prolog so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies. 
------------------------------------------------------------------ */
/* load_work.c */
/* #define  DEBUG_LOADWORK  */
#define  DEBUG_OVERFLOW
/*
#define  DEBUG_NAMESTRING
*/

#include "sim.h"
#include "aux.h"
#include <stdio.h>

#define  ALIGN(type,ptr) ptr = (type)(((LONG)ptr + 3) & 0xfffffffc)

/************************************************************************
*                                                                       *
*  The hash function uses the arity and character string associated     *
*  with a predicate, constant, or structure to find the proper bucket   *
*  (a bucket is a linked list within the pcs table) to insert or locate *
*  pcs entries.                                                         *
*                                                                       *
************************************************************************/

int hash( name, length, arity)  /* hashing function on name,returning a */
CHAR_PTR  name;                 /* bucket number in the hash table      */
WORD      length;
BYTE      arity;
{
   int   bucknum;

   bucknum = arity+1;
   if ( length > 0 ) {   /* first */
      bucknum = bucknum + *name;
      if ( length > 1 ) {   /* last */
         bucknum = (bucknum << 2) + *(name+length-1);
         if (length > 2) {  /* middle */
            bucknum = (bucknum << 2) + *(name+length/2);
            if (length > 3) {
               bucknum = (bucknum << 2) + *(name+(length/2)-1);
            }
         }
      }
   }
   return (abs(bucknum % BUCKET_CHAIN));

}  /* hash */
/****************************************************************************/

LONG_PTR  search( name, length, arity, hash_ptr )
CHAR_PTR          name;
BYTE              arity;
WORD              length;
LONG_PTR          hash_ptr;
{
   PSC_REC_PTR  hash_psc;

   struct booleans {
      unsigned eq   : 1;
      unsigned stop : 1;
   } flag;

   unsigned short i;

   flag.eq   = FALSE;
   flag.stop = FALSE;

#ifdef  DEBUG_LOADWORK
   printf( "search: name = %s   len = %d   arity = %d  ", name, length, arity );
   printf( "hash = %08x\n", hash_ptr );
#endif

   while (!ISFREE(hash_ptr) && flag.stop == FALSE) {

      hash_ptr = (LONG_PTR)FOLLOW(hash_ptr);    /* pointer to pair */
      hash_psc = (PSC_REC_PTR)FOLLOW(hash_ptr); /* pointer to psc ent */

#ifdef  DEBUG_LOADWORK
      printf( "   hash_ptr  %08x  hash_psc  %08x     *hash_ptr  %08x  \n",
                  hash_ptr, hash_psc, *hash_ptr );
#endif

      if ( (arity  == GET_ARITY(hash_psc) ) &&
           (length == GET_LENGTH(hash_psc) ) ) {
         flag.eq = TRUE;
         for ( i=0; ((i < length) && (flag.eq == TRUE)); i++ ) {
            if ( *(name + i) != *(GET_NAME(hash_psc) + i) ) {
                flag.eq = FALSE;
            }
         } /* for */
      }
      if (flag.eq == TRUE) flag.stop = TRUE;
      else hash_ptr++;
  }

  return hash_ptr;

}   /* search */
/****************************************************************************/
LONG_PTR  insert_temp( name, length, hash_ptr)
CHAR_PTR               name;
WORD                   length;
LONG_PTR               hash_ptr;
{
   PSC_REC_PTR       temp;
   LONG_PTR          new_pair;
   register CHAR_PTR threg;
   int               i;

   new_pair = hreg++;
   FOLLOW(hash_ptr) = (LONG)new_pair;
   PUSHTRAIL((LONG)hash_ptr);         /* trail for backtracking */
   MAKE_FREE(LONG, *hreg );           /* 2nd of pair free */
   FOLLOW(new_pair) = (LONG)++hreg;   /* 1st of pair points to psc_rec */
   temp = (PSC_REC_PTR)hreg;          /* temp points to the psc entry */
   hreg += 2;                         /* reserve the space on the heap */
                                      /*   NO EP FIELD */

   /* make nameptr point to next available space on heap */

   GET_ETYPE(temp)  = T_ORDI;
   GET_ARITY(temp)  = 0;
   GET_LENGTH(temp) = length;
  /*  GET_NAME(temp) = name; */

   threg = (CHAR_PTR)hreg;
   GET_NAME(temp) = threg;     /* copy name, since might write over it !! */
   for (i=0; i < length; i++)
      *(threg++) = *(name++);

   hreg = (LONG_PTR)threg;
   ALIGN( LONG_PTR, hreg );

   return new_pair;

}  /* insert_temp */

LONG_PTR  insert_perm( name, length, arity, hash_ptr )
CHAR_PTR               name;
BYTE                   arity;
WORD                   length;
LONG_PTR               hash_ptr;
{
   PSC_REC_PTR    temp;
   LONG_PTR       new_pair;
   register int   i;

   ALIGN( CHAR_PTR, curr_fence);

   new_pair = (LONG_PTR)curr_fence;
   FOLLOW(hash_ptr) = (LONG)new_pair;       /* prev link to here */
   curr_fence += 4;                         /* point to 2nd of pair */
   MAKE_FREE(LONG_PTR,*(LONG_PTR*)curr_fence);          /* set 2nd free */

#ifdef  DEBUG_LOADWORK
   printf("insert_perm %8x %8x\n", curr_fence, *(LONG_PTR)curr_fence);
#endif

   curr_fence += 4;                         /* where we'll put psc_rec */
   FOLLOW(new_pair) = (LONG)curr_fence;     /* set 1st to point to psc_rec */
   temp = (PSC_REC_PTR)curr_fence;          /* temp points there too */
   curr_fence += 12;                        /* 12 bytes for psc_rec */

   GET_ETYPE(temp)  = T_ORDI;
   GET_ARITY(temp)  = arity;
   GET_LENGTH(temp) = length;
   GET_NAME(temp)   = curr_fence;

   for ( i=0;  i < length;  i++) *(curr_fence++) = *(name++);

   ALIGN( CHAR_PTR, curr_fence);

   if ((LONG_PTR)curr_fence > max_fence) {
#ifdef DEBUG_OVERFLOW
   printf( "Overflow in \"insert_perm\" curr_fence = %08x  max_fence = %08x\n",
            curr_fence, max_fence );
#endif
       quit("Program area overflow\n");
   }

   return new_pair;

}  /* insert_perm */
/****************************************************************************/

LONG_PTR  insert( name, length, arity, perm )
CHAR_PTR          name;
WORD              length;
BYTE              arity;
BYTE_PTR          perm;
{
   int       bucket_no;
   LONG_PTR  temp_ptr, perm_ptr, perm_hashptr, temp_hashptr, ret_ptr;


   bucket_no = hash( name, length, arity);
   perm_hashptr = (LONG_PTR)&hash_table[bucket_no][PERM];

#ifdef  DEBUG_LOADWORK
   printf( "insert: name = %s   len = %d   arity = %d  ", name, length, arity );
   printf( "bucket = %d  perm = %08x  hash = %08x\n", bucket_no, perm_hashptr,
                *perm_hashptr );
#endif

   perm_ptr = search (name, length, arity, perm_hashptr);
   if (!ISFREE(perm_ptr)) {   /* found perm psc record */
      if (!(*perm))
         *perm = PERM;        /* set perm flag parameter */
      return perm_ptr;        /* return permanent */
   }

   temp_hashptr = (LONG_PTR)&hash_table[bucket_no][TEMP];  /* look for temp */
   temp_ptr = search(name, length, arity, temp_hashptr);
   if (!ISFREE(temp_ptr)) {  /* found temp psc record */
      if (!(*perm)) {        /* Temporary wanted */
         return(temp_ptr);   /* return ptr to psc record */
      } else {               /* Perm wanted - convert temp to perm */
         /*printf("cvting temp to perm: %c %d %d\n", *name, length, arity);*/
         perm_ptr = insert_perm(name,length,arity,perm_ptr);
         FOLLOW(temp_ptr) = FOLLOW(perm_ptr);
         return(perm_ptr);
      }
    } else {                 /* Insert constant where indicated */
      if (*perm)
         return insert_perm(name, length, arity, perm_ptr);
      else
         return insert_temp(name, length, temp_ptr);
   }
}  /* insert */
/************************************?????????????????????*********/
set_temp_ep( pscrec, ep )
PSC_REC_PTR  pscrec;
{
   if (ep >= 0) {
      GET_ETYPE(pscrec) = T_TEMP_PRED;
      GET_EP(pscrec) = (LONG_PTR)ep;
   }
}

set_real_ep( psc_p, base )
PSC_REC_PTR  psc_p;
char        *base;
{
   if (GET_ETYPE(psc_p) == T_TEMP_PRED) {
      GET_EP(psc_p) = (LONG_PTR) (base + (int)GET_EP(psc_p)); /*???*/
      GET_ETYPE(psc_p) = T_PRED;
   }
}

/*
set_file_ptr( pscrec, file_ptr )
PSC_REC_PTR   pscrec;
FILE           *file_ptr;
{
   GET_ETYPE(pscrec) = T_FILE;
   GET_EP(pscrec) = (WORD_PTR) (file_ptr);
}
*/

/*
unset_file_ptr( pscrec )
PSC_REC_PTR     pscrec;
{
   GET_ETYPE(pscrec) = T_ORDI;
   GET_EP(pscrec) = 0;
}
*/

namestring( p, s )
PSC_REC_PTR p;
char       *s;
{
   int    i, len;
   char  *st;

   len = GET_LENGTH(p);
   st  = GET_NAME(p);

#ifdef  DEBUG_NAMESTRING
   printf( "namestring: len = %d    string = %s\n", len, st );
#endif
   for ( i=0; i < len; i++) {
      *(s++) = *(st++);
   }
   *s = '\0';
}

alloc_perm( size, baddr )   /* size should be a multiple of 4 */
LONG_PTR         *baddr;
{
   PSC_REC_PTR  ptr;

   ALIGN( CHAR_PTR, curr_fence);
   *((LONG_PTR)curr_fence) = (LONG)(curr_fence) + 4; /* fake 1st of pair */
   *baddr = (LONG_PTR)curr_fence;
   curr_fence  += 4;
   ptr = (PSC_REC_PTR)curr_fence;
   GET_ETYPE(ptr)  = T_BUFF;
   GET_ARITY(ptr)  = 0;
   GET_LENGTH(ptr) = size;
   GET_NAME(ptr)   = (curr_fence += 8);      /* no ep */
   curr_fence += size;
   ALIGN( CHAR_PTR, curr_fence);
   if ((LONG_PTR)curr_fence > max_fence) {
#ifdef DEBUG_OVERFLOW
   printf( "Overflow in \"alloc_perm\" curr_fence = %08x  max_fence = %08x\n",
            curr_fence, max_fence );
#endif
      quit("Program area overflow\n");
   }
}  /* alloc_perm */
